Require Import scap_rules.
Require Import scap.
Require Import cdsa.
Require Import heap2.
Require Import lib_lemma.
Require Import Omega.
Require Import properties.
Require Import properties2.
Require Import copydefs.
Require Import copyprops.
Require Import malloc.
Require Import List.

Axiom MYTOP : TOP > 10.

Notation psi_copy_entry := (
  consCT L_copy_test T_copy_test emptyCT).

Notation psi_copy_test := (consCT 
  L_copy_new T_copy_new (consCT 
  L_copy_halt T_copy_halt emptyCT)).

Notation psi_copy_new := (consCT
  L_malloc_prolog T_malloc_prolog (consCT
  L_copy_del T_copy_del emptyCT)).

Notation psi_copy_del := (consCT
  L_copy_halt T_copy_halt (consCT
  L_copy_reorder_t T_copy_reorder_t emptyCT)).

Notation psi_copy_reorder_t := (consCT
  L_free_prolog T_free_prolog (consCT
  L_copy_ahead T_copy_ahead emptyCT)).

Notation psi_copy_ahead := (consCT 
  L_copy_lnkprv T_copy_lnkprv (consCT
  L_copy_test T_copy_test emptyCT)).

Notation psi_copy_lnkprv := (consCT
  L_copy_test T_copy_test emptyCT).

Notation psi_copy_halt := (consCT
  L_copy_halt T_copy_halt emptyCT).

(**************************************************)

Theorem WF_copy_halt :
  Infer (itpx2 (itpx1 T_copy_halt) psi_copy_halt) C_copy_halt.
Proof.
unfold C_copy_halt.
apply WF_j with P_copy_halt G_copy_halt.
constructor. intros; auto. intros; auto.
Save.

(**************************************************)

Theorem WF_copy_lnkprv :
  Infer (itpx2 (itpx1 T_copy_lnkprv) psi_copy_lnkprv) C_copy_lnkprv.
Proof.
unfold C_copy_lnkprv.
(* sw prnew 1 s3**)
apply WF_sw with 
  (fun s : state => COPY_FRAME s /\ getValinR s prnew <> pnil /\
    exists fv : addr, exists key : nat, exists key' : nat, 
      exists l1 : lint, exists l2 : lint, 
      star (pto flist fv) (
	star (Good fv) (
	  star (Stk TOP) (
	    star (Pair (getValinR s s3) key (getValinR s src)) (
	      star (Slist l1 (getValinR s src) 0) (
	      star (Slist l2 (getValinR s tgt) (getValinR s prnew)) 
	      (Pair (getValinR s prnew) key' (getValinR s s3))))))) s)
  NOG.
(** prnew->next in dom **)
unfold P_copy_lnkprv.
intros. simplRf H. splitland. splitex.
repeat splitstar. inrh. inrh. inrh. inrh. inrh. inrh. pairindom1.
(** P S -> P' nextS **)
unfold P_copy_lnkprv. intros. induction s as [h rf].
inversion H0. simplRf H; simplRf'. splitland; getland.
splitex. exists x; exists x0; exists x1; exists x3; exists x4.
splitstar. writerh. repeat splitstar. inrh. inrh. inrh. inrh. inrh. pairindom1.
splitstar. writerh. repeat splitstar. inrh. inrh. inrh. inrh. pairindom1.
splitstar. writerh. repeat splitstar. inrh. inrh. inrh. pairindom1.
splitstar. writerh. repeat splitstar. inrh. inrh. pairindom1.
splitstar. writerh. repeat splitstar. inrh. pairindom1.
splitstar. writerh. pairindom1. p2pw1. unfold G_copy_lnkprv. intros; auto.
(** addiu prnew s3 0 **)
apply WF_addiu with 
  (fun s : state => COPY_FRAME s /\ getValinR s prnew <> pnil /\
    exists fv : addr, exists key : nat, 
      exists l1 : lint, exists l2 : lint, 
      star (pto flist fv) (
	star (Good fv) (
	  star (Stk TOP) (
	    star (Pair (getValinR s prnew) key (getValinR s src)) (
	      star (Slist l1 (getValinR s src) 0)
	      (Slist l2 (getValinR s tgt) (getValinR s prnew)))))) s)
  NOG.
intros. induction s as [h rf]. inversion H0. simpl0p.
simplRf H; simplRf'. splitland; getland. splitex. repeat splitstar.
move H20 after H28. pairnot0. splitex.  exists x; exists x0.
exists x2. exists (app x3 (cons x1 nil)). rfinsen. repeat(intros; rfinsen).
intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
splitstar. getstar'.  splitstar. getstar'. splitstar. getstar'. 
splitstar. getstar'. splitstar. getstar'. splitstar.
apply SlistCons' with x14 x15 (lookupR rf prnew); auto. 
intros; auto.
(**j L_copy_test **)
apply WF_j with P_copy_test G_copy_test.
constructor.
unfold P_copy_test. intros. right.
splitland; getland. splitex. exists x; exists x0; exists x1; exists x2.
induction s as [h rf]. splitstar; getstar'. splitstar; getstar'. 
splitstar; getstar'. assocrgoal. starrev'. trivial.
intros; auto.
Save.


(**************************************************)

Theorem WF_copy_ahead :
  Infer (itpx2 (itpx1 T_copy_ahead) psi_copy_ahead) C_copy_ahead.
Proof.
unfold C_copy_ahead.
(** bgtz prnew L_copy_lnkprv ***)
apply WF_bgtz with P_copy_lnkprv 
  (fun s : state => COPY_FRAME s /\ getValinR s prnew = pnil /\
    (exists fv : addr, exists key : nat, exists l : lint,
      star (pto flist fv) (star (Good fv) (star (Stk TOP) (
	star (Pair (getValinR s s3) key (getValinR s src)) 
	(Slist l (getValinR s src) 0)))) s))
  
  NOG NOG.
constructor.
(**P => P' **) Focus 2.
unfold P_copy_ahead; unfold P_copy_lnkprv.
intros. simplRf H; simplRf'.  assert (lookupR rf prnew <> 0); [omega|idtac].
inversion_clear H. splitland; contradiction. unfold FR_OTH2 in H2.
splitland. splitex. getland. exists x. repeat (splitstar).
splitex. exists x6; exists x7; exists x8; exists x9; exists x10.
getstar'. getstar'. getstar'.
(*** P => G' => G **) Focus 2.
unfold G_copy_ahead; intros; auto.
(*** <=0 => P => P'' **) Focus 2. 
unfold P_copy_ahead. intros. simplRf H; simplRf'.
assert (lookupR rf prnew = 0); [omega|idtac]. inversion_clear H.
splitland; getland. splitex. exists x. unfold FR_OTH1 in H.
repeat splitstar. splitex. exists x6; exists x7. getstar'.
getstar'. getstar'. splitland; contradiction.
(** <=0 => P => G'' => G**) Focus 2.
unfold G_copy_ahead; intro; auto.
(**addiu tgt s3 0 **)
apply WF_addiu with 
  (fun s : state => COPY_FRAME s /\ getValinR s prnew = pnil /\
    (exists fv : addr, exists key : nat, exists l : lint,
      star (pto flist fv) (star (Good fv) (star (Stk TOP) (star 
	(star (Slist nil (getValinR s tgt) (getValinR s tgt))
	  (Pair (getValinR s tgt) key (getValinR s src)))
	(Slist l (getValinR s src) 0)))) s))
  NOG.
intros. induction s as [h rf]. inversion H0.
simplRf H. simplRf'. splitland. getland. simpl0p.
splitex. exists x; exists x0; exists x1.
rfinsen. repeat (intros; rfinsen). 
intros; apply SlistInsen with rf; auto.
intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
splitstar; getstar'. splitstar; getstar'. splitstar; getstar'. 
splitstar; getstar'. unfold star. exists hempty; exists x8.
getland. apply disj_emp. apply append_emp.
constructor. unfold emp. unfold hempty. ungetVinH. intros; auto.
intros; auto.
(**addiu prnew tgt 0 **)
apply WF_addiu with
  (fun s : state => COPY_FRAME s /\ getValinR s prnew <> 0 /\
    (exists fv : addr, exists key : nat, exists l : lint,
      star (pto flist fv) (star (Good fv) (star (Stk TOP) (star 
	(star (Slist nil (getValinR s tgt) (getValinR s prnew))
	  (Pair (getValinR s prnew) key (getValinR s src)))
	(Slist l (getValinR s src) 0)))) s))
  NOG.
intros. induction s as [h rf]. inversion H0. 
simpl0p. simplRf H; simplRf'. splitland.  getland.
splitex. repeat splitstar. pairnot0. splitex.
exists x; exists x0; exists x1. repeat (intros; rfinsen).
intros; apply SlistInsen with rf; auto.
intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto. intros; auto.
(**j L_copy_test **)
apply WF_j with P_copy_test G_copy_test.
constructor; [constructor |omega].
unfold P_copy_test.
intros. right. splitland; getland. splitex.
exists x. exists x0. exists x1. exists (nil: lint). 
induction s as [h rf]. splitstar; getstar'.
splitstar; getstar'. splitstar; getstar'. starrev H12.
splitstar; getstar'. intros; auto.
Save.


(**************************************************)

Theorem WF_copy_reorder_t :
  Infer (itpx2 (itpx1 T_copy_reorder_t) psi_copy_reorder_t) C_copy_reorder_t.
Proof. 
unfold C_copy_reorder_t.
apply WF_addiu with 
  (fun s : state => COPY_FRAME s /\ getValinR s tgt = pnil /\ getValinR s prnew =pnil /\
    (exists fv : addr, exists lnk : addr, exists siz : nat,
      star (MBlk (getValinR s a0) lnk siz) (star (pto flist fv) (
	star (Good fv) (star (Stk TOP) (FR_OTH1 (getValinR s s3)(getValinR s src))))) s)
    \/
    COPY_FRAME s /\ getValinR s prnew <> pnil /\ (exists fv : addr, exists lnk : addr,
      exists siz : nat, star (MBlk (getValinR s a0) lnk siz) (star (pto flist fv) (
	star (Good fv) (star (Stk TOP) 
	  (FR_OTH2 (getValinR s s3) (getValinR s src) 
	    (getValinR s tgt) (getValinR s prnew))))) s))
  NOG.
unfold P_copy_reorder_t.
intros. induction s as [h rf]. inversion H0.
simplRf H; simplRf'. inversion_clear H. left.
splitland; getland. splitex. exists x; exists x0; exists x1.
simpl0p. repeat (intros; rfinsen). unfold FR_OTH1. intros. 
splitex. exists x2; exists x3. repeat (intros; rfinsen). 
intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
right. splitland. getland. splitex. exists x; exists x0; exists x1.
simpl0p. repeat (intros; rfinsen). unfold FR_OTH2. intros.
splitex. exists x2; exists x3; exists x4; exists x5; exists x6.
repeat (intros; rfinsen). 
intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
intros; apply PairInsen with rf; auto.
unfold G_copy_reorder_t. intros; auto.
(** jal L_free_prolog L_copy_ahead **)
apply WF_jal with P_free_prolog P_copy_ahead G_free_prolog G_copy_ahead.
constructor. constructor; [constructor|omega].
(** P => P' (ra~>_)**)
unfold P_free_prolog. intros. simplRf H; simplRf'.
inversion_clear H. splitland. assert (TOP>10); [apply MYTOP|idtac].
rewrite H2; rewrite H4. getland.  omega. omega. omega.
splitex. exists x; exists x0; exists x1. rfinsen. 
intros; rfinsen. repeat (intros; rfinsen).
splitstar; getstar'. splitstar; getstar'. splitstar; getstar'.
splitstar; getstar'. unfold arbh. auto.
assert (TOP>10); [apply MYTOP|idtac].
splitland;getland. omega. omega. omega. 
splitex. exists x; exists x0; exists x1.
rfinsen. intros; rfinsen. repeat (intros; rfinsen).
splitstar; getstar'. splitstar; getstar'. splitstar; getstar'. 
splitstar; getstar'. unfold arbh; auto.
(** P => G' => P'' ***)
unfold G_free_prolog; unfold P_copy_ahead. 
intros. unfold Rqd in H0. mc_unall H0. induction s' as [h' rf'].
simplRf H0; simplRf H; simplRf'.
inversion_clear H. left. splitland; getland.
rewrite <- H8; auto. rewrite <- H9; auto. rewrite <- H6; auto.
rewrite <- H12; auto. apply H7. unfold FR_OTH1; intros. clear H4.
splitex. exists x; exists x0. repeat (intros; rfinsen). 
intros; apply PairInsen with rf0; auto.
intros; apply SlistInsen with rf0; auto.
splitex. exists x; exists x0; exists x1.
rewrite <- H13; rewrite <- H10. repeat (intros; rfinsen).
unfold FR_OTH1; intros. splitex. exists x2; exists x3. repeat (intros; rfinsen).
intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
right. splitland. getland. rewrite <- H7; auto.
rewrite <- H8; auto. rewrite <- H11; auto.
apply H6. clear H3. unfold FR_OTH2. intros. splitex.
exists x; exists x0; exists x1; exists x2; exists x3. repeat (intros; rfinsen).
intros; apply PairInsen with rf0; auto.
intros; apply SlistInsen with rf0; auto.
intros; apply SlistInsen with rf0; auto.
intros; apply PairInsen with rf0; auto.
splitex. exists x; exists x0; exists x1. rewrite <- H12.
rewrite <- H9. rewrite <- H5. rewrite <- H11.
repeat (intros; rfinsen). unfold FR_OTH2. intros. splitex.
exists x2; exists x3; exists x4; exists x5; exists x6. repeat (intros; rfinsen).
intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
intros; apply PairInsen with rf; auto.
intros; auto.
unfold G_free_prolog. unfold Rqd. intros. mc_unall H. simplRf H.
intuition.
Save.


(**************************Well-formed Code Block**)

Theorem WF_copy_del :
  Infer (itpx2 (itpx1 T_copy_del) psi_copy_del) C_copy_del.
Proof.
unfold C_copy_del. unfold P_copy_del; unfold G_copy_del.
(** beq new zero L_copy_halt **)
apply WF_beq with P_copy_halt
  (fun s : state => (COPY_FRAME s /\ getValinR s tgt = pnil /\ getValinR s prnew =pnil /\
    getValinR s src <> pnil /\ 
    ((exists fv : addr, exists lnk : addr, exists siz : nat, 
      star (pto flist fv) 
      (star (MBlk (getValinR s retptr) lnk siz) 
	(star (Good fv)
	  (star (Stk TOP) (MC_OTH1(getValinR s src))))) s /\ siz >= 4)))
  \/ 
  (COPY_FRAME s /\ getValinR s prnew <> pnil /\ getValinR s src <> pnil /\ 
    ((exists fv : addr, exists lnk : addr, exists siz : nat, 
      star (pto flist fv) 
      (star (MBlk (getValinR s retptr) lnk siz) 
	(star (Good fv)
	  (star (Stk TOP) (MC_OTH2(getValinR s src)(getValinR s tgt)(getValinR s prnew))))) s 
      /\ siz >= 4))))
  NOG
  NOG.
constructor. 
(**P => P'**) Focus 2.
intros. unfold P_copy_halt; auto.
(**P => P' => G**) Focus 2.
intros; auto.
(**<> -> P => G'' => G*) Focus 2.
intros. simplz'. simplRf'; simplRf H. inversion_clear H.
splitland. inversion_clear H5. left. getland. splitex.
inversion_clear H5. contradiction. splitland. inversion_clear H4.
right. getland. splitex.
inversion_clear H4. contradiction.
(**<> -> P => G'' => G ***) Focus 2.
intros. auto.
unfold MC_OTH1; unfold MC_OTH2.
(** lw keyval src pnil **)
apply WF_lw with 
  (fun s : state =>
    ((getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1) /\
    getValinR s tgt = 0 /\ getValinR s prnew = 0 /\ getValinR s src <> 0 /\
    (exists fv : nat, (exists lnk : nat, (exists siz : nat,
      star (pto flist fv) (star (MBlk (getValinR s new) lnk siz)
	(star (Good fv) (star (Stk TOP)
          (fun s0 : state => exists lnk : addr, exists l1 : lint, 
            star (Pair (getValinR s src) (getValinR s keyval) lnk)
	    (Slist l1 lnk 0)s0 )))) s /\ siz >= 4))))
    \/
    (getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1) /\ getValinR s prnew <> 0 /\
    getValinR s src <> 0 /\ (exists fv : nat,(exists lnk : nat, (exists siz : nat,
      star (pto flist fv) (star (MBlk (getValinR s new) lnk siz)
        (star (Good fv) (star (Stk TOP)
          (fun s0 : state => ((exists lnk : addr, exists key : nat, 
	    exists l1 : lint, exists l2 : lint, star (
	      star (Pair (getValinR s src) (getValinR s keyval) lnk ) (Slist l2 lnk 0)
	    )(star (Slist l1 (getValinR s tgt)(getValinR s prnew))
                (Pair (getValinR s prnew) key
                  (getValinR s src))) s0)))))) s /\
          siz >= 4))))
  NOG.
(** src->key in dom **)
intros. simplRf H. inversion_clear H. simpl0p. splitland.
splitex. splitland. repeat splitstar. inrh. inrh. inrh. inrh.
splitex. slistindom0. simpl0p. splitland. splitex. splitland.
repeat splitstar. inrh. inrh. inrh. inrh. splitex. 
splitstar. inlh. slistindom0.
(**P S -> P' nextS **)
intros. induction s as [h rf]. inversion H0. simplRf H. simplRf'. simpl0p'.
inversion_clear H. left. splitland. getland. splitex. exists x0;
exists x1; exists x2. splitland; getland. rfinsen. clear H12. intros.
repeat (intros; rfinsen). intros. splitex. exists x3; exists x4.
repeat (intros; rfinsen). intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
splitstar; getstar'. splitstar; getstar'. splitstar; getstar'.
splitstar; getstar'. splitex.  inversion H25. contradiction.
exists lptr; exists l. replace x with (hread' h (lookupR rf src)).
unfold hread'. hread2. readrh. hread2. readrh. hread2. readrh.
hread2. readrh. hread2. readlh. foldhread'. getstar'. 
replace (hread' h1 (lookupR rf src)) with key; auto.
pairkey. fauto. pairindom0. inlh. pairindom0. inrh. inlh. pairindom0.
inrh. inrh. inlh. pairindom0. inrh. inrh. inrh. inlh. pairindom0.
unfold hread'. rewrite H7; auto. 
right. splitland. getland. splitex. exists x0; exists x1; exists x2. splitland. 
getland. rfinsen. clear H11. repeat (intros; rfinsen). intros. 
splitex. exists x3; exists x4; exists x5; exists x6. 
repeat (intros; rfinsen). intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto. intros; apply SlistInsen with rf; auto.
intros; apply PairInsen with rf; auto.
replace x with (hread' h (lookupR rf src)); [idtac|unfold hread'; rewrite H7; trivial].
unfold hread'. splitstar; getstar'. hread2. readrh.
splitstar; getstar'. hread2. readrh. splitstar; getstar'. hread2. readrh.
splitstar; getstar'. hread2. readrh. splitex. splitstar.
inversion H26. contradiction. exists lptr; exists x11; exists x13; exists l.
hread2. readlh. getstar'. hread2. readlh. foldhread'. 
 replace (hread' h1 (lookupR rf src)) with key. getstar'.
pairkey. fauto. pairindom0. fauto. inlh. pairindom0.
splitex. splitstar. inlh. slistindom0. splitstar; splitex. splitstar.
inrh. inlh. slistindom0. repeat splitstar. splitex. inrh.
inrh. splitstar. inlh. slistindom0. repeat splitstar. splitex. inrh.
inrh. splitstar. inrh. inlh. slistindom0.
intros; auto.
(** sw new 0 keyval **)
apply WF_sw with
  (fun s : state =>(getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1 /\
    getValinR s tgt = 0 /\ getValinR s prnew = 0 /\ getValinR s src <> 0 /\
    (exists fv : nat, exists lnk : nat, exists key : nat,
      star (pto flist fv) (star (Pair (getValinR s new) key lnk)
	(star (Good fv) (star (Stk TOP)
          (fun s0 : state => exists lnk : addr, exists l : lint, 
            star (Pair (getValinR s src) (getValinR s keyval) lnk)
	    (Slist l lnk 0)s0 )))) s ))
  \/
  getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1 /\ getValinR s prnew <> 0/\
  getValinR s src <> 0 /\ (exists fv : nat, exists lnk1 : nat, exists key1 : nat,
    star (pto flist fv) (star (Pair (getValinR s new) key1 lnk1)
      (star (Good fv) (star (Stk TOP)
        (fun s0 : state => (exists lnk : addr, exists key : nat, 
	  exists l1 : lint, exists l2 : lint, star (
	  star (Pair (getValinR s src) (getValinR s keyval) lnk ) (Slist l2 lnk 0)
	)(star (Slist l1 (getValinR s tgt)(getValinR s prnew))
          (Pair (getValinR s prnew) key (getValinR s src))) s0))))) s ))
  NOG.
(** new->key in dom **)
intros. simpl0p. simplRf H. inversion_clear H. splitland. 
splitex. splitland. repeat splitstar. inrh. inlh. fromb2p. splitex.
pairindom0. splitland. splitex. splitland.  repeat splitstar. inrh. inlh.
fromb2p. splitex. pairindom0.
(** P S -> P' nextS **)
intros. induction s as [h rf]. inversion H0.  simplRf H; simplRf'. 
simplRf H; simplRf'. inversion_clear H. left. splitland; getland.
simpl0p. splitex. splitland. exists x. do 2 splitstar. fromb2p.
splitex. exists x7; exists (lookupR rf keyval). 
move H after H19. writerh. inlh. pairindom0. writelh. pairindom0.
p2pw0. right. splitland. getland. splitex. exists x. splitland.
do 2 splitstar. fromb2p. splitex. exists x7; exists (lookupR rf keyval). 
move H after H18. writerh. simpl0p. inlh. pairindom0.
writelh. simpl0p. pairindom0. simpl0p. p2pw0. intros; auto.
(** addu a0 src zero **)
apply WF_addu with
  (fun s : state =>(getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1 /\
    getValinR s tgt = 0 /\ getValinR s prnew = 0 /\ getValinR s src <> 0 /\
    getValinR s a0 = getValinR s src /\
    (exists fv : nat, exists lnk : nat, exists key : nat,
      star (pto flist fv) (star (Pair (getValinR s new) key lnk)
	(star (Good fv) (star (Stk TOP)
          (fun s0 : state => exists lnk' : addr, exists l : lint, 
            star (Pair (getValinR s src) (getValinR s keyval) lnk')
	    (Slist l lnk' 0)s0 )))) s ))
  \/
  getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1 /\ getValinR s prnew <> 0 /\
  getValinR s src <> 0 /\ getValinR s a0 = getValinR s src /\
  (exists fv : nat, exists lnk1 : nat, exists key1 : nat,
    star (pto flist fv) (star (Pair (getValinR s new) key1 lnk1)
      (star (Good fv) (star (Stk TOP)
        (fun s0 : state => (exists lnk : addr, exists key : nat, 
	  exists l1 : lint, exists l2 : lint, star (
	  star (Pair (getValinR s src) (getValinR s keyval) lnk ) (Slist l2 lnk 0)
	)(star (Slist l1 (getValinR s tgt)(getValinR s prnew))
          (Pair (getValinR s prnew) key (getValinR s src))) s0))))) s ))
  NOG.
(** P S => P' nextS  *)
intros. induction s as [h rf]. inversion H0. simplz; [idtac| fauto].
simpl0p. inversion_clear H. left. splitland; getland. splitex.
exists x; exists x0; exists x1. simplRf H11; simplRf'.
repeat (intros; rfinsen). intros; apply PairInsen with rf; auto.
intros. splitex. exists x2; exists x3. repeat (intros; rfinsen). 
intros; apply PairInsen with rf; auto. 
intros; apply SlistInsen with rf; auto.
right. simplRf H7; simplRf'. splitland; getland. splitex.
exists x; exists x0; exists x1. repeat (intros; rfinsen).
intros; apply PairInsen with rf; auto. intros; splitex.
exists x2; exists x3; exists x4; exists x5. repeat (intros; rfinsen).
intros; apply PairInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
intros; apply SlistInsen with rf; auto.
intros; apply PairInsen with rf; auto.
intros; auto.
(** lw src src 1 ***)
apply WF_lw with 
  (fun s : state =>(getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1 /\
    getValinR s tgt = 0 /\ getValinR s prnew = 0 /\
    (exists fv : nat, exists lnk : nat, exists key : nat, exists lnk' : addr, 
      exists siz : nat, exists l : lint, 
      star (pto flist fv) (star (Pair (getValinR s new) key lnk)
	(star (Good fv) (star (Stk TOP)
          (star (MBlk (getValinR s a0) lnk' siz )
	    (Slist l (getValinR s src) 0))))) s))
  \/
  getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1 /\ getValinR s prnew <> 0
  /\ (exists fv : nat, exists lnk1 : nat, exists key1 : nat,
    exists lnk' : addr, exists siz : nat, exists key, exists l1 : lint, exists l2 : lint,
      star (pto flist fv) (star (Pair (getValinR s new) key1 lnk1)
	(star (Good fv) (star (Stk TOP) (star(
	  star (MBlk (getValinR s a0) lnk' siz) (Slist l1 (getValinR s src) 0)
	)(star (Slist l2 (getValinR s tgt)(getValinR s prnew))
          (Pair (getValinR s prnew) key (getValinR s a0))))))) s))
  NOG.
(** src->next in dom ***)
intros. simplRf H. inversion_clear H. splitland. splitex. repeat splitstar.
inrh. inrh. inrh. inrh. splitex. splitstar. inlh. pairindom1.
splitland. splitex. repeat splitstar. inrh. inrh. inrh. inrh. splitex.
repeat splitstar. inlh. inlh. pairindom1.
(** P S -> P' nextS **)
intros. induction s as [h rf]. inversion H0.
simplRf H; simplRf'. inversion_clear H. left.
splitland; getland. splitex. repeat splitstar.
splitex. repeat splitstar. rewrite H12.
exists x0; exists x1; exists x2. 
replace x with (hread' h (lookupR rf src+1));[idtac|unfold hread'; rewrite H7; auto].
assert (x11 = hread' h (lookupR rf src+1)). unfold hread'.
hread2. readrh. hread2. readrh. hread2. readrh. hread2. readrh. hread2. readlh.
foldhread'. pairlnk. fauto. pairindom1. inlh. pairindom1. inrh. inlh. pairindom1.
inrh. inrh. inlh. pairindom1. inrh. inrh. inrh. inlh. pairindom1.
rewrite <- H28. fromp2b. splitex.
exists x15; exists x16; exists x12. getstar'. getstar'.
apply PairInsen with rf; auto. getstar'. rfinsen.
getstar'. rfinsen. getstar'. rfinsen.
apply SlistInsen with rf; auto.
right. splitland. getland. splitex. exists x0.
exists x1. exists x2. repeat splitstar. splitex. 
rewrite H11. repeat splitstar. 
replace x with (hread' h (lookupR rf src+1));[idtac|unfold hread'; rewrite H7; auto].
assert (x11 = hread' h (lookupR rf src+1)). unfold hread'.
hread2. readrh. hread2. readrh. hread2. readrh. hread2. readrh.
hread2. readlh. hread2. readlh. foldhread'. pairlnk. fauto.
pairindom1. fauto. inlh. pairindom1.  inlh. inlh. pairindom1. 
inrh. inlh. inlh. pairindom1. inrh. inrh. inlh. inlh. pairindom1.
inrh. inrh. inrh. inlh. inlh. pairindom1. fromp2b. splitex.
exists x21; exists x22. exists x12. exists x14. exists x13.
rewrite <- H33. getstar'.
getstar'. apply PairInsen with rf; auto. getstar'. rfinsen.
getstar'. rfinsen. getstar'. getstar'. rfinsen. 
apply SlistInsen with rf; auto. getstar'. 
apply SlistInsen with rf; auto. apply PairInsen with rf; auto. 
intros; auto.
(*sw new 1 src **)
apply WF_sw with
  (fun s : state =>(getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1 /\
    getValinR s tgt = 0 /\ getValinR s prnew = 0 /\
    (exists fv : nat, exists key : nat, exists lnk' : addr, exists siz : nat,
      exists l : lint, 
      star (pto flist fv) (star (Pair (getValinR s new) key (getValinR s src))
	(star (Good fv) (star (Stk TOP)
          (star (MBlk (getValinR s a0) lnk' siz )
	    (Slist l (getValinR s src) 0))))) s))
  \/
  getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1 /\ getValinR s prnew <> 0
  /\ (exists fv : nat, exists key1 : nat, exists lnk' : addr, 
    exists siz : nat, exists key, exists l1 : lint, exists l2 : lint,
      star (pto flist fv) (star (Pair (getValinR s new) key1 (getValinR s src))
	(star (Good fv) (star (Stk TOP) (star(
	  star (MBlk (getValinR s a0) lnk' siz) (Slist l2 (getValinR s src) 0)
	)(star (Slist l1 (getValinR s tgt)(getValinR s prnew))
          (Pair (getValinR s prnew) key (getValinR s a0))))))) s))
  NOG.
(** new-> next in dom **)
intros. simplRf H. inversion_clear H. splitland. splitex. repeat splitstar.
inrh. inlh. pairindom1. splitland. splitex. repeat splitstar.
inrh. inlh. move H7 after H24. pairindom1.
(** P S -> P' nextS ***)
intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'.
inversion_clear H. left. splitland; getland. splitex.
exists x; exists x1; exists x2; exists x3; exists x4. splitstar. writerh.
splitstar. inlh. pairindom1. splitstar. writelh. pairindom1.
p2pw1. right. splitland. getland. splitex. exists x.
exists x1; exists x2; exists x3; exists x4; exists x6; exists x5. splitstar.
writerh. splitstar. inlh. pairindom1. splitstar.  writelh.
pairindom1. p2pw1. intros; auto.
(** j L_copy_reorder_t ***)
apply WF_j with P_copy_reorder_t G_copy_reorder_t.
constructor;[constructor|omega].
(**P => P'**)
unfold P_copy_reorder_t. intros. induction s as [h rf]. simplRf H. simplRf'.
inversion_clear H. left. splitland; getland. splitex. 
exists x; exists x1; exists x2. starrev'. assoclgoal. splitstar. getstar'.
starrev H7. starassocl H6. assoclgoal. splitstar. getstar'. 
starassocl H10. assoclgoal. splitstar. getstar'. starassocl H13.
starrev'. splitstar; getstar'. unfold FR_OTH1. exists x0; exists x3. starrev'. trivial.
right. splitland. getland. splitex. exists x; exists x1; exists x2.
starrev'. assoclgoal. splitstar. getstar'. starrev H6. starassocl H5. 
assoclgoal. splitstar; getstar'. starassocl H9. assoclgoal. 
splitstar; getstar'. starassocl H12. starrev'. starassocl H11.
splitstar; getstar'. unfold FR_OTH2. exists x0; exists x3; exists (lookupR rf reqsize).
exists x5; exists x4.
starassocr H15. starrev H14. splitstar. getstar'.
intros; auto.
Save.


(**************************************************)

Theorem WF_copy_new :
  Infer (itpx2 (itpx1 T_copy_new) psi_copy_new) C_copy_new.
Proof.
unfold C_copy_new.
(** addiu reqsize zero 2 ***)
apply WF_addiu with 
  (fun s : state => (COPY_FRAME s /\ getValinR s tgt = pnil /\ getValinR s prnew =pnil /\
    getValinR s src <> pnil /\ getValinR s reqsize = 2 /\
    exists l : lint, exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) 
      (Slist l (getValinR s src) pnil))) s)
  \/ 
  (COPY_FRAME s /\ getValinR s prnew <> pnil /\ getValinR s src <> pnil /\ 
    getValinR s reqsize = 2 /\ exists fv : addr, 
      exists la : lint, exists lb : lint, exists b : int, 
	(star (pto flist fv) (star (Good fv) (star (Stk TOP) 
	  (star (Slist la (getValinR s src) pnil) (star
	    (Slist lb (getValinR s tgt) (getValinR s prnew)) 
	    (Pair (getValinR s prnew) b (getValinR s src))))
	)) s )))
  NOG.
unfold P_copy_new. intros. induction s as [h rf]. inversion H0.
simplRf H; simplRf'. inversion_clear H. left. splitland; getland. 
simplz. fauto. splitex. exists x0; exists x. repeat (intros; rfinsen).
intros; apply SlistInsen with rf; auto. right. splitland; getland.
simplz. fauto. splitex. exists x; exists x1; exists x2; exists x0.
splitland. repeat (intros; rfinsen).
intros; apply SlistInsen with rf; auto. intros; apply SlistInsen with rf; auto.
intros; apply PairInsen with rf; auto. unfold G_copy_new. auto.
(** jal L_malloc_prolog L_copy_del **)
apply WF_jal with P_malloc_prolog P_copy_del G_malloc_prolog G_copy_del.
constructor. constructor. constructor. omega.
(**P -> P' **)
unfold P_malloc_prolog. intros. simplRf H. simplRf'.
assert (TOP > 10); [apply MYTOP|idtac]. inversion_clear H.
splitland; getland. omega. omega. omega. omega. splitex.
exists x0. rfinsen. splitstar. intros. repeat (intros; rfinsen).
splitstar. getstar'. splitstar. getstar'. splitstar. getstar'.
unfold arbh; auto. splitland; getland. omega. omega. omega. omega.
splitex. splitland. exists x. rfinsen. repeat (intros; rfinsen).
splitstar. getstar'. splitstar. getstar'. splitstar. getstar'.
unfold arbh. auto.
(** P => G' => P'' **)
unfold G_malloc_prolog; unfold P_copy_del. unfold Rqd.
intros. mc_unall H0. induction s' as [h' rf'].
simplRf H; simplRf H0; simplRf'. inversion_clear H. left.
splitland; getland. rewrite <- H10. rewrite H5; auto.
rewrite <- H11. rewrite H7; auto. rewrite <- H8. rewrite H1; auto.
rewrite <- H22. rewrite H2; auto. rewrite <- H20. trivial.
unfold MCPOST in H9. simplRf H9. rewrite H4 in H19. rewrite <- H19 in H9.
replace (2+2) with 4 in H9;[idtac|auto]. apply H9. unfold RfInsA.
unfold MC_OTH1. intros. clear H6. splitex. exists x.
apply SlistInsen with rf0; trivial. splitex. exists x0. unfold MC_OTH1.
rewrite <- H20. rfinsen. intros. repeat (intros; rfinsen). intros. 
splitex. exists x1. apply SlistInsen with rf; auto. splitstar. getstar'.
splitstar. getstar'. splitstar. getstar'. exists x. trivial.
right. splitland.  getland. rewrite <- H9; trivial. rewrite <- H10; trivial.
rewrite <- H21; trivial. rewrite <- H19; trivial. unfold MCPOST in H8.
simplRf H8. rewrite <- H18 in H8. rewrite H3 in H8. 
replace (2+2) with 4 in H8;[idtac|auto]. apply H8.
unfold RfInsA. unfold MC_OTH2. clear H5. intros. 
splitex. exists x; exists x0; exists x1. 
rfinsen. intros; apply SlistInsen with rf0; auto. intros; rfinsen.
intros; apply SlistInsen with rf0; auto. intros; apply PairInsen with rf0; auto.
rewrite <- H19. rewrite <- H7. rewrite <- H21. unfold MC_OTH2.
splitex. exists x. rfinsen. clear H5. intros.
repeat (intros; rfinsen). intros; splitex. exists x3; exists x4; exists x5.
rfinsen. intros; apply SlistInsen with rf; auto.
intros; rfinsen. intros; apply SlistInsen with rf; auto. 
intros; apply PairInsen with rf; auto. getland. splitstar. getstar'.
splitstar; getstar'. splitstar; getstar'. exists x2; exists x0; exists x1.
auto. intros; auto. unfold G_malloc_prolog.
unfold Rqd. intros. mc_unall H. simplRf H. intuition.
Save.


(**************************************************)
Theorem WF_copy_test :
  Infer (itpx2 (itpx1 T_copy_test) psi_copy_test) C_copy_test.
Proof.
unfold C_copy_test. 
(**bgtz src L_copy_new ***)
apply WF_bgtz with P_copy_new P_copy_halt G_copy_new G_copy_halt.
constructor. Focus 2. unfold P_copy_test; unfold P_copy_new.
intros. simplRf'; simplRf H. inversion_clear H. left. intuition.
right; intuition. Focus 2. unfold G_copy_test. intros; auto.
Focus 2. unfold P_copy_test; unfold P_copy_halt. intros. auto.
Focus 2. unfold G_copy_test; auto.
(**j L_copy_halt **)
apply WF_j with P_copy_halt G_copy_halt.
constructor. constructor. omega. intros; trivial.
intros; trivial.
Save.

(**************************************************)

Theorem WF_copy_entry :
  Infer (itpx2 (itpx1 T_copy_entry) psi_copy_entry) C_copy_entry.
Proof.
unfold C_copy_entry.
(** addiu tgt zero 0 **)
apply WF_addiu with
  (fun s : state => COPY_FRAME s /\ getValinR s tgt = pnil /\
    exists l : lint, exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) 
      (Slist l (getValinR s src) pnil))) s)
  NOG.
unfold P_copy_entry. intros. induction s as [h rf].
inversion H0. simplRf H; simplRf'. simplz. simplz'. simpl0p. simpl0p'.
intuition. splitex. exists x0; exists x. repeat (intros; rfinsen).
intros. apply SlistInsen with rf; auto. fauto. unfold G_copy_entry.
intros; auto.
(** addiu prnew zero 0 **)
apply WF_addiu with
  (fun s : state => COPY_FRAME s /\ getValinR s tgt = pnil /\ getValinR s prnew = pnil /\
    exists l : lint, exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) 
      (Slist l (getValinR s src) pnil))) s)
  NOG.
intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'. 
simplz. simplz'. simpl0p. simpl0p'. intuition. splitex. exists x; exists x0.
repeat (intros; rfinsen). intros. apply SlistInsen with rf; auto. fauto.
intros. auto.
(**j L_copy_test **)
apply WF_j with P_copy_test G_copy_test.
constructor. unfold P_copy_test. 
intros. left. intuition. splitex. exists x0; exists x. trivial.
intros; auto.
Save.
