Require Import heap2.
Require Import scap.
Require Import scap_rules.
Require Import cdsa.
Require Import Omega.
Require Import properties.
Require Import properties2.
Require Import List.
Require Import malloc.

(***Single Linked List Assertions ********)
(**** 2005-7-6 : From original cdsa ******)
(**nodes***)
Notation lint := (list int).
Definition Pair (hptr : addr) (key : nat) (lp : addr) : assert :=
  fun s : state => exists lnk : addr, exists siz, 
    getVinH s hptr = key /\ getVinH s (hptr + 1) = lp /\
    MBlk hptr lnk siz s /\ (siz >= 4).

Inductive Slist : lint -> addr -> addr -> assert :=
  | l_nil : forall (s : state) (i : addr),
    (emp s) -> Slist nil i i s
  | l_cons : forall (h1 h2 h : heap) (rf : rfile) (hptr lptr tptr : addr)
    (l : lint) (key : nat),
    hdisj h1 h2 ->
    heap_eq (happend h1 h2) h ->
    Pair hptr key lptr (h1, rf) ->
    Slist l lptr tptr (h2, rf) ->
    Slist (cons key l) hptr tptr (h, rf).


(****************** code heap of copy ****)
Notation L_copy_entry := 30.
Notation L_copy_test := 31.
Notation L_copy_new := 32.
Notation L_copy_del := 33.
Notation L_copy_ahead := 34.
Notation L_copy_lnkprv := 35.
Notation L_copy_halt := 36.
Notation L_copy_reorder_t := 37.

(**************Registers alias ************)
Notation src := s0.
Notation tgt := s1.
Notation new := v0.
Notation prnew := s2.
Notation nodesize := 2.
Notation nextnode := t0.
Notation keyval := t1.

(******************************************)
(********** Code Blocks *******************)
(******************************************)
Definition C_copy_entry : iseq := seq (
  addiu tgt zero pnil)            (seq (
  addiu prnew zero pnil)          (
  j L_copy_test)).

Definition C_copy_test : iseq := seq (
  bgtz src L_copy_new)           (
  j L_copy_halt).

Definition C_copy_new : iseq := seq (
  addiu a0 zero nodesize)       (
  jal L_malloc_prolog L_copy_del).

Definition C_copy_del := seq (
  beq new zero L_copy_halt)   (seq (
  lw keyval src 0)       (seq (
  sw new 0 keyval)       (seq (
  addu a0 src zero)     (seq (
  lw src src 1)          (seq (
  sw new 1 src)          (
  j L_copy_reorder_t     )))))).

Definition C_copy_reorder_t := seq (
  addiu s3 new pnil)               (
  jal L_free_prolog L_copy_ahead).

Definition C_copy_ahead := seq (
  bgtz prnew L_copy_lnkprv) (seq (
  addiu tgt s3 pnil)        (seq (
  addiu prnew tgt pnil)      (
  j L_copy_test))).
  
Definition C_copy_lnkprv := seq (
  sw prnew 1 s3)           (seq (
  addiu prnew s3 0)        (
  j L_copy_test)).

Definition C_copy_halt :=  (j L_copy_halt).


(***********************Spec of Copy********************)
Notation COPY_FRAME :=
  (fun s : state => getValinR s sp = TOP - 2 /\ getValinR s fp = TOP - 1).

Notation NOG := (fun s s' : state => getValinR s' ra = 0).

Definition P_copy_entry : assert :=
  fun s : state => COPY_FRAME s /\ exists fv : addr, exists l : lint,
    star (pto flist fv) (star (Good fv) (star (Stk TOP) 
      (Slist l (getValinR s src) pnil))) s.
Definition G_copy_entry : action := NOG.
Notation T_copy_entry := (cdspec P_copy_entry G_copy_entry).

Definition P_copy_test : assert :=
  fun s : state => (COPY_FRAME s /\ getValinR s tgt = pnil /\ getValinR s prnew =pnil /\
    exists fv : addr, exists l : lint,
      star (pto flist fv) (star (Good fv) (star (Stk TOP) 
      (Slist l (getValinR s src) pnil))) s)
  \/ 
  (COPY_FRAME s /\ getValinR s prnew <> pnil /\ 
    exists fv : addr, exists b : int, exists l1 : lint, exists l2 : lint, 
      (star (pto flist fv) (star (Good fv) (star (Stk TOP) 
	(star (Slist l1 (getValinR s src) pnil) (star
	  (Slist l2 (getValinR s tgt) (getValinR s prnew)) 
	  (Pair (getValinR s prnew) b (getValinR s src))))
      )) s)).
Definition G_copy_test : action := NOG.
Notation T_copy_test := (cdspec P_copy_test G_copy_test).

Definition P_copy_new : assert :=
  fun s : state => (COPY_FRAME s /\ getValinR s tgt = pnil /\ getValinR s prnew =pnil /\
    getValinR s src <> pnil /\ exists fv : addr, exists l : lint,
      star (pto flist fv) (star (Good fv) (star (Stk TOP) 
      (Slist l (getValinR s src) pnil))) s)
  \/ 
  (COPY_FRAME s /\ getValinR s prnew <> pnil /\ getValinR s src <> pnil /\ 
    exists fv : addr, exists b : int, exists l1 : lint, exists l2 : lint, 
	(star (pto flist fv) (star (Good fv) (star (Stk TOP) 
	  (star (Slist l1 (getValinR s src) pnil) (star
	    (Slist l2 (getValinR s tgt) (getValinR s prnew)) 
	    (Pair (getValinR s prnew) b (getValinR s src))))
	)) s )).

Definition G_copy_new : action := NOG.
Notation T_copy_new := (cdspec P_copy_new G_copy_new).

  
Definition MC_OTH1 (ptr : nat) : assert :=
  fun s : state => exists l : lint, Slist l ptr pnil s.

Definition MC_OTH2 (ptr1 :nat) (ptr2 : nat) (ptr3 : nat) : assert :=
  fun s : state => exists b : int, exists l1 : lint, exists l2 : lint, 
      star (Slist l1 ptr1 pnil) 
      (star (Slist l2 ptr2 ptr3) 
	(Pair ptr3 b ptr1)) s.


Definition P_copy_del : assert :=
  fun s : state => (COPY_FRAME s /\ getValinR s tgt = pnil /\ getValinR s prnew =pnil /\
    getValinR s src <> pnil /\ 
    ((exists fv : addr, exists lnk : addr, exists siz : nat, 
      star (pto flist fv) 
      (star (MBlk (getValinR s retptr) lnk siz) 
	(star (Good fv)
	  (star (Stk TOP) (MC_OTH1(getValinR s src))))) s /\ siz >= 4) \/
    (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) 
      (MC_OTH1 (getValinR s src)))) s /\ 
    getValinR s retptr = pnil)))
  \/ 
  (COPY_FRAME s /\ getValinR s prnew <> pnil /\ getValinR s src <> pnil /\ 
    ((exists fv : addr, exists lnk : addr, exists siz : nat, 
      star (pto flist fv) 
      (star (MBlk (getValinR s retptr) lnk siz) 
	(star (Good fv)
	  (star (Stk TOP) (MC_OTH2(getValinR s src)(getValinR s tgt)(getValinR s prnew))))) s 
      /\ siz >= 4)
    \/
    (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) 
      (MC_OTH2(getValinR s src)(getValinR s tgt)(getValinR s prnew)))) s 
    /\ getValinR s retptr = pnil))).

Definition G_copy_del : action := NOG.
Notation T_copy_del := (cdspec P_copy_del G_copy_del).


Definition P_copy_halt := (fun s : state => True).
Definition G_copy_halt := NOG.
Notation T_copy_halt := (cdspec P_copy_halt G_copy_halt).

Definition FR_OTH1 (new1 : addr) (src1 : addr)  : assert :=
  fun s : state => exists key, exists l : lint, 
    star (Pair new1 key src1) (Slist l src1 pnil) s.

Definition FR_OTH2 (new1 : addr) (src1 : addr) (tgt1 : addr) (prnew1 : addr) : assert :=
  fun s : state => exists key : nat, exists key': nat, exists nptr : addr,
    exists l1 : lint, exists l2 : lint, 
    star (Pair new1 key src1) (star (Slist l1 src1 0) (
      star (Slist l2 tgt1 prnew1) (Pair prnew1 key' nptr))) s.

Definition P_copy_ahead := 
  fun s : state => COPY_FRAME s /\ getValinR s tgt = pnil /\ getValinR s prnew =pnil /\
    (exists fv : addr, 
      star (pto flist fv) (
	star (Good fv) (star (Stk TOP) 
	  (FR_OTH1 (getValinR s s3)(getValinR s src)))) s)    
    \/
    COPY_FRAME s /\ getValinR s prnew <> pnil /\ (exists fv : addr,
      star (pto flist fv) (
	star (Good fv) (star (Stk TOP) 
	  (FR_OTH2 (getValinR s s3) (getValinR s src) 
	    (getValinR s tgt) (getValinR s prnew)))) s).

Definition G_copy_ahead := NOG.

Notation T_copy_ahead := (cdspec P_copy_ahead G_copy_ahead).

    
Definition P_copy_reorder_t := 
  fun s : state => COPY_FRAME s /\ getValinR s tgt = pnil /\ getValinR s prnew =pnil /\
    (exists fv : addr, exists lnk : addr, exists siz : nat,
      star (MBlk (getValinR s a0) lnk siz) (star (pto flist fv) (
	star (Good fv) (star (Stk TOP) (FR_OTH1 (getValinR s new)(getValinR s src))))) s)
    \/
    COPY_FRAME s /\ getValinR s prnew <> pnil /\ (exists fv : addr, exists lnk : addr,
      exists siz : nat, star (MBlk (getValinR s a0) lnk siz) (star (pto flist fv) (
	star (Good fv) (star (Stk TOP) 
	  (FR_OTH2 (getValinR s new) (getValinR s src) 
	    (getValinR s tgt) (getValinR s prnew))))) s).


Definition G_copy_reorder_t := NOG.

Notation T_copy_reorder_t := (cdspec P_copy_reorder_t G_copy_reorder_t).

Definition P_copy_lnkprv :=
  fun s : state => COPY_FRAME s /\ getValinR s prnew <> pnil /\
    exists fv : addr, exists key : nat, exists key' : nat, exists nptr : addr, 
      exists l1 : lint, exists l2 : lint, 
      star (pto flist fv) (
	star (Good fv) (
	  star (Stk TOP) (
	    star (Pair (getValinR s s3) key (getValinR s src)) (
	      star (Slist l1 (getValinR s src) 0) (
	      star (Slist l2 (getValinR s tgt) (getValinR s prnew)) 
	      (Pair (getValinR s prnew) key' nptr)))))) s.

Definition G_copy_lnkprv := NOG.

Notation T_copy_lnkprv := (cdspec P_copy_lnkprv G_copy_lnkprv).



