Require Import heap2.
Require Import scap.
Require Import scap_rules.
Require Import cdsa.
Require Import Omega.
Require Import properties.
Require Import properties2.
Require Import copydefs.
Require Import properties.
Require Import List.

Lemma h_subst_Slist : forall (h1 h2 : heap)(rf : rfile)
  (hptr tptr : addr)(l : lint),
  heap_eq h1 h2 -> Slist l hptr tptr (h1, rf) 
  -> Slist l hptr tptr (h2, rf).
Proof.
intros.
induction l.
inversion_clear H0; apply l_nil.
unfold emp in H1; unfold emp.
unfold getValinH; unfold getValinH in H1.
intros.
assert (hread h1 i = hread h2 i).
apply h_subst_read; auto.
rewrite <- H0; auto.
inversion_clear H0.
apply l_cons with h0 h3 lptr; auto.
apply heap_eq_trans with h1; auto.
Save.

Lemma SlistCons' :
  forall (l : lint) (h1 h2 h : heap) (rf : rfile) (hptr lptr tptr : addr)
    (key : nat),
    hdisj h1 h2 ->
    heap_eq (happend h1 h2) h ->
    Slist l hptr tptr (h1, rf) ->
    Pair tptr key lptr (h2, rf) ->
    Slist (app l (cons key nil)) hptr lptr (h, rf).
Proof.
induction l; intros.
assert (Slist (cons key nil) tptr lptr (h2, rf)).
apply l_cons with h2 hempty lptr; auto.
apply disj_sym; apply disj_emp.
assert (heap_eq (happend h2 hempty)(happend hempty h2)).
apply append_sym; apply disj_sym; apply disj_emp.
assert (heap_eq (happend hempty h2) h2).
apply append_emp.
apply heap_eq_trans with (happend hempty h2); [auto | apply append_emp].
apply l_nil.
unfold emp; unfold getValinH; unfold hread; auto.
clear H2.
inversion_clear H1; simpl.
apply h_subst_Slist with h2; auto.
apply heap_eq_trans with (happend h1 h2); auto.
unfold emp in H2; unfold getValinH in H2; unfold hread in H2.
apply all_same; intros; unfold happend.
assert (h1 i = None).
auto.
rewrite H1; auto.
replace ((a :: l) ++ key :: nil) with (a :: (l ++ key :: nil)); auto.
inversion_clear H1.
assert (hdisj h2 h0 /\ hdisj h2 h3).
apply disj_append_R; apply disj_heap_eq with h1.
apply disj_sym; auto.
apply heap_eq_sym; auto.
inversion_clear H1.
apply l_cons with h0 (happend h3 h2) lptr0; auto.
apply disj_append_L; getland.
apply disj_sym; auto.
assert (heap_eq (happend h3 h2)(happend h2 h3)).
apply append_sym; apply disj_sym; auto.
assert (heap_eq (happend (happend h3 h2) h0)
  (happend (happend h2 h3) h0)).
apply heap_eq_append; auto.
assert (heap_eq  (happend h0 (happend h3 h2))
  (happend (happend h3 h2) h0)).
apply append_sym.
apply disj_append_L; split; [auto | apply disj_sym; auto].
apply heap_eq_trans with (happend (happend h3 h2) h0 ); auto.
apply heap_eq_trans with (happend (happend h2 h3) h0); auto.
assert (heap_eq (happend h0 (happend h2 h3))
  (happend (happend h2 h3) h0)).
apply append_sym.
apply disj_append_L; split; [apply disj_sym; auto | auto].
apply heap_eq_trans with (happend h0 (happend h2 h3)).
apply heap_eq_sym; auto.
apply cross_happend with h1; auto.
apply IHl with h3 h2 tptr; auto.
apply disj_sym; auto.
apply heap_eq_refl.
Save.

(*    Slist (rev(cons key (rev l))) hptr tptr (h, rf).*)

Lemma PairNot0 : forall (h : heap) (rf : rfile) (hptr nptr : addr) (v : nat),
  Pair hptr v nptr (h, rf) -> hptr <> pnil.
Proof.
intros.
inversion_clear H.
splitex; splitland.
inversion_clear H1; omega.
Save.

Ltac pairnot0 :=
  match goal with
    | [ H : Pair ?hptr ?v ?nptr (?h, ?rf) |- _] =>
      apply PairNot0 with h rf nptr v; auto
  end.

Lemma PairInsen :
  forall (h : heap) (rf rf' : rfile) (hptr nptr : addr) (v : nat),
    Pair hptr v nptr (h, rf) -> Pair hptr v nptr (h, rf').
Proof.
intros.
inversion_clear H;inversion_clear H0;intuition.
unfold Pair.
exists x;exists x0;getland.
apply MBlkInsen with rf;auto.
Save.

Lemma SlistInsen :
  forall (l : lint),
    forall (h : heap) (rf rf' : rfile) (hptr tptr : addr),
      Slist l hptr tptr (h, rf) -> Slist l hptr tptr (h, rf').
Proof.
induction l; intros.
 inversion_clear H.
   apply l_nil; intuition.
 inversion_clear H.
   apply l_cons with h1 h2 lptr.
  exact H0.
  exact H1.
  apply PairInsen with rf; exact H2.
  apply IHl with rf; auto.
Save.

Lemma PairInDom0 : forall (h : heap) (rf : rfile) (hptr nptr : addr) (v : nat),
  Pair hptr v nptr (h, rf) -> hindom hptr h.
Proof.
intros.
inversion_clear H; inversion_clear H0; intuition.
inversion_clear H1; inversion_clear H4.
splitstar1.
unfold removeS in H7.
assert (hindom hptr (hfree (hfree h (hptr - 2)) (hptr - 1))).
 apply pnalPtrIn with (x0 - 2) rf hptr; auto.
   omega.
 inversion_clear H6; unfold hfree in H8.
   apply indom_some with v0; rewrite <- H8.
   assert (hptr - 1 <> hptr); [ omega | simplneq ].
  assert (hptr - 2 <> hptr); [ omega | simplneq ].
    apply beq_nat_neq; exact H9.
  apply beq_nat_neq; exact H6.
Save.


Lemma PairInDom1 : forall (h : heap) (rf rf' : rfile) (hptr nptr : addr) (v : nat),
  Pair hptr v nptr (h, rf) -> hindom (hptr+1) h.
Proof.
intros.
inversion_clear H; inversion_clear H0; intuition.
inversion_clear H1; inversion_clear H4.
splitstar1.
unfold removeS in H7.
assert (hindom (hptr + 1) (hfree (hfree h (hptr - 2)) (hptr - 1))).
 apply pnalPtrIn with (x0 - 2) rf hptr; auto.
  omega.
  omega.
 inversion_clear H6; unfold hfree in H8.
   apply indom_some with v0; rewrite <- H8.
   assert (hptr - 1 <> hptr + 1); [ omega | simplneq ].
  assert (hptr - 2 <> hptr + 1); [ omega | simplneq ].
    apply beq_nat_neq; exact H9.
  apply beq_nat_neq; exact H6.
Save.



Ltac pairindom0 :=
  match goal with
    | [H : Pair ?hptr ?v ?nptr (?h, ?rf) |- _] =>
      apply PairInDom0 with rf nptr v; auto
  end.

Ltac pairindom1 :=
  match goal with
    | [H : Pair ?hptr ?v ?nptr (?h, ?rf) |- _] =>
      apply PairInDom1 with rf nptr v; auto
  end.

Lemma SlistInDom0 : forall (h : heap) (hptr tptr: addr) (rf : rfile) (l : lint),
  hptr <> tptr ->
  Slist l hptr tptr (h, rf)->
  hindom hptr h.
Proof.
intros.
inversion H0;intuition.    
apply h_subst_indom with (happend h1 h2);auto.
apply append_preserve_indom_L.
pairindom0.
Save.


Ltac slistindom0 := 
  match goal with
    | [ H : Slist ?l ?hptr ?tptr (?h, ?rf) |- _] =>
      apply SlistInDom0 with tptr rf l; auto
  end.

Lemma PairKey : forall (h : heap) (rf : rfile) (hptr nptr : addr) (v : nat),
  Pair hptr v nptr (h, rf) -> v = hread' h hptr .
Proof.
intros.
inversion_clear H;inversion_clear H0;intuition.
Save.

Lemma PairLnk : forall (h : heap) (rf : rfile) (hptr nptr : addr) (v : nat),
  Pair hptr v nptr (h, rf) -> nptr = hread' h (hptr+1) .
Proof.
intros.
inversion_clear H.
splitex; splitland.
rewrite <- H.
unfold getVinH; auto.
Save.


Ltac pairkey :=
  match goal with 
    | [H : Pair ?hptr ?v ?nptr (?h, ?rf) |- _] =>
      apply PairKey with rf nptr; auto
  end.

Ltac pairlnk :=
  match goal with 
    | [H : Pair ?hptr ?v ?nptr (?h, ?rf) |- _] =>
      apply PairLnk with rf v; auto
  end.

Lemma MBlk_Pair :
  forall (hptr lnk : addr) (siz : nat) (h : heap) (rf : rfile),
    MBlk hptr lnk siz (h, rf) ->
    siz >= 4 ->
    (exists key : nat, exists nptr : addr, (Pair hptr key nptr (h, rf))).
Proof.
intros.
unfold Pair.
exists (getVinH (h,rf) hptr); exists (getVinH (h,rf) (hptr + 1)).
exists lnk; exists siz.
getland.
Save.

Ltac fromb2p :=
  match goal with
    | [ H : MBlk ?hptr ?lnk ?siz (?h, ?rf) |- _ ] =>
      assert (exists key : nat, exists nptr : addr,
	Pair hptr key nptr (h, rf)); [apply MBlk_Pair with lnk siz; auto| clear H]
  end.

Lemma Pair_MBlk :
  forall (hptr nptr : addr) (key : nat) (h : heap) (rf : rfile),
    Pair hptr key nptr (h, rf) ->
    (exists lnk : addr, exists siz : nat, (MBlk hptr lnk siz (h, rf))).
Proof.
intros.
inversion_clear H.
splitex; splitland.
exists x; exists x0; auto.
Save.


Ltac fromp2b :=
  match goal with
    | [ H : Pair ?hptr ?key ?nptr (?h, ?rf) |- _] =>
      assert (exists lnk : addr, exists siz : nat, 
	MBlk hptr lnk siz (h, rf)); [apply Pair_MBlk with nptr key; auto| clear H]
  end.

Lemma MBlk_hptr_write:
  forall (hptr new : addr)(v lnk siz : nat)(h : heap)(rf : rfile),
    MBlk hptr lnk siz (h, rf) -> siz >= 4 -> new = hptr \/ new = hptr +1
    ->   MBlk hptr lnk siz (hwrite h new v, rf).
Proof.
unfold MBlk.
intros.
splitland; getland.
splitstar1; getstar1.
unfold getVinH; unfold hwrite; unfold getValinH; unfold hread.
simplneq.
apply beq_neq_false; omega.
rewrite <- H4.
unfold removeS; unfold getValinH; unfold hread.
unfold hfree; unfold hwrite.
simplneq.
simplneq.
apply beq_neq_false; omega.
apply beq_neq_false; omega.
unfold removeS; unfold removeS in H4.
assert (heap_eq (hfree (hfree (hwrite h new v) (hptr - 2)) (hptr - 1))
       (hwrite (hfree (hfree h (hptr - 2)) (hptr - 1)) new v)).
apply all_same.
intros.
unfold hfree; unfold hwrite.
compare new i.
intros.
assert (true = EqNat.beq_nat new i).
apply beq_eq_true; auto.
rewrite <- H5.
simplneq.
simplneq.
apply beq_neq_false; omega.
apply beq_neq_false; omega.
compare (hptr - 1) i.
intros.
assert (true = EqNat.beq_nat (hptr - 1) i).
apply beq_eq_true; auto.
rewrite <- H5.
simplneq.
apply beq_neq_false; auto.
compare (hptr -2) i.
intros.
assert (true = EqNat.beq_nat (hptr - 2) i).
apply beq_eq_true; auto.
rewrite <- H5.
simplneq.
simplneq.
apply beq_neq_false; auto.
apply beq_neq_false; auto.
intros.
simplneq.
simplneq.
apply beq_neq_false; auto.
apply beq_neq_false; auto.
assert (ptoanynlist hptr (siz - 2)
      ((hwrite (hfree (hfree h (hptr - 2)) (hptr - 1)) new v),rf)).
apply pnalRet; auto.
apply pnalPtrIn with (siz - 2) rf hptr; auto.
omega.
omega.
apply h_subst_ptoanynlist with
    (hwrite (hfree (hfree h (hptr - 2)) (hptr - 1)) new v); auto.
apply heap_eq_sym; auto.
Save.

Lemma Pair2Pair0 :
  forall (hptr nptr : addr) (key key' : nat) (h : heap) (rf : rfile),
    Pair hptr key nptr (h, rf) ->
    Pair hptr key' nptr (hwrite h hptr key', rf).
Proof.
unfold Pair.
intros.
splitex; splitland.
exists x; exists x0; getland.
unfold getVinH; unfold hwrite; unfold getValinH; unfold hread.
simpleq.
apply beq_eq_true; auto.
unfold getVinH; unfold hwrite; unfold getValinH; unfold hread.
simplneq.
apply beq_neq_false; omega.
apply MBlk_hptr_write; auto.
Save.

Ltac p2pw0 := 
  match goal with
    | [ H : Pair ?hptr ?key ?nptr (?h, ?rf) |- _] =>
      apply Pair2Pair0 with  key; auto
  end.

Lemma Pair2Pair1 :
  forall (hptr nptr nptr' : addr) (key : nat) (h : heap) (rf : rfile),
    Pair hptr key nptr (h, rf) ->
    Pair hptr key nptr' (hwrite h (hptr+1) nptr', rf).
Proof.
unfold Pair.
intros.
splitex; splitland.
exists x; exists x0.
getland.
unfold getVinH; unfold hwrite; unfold getValinH; unfold hread.
simplneq.
apply beq_neq_false; omega.
unfold getVinH; unfold hwrite; unfold getValinH; unfold hread.
simpleq.
apply beq_eq_true; auto.
apply MBlk_hptr_write; auto.
Save.

Ltac p2pw1 := 
  match goal with
    | [ H : Pair ?hptr ?key ?nptr (?h, ?rf) |- _] =>
      apply Pair2Pair1 with  nptr; auto
  end.

