Require Import Arith.
Require Import heap2.
Require Import scap.
Require Import scap_rules.
Require Import List.
Require Import cdsa.
Require Import properties.
Require Import properties2.
Require Import lib_lemma.
Require Import Omega.

Notation L_prolog := 1.
Notation L_fact := 2.
Notation L_nonzero := 3.
Notation L_mul := 4.
Notation L_epilog := 5.

Notation C_prolog := (seq
  (subiu sp sp 3) (seq 
  (sw sp 3 fp) (seq
  (addiu fp sp 3) (seq
  (sw_ fp 1 ra) (seq
  (sw_ fp 2 s0) 
  (j L_fact)))))).

Notation C_fact := (seq
  (bgtz a0 L_nonzero) (seq
  (addiu v0 zero 1) 
  (j L_epilog))).

Notation C_nonzero := (seq
  (addiu s0 a0 0) (seq 
  (subiu a0 a0 1) 
  (jal L_prolog L_mul))).

Notation C_mul := (seq
  (addu v0 s0 v0)
  (j L_epilog)).

Notation C_epilog := (seq
  (lw_ s0 fp 2) (seq
  (lw_ ra fp 1) (seq
  (lw fp fp 0) (seq
  (addiu sp sp 3) 
  (jr ra))))).

Definition GoodFr(n : nat) :=
  fun s : state => getValinR s sp > 0 /\ getValinR s fp <= TOP /\
    getValinR s fp = getValinR s sp + n.

Notation ALLT := (
  Gand (RIntact t0) (Gand (RIntact t1) (Gand (RIntact t2) (Gand (RIntact t3) (
  Gand (RIntact t4) (Gand (RIntact t5) (Gand (RIntact t6) (Gand (RIntact t7) (
  Gand (RIntact t8) (RIntact t9)))))))))).

Definition P_prolog : assert := 
  fun s : state => (exists n : nat, GoodFr n s) /\ getValinR s sp > ((getValinR s a0 + 1) * 3 )
    /\ star (Stk TOP) arbh s /\ getValinR s ra <> 0.

Definition G_prolog : action :=
  fun s s' : state => Gand S0S1S2 (Gand ALLT Rqd) s s' /\
    HIntact TOP (getValinR s sp) s s' /\ forall oth : assert, RfInsA oth ->
      star (Stk TOP) oth s -> star (Stk TOP) oth s'.

Notation T_prolog := (cdspec P_prolog G_prolog).

Definition P_fact : assert :=
  fun s : state => GoodFr 3 s /\ getValinR s fp > ((getValinR s a0 + 1) * 3 )
    /\ star (Stk TOP) arbh s /\ getVinH s (getValinR s fp - 1) <> 0. 

Definition G_fact : action :=
  fun s s' : state => Gand ES0 ALLT s s' /\ 
    getValinR s' fp = getVinH s (getValinR s fp) /\
    getValinR s' sp = getValinR s sp + 3 /\
    getValinR s' ra = getVinH s (getValinR s fp - 1) /\
    getValinR s' s0 = getVinH s (getValinR s fp - 2) /\
    HIntact TOP (getValinR s fp) s s' /\ 
    forall oth : assert, RfInsA oth -> star (Stk TOP) oth s -> star (Stk TOP) oth s'.

Notation T_fact := (cdspec P_fact G_fact).

Definition P_nonzero : assert :=
  fun s : state => P_fact s /\ getValinR s a0 > 0.

Notation G_nonzero := G_fact.

Notation T_nonzero := (cdspec P_nonzero G_nonzero).

Definition P_epilog : assert :=
  fun s : state => GoodFr 3 s /\ star (Stk TOP) arbh s /\ getVinH s (getValinR s fp - 1) <> 0.
Notation G_epilog := G_fact.
Notation T_epilog := (cdspec P_epilog G_epilog).

Notation P_mul := P_epilog.
Notation G_mul := G_epilog.
Notation T_mul := (cdspec P_mul G_mul).
  
(**************************************************************)
(**********local spec. env.************************************)
(**************************************************************)
Notation psi_prolog := (consCT L_fact T_fact emptyCT).
Notation psi_fact := 
  (consCT L_nonzero T_nonzero 
  (consCT L_epilog T_epilog emptyCT)).

Notation psi_nonzero := 
  (consCT L_prolog T_prolog
  (consCT L_mul T_mul emptyCT)).

Notation psi_mul := (consCT L_epilog T_epilog emptyCT).

(**************************************************************)
(********************Well-formed code block *******************)
(**************************************************************)
Ltac unall H1 := try (unfold S0S1S2 in H1; unfold S0S1S2; unfold Rqd in H1;
   unfold Rqd; unfold ES0 in H1; unfold ES0; unGand H1; unRIntact H1).

Ltac simplps' :=
  match goal with
    | [ H1 : context [(?X -?N + ?N)] |- _] =>
      replace (X - N + N) with X in H1; auto
  end.
Ltac othhold H rf := try (unfold RfInsA in H; intros; apply H with rf; auto).
Ltac ungetVinH' H := try (unfold getVinH in H;[unfold getValinH in H;
  [unfold hread in H;[unfold hwrite in H]]]).

Axiom HIntact_trans :
  forall (b e e': nat) (h h' h'' : heap) (rf rf' rf'' : rfile),
    e' <= e ->
    HIntact b e' (h, rf) (h', rf')->
    HIntact b e (h', rf') (h'', rf'') ->
    HIntact b e (h, rf) (h'', rf'').

(***************************************************************)
Theorem WF_epilog :
  Infer (itpx2 (itpx1 T_epilog) emptyCT) C_epilog.
Proof.
apply WF_lw_ with
  P_epilog
  (fun s s' : state => Gand ES0 ALLT s s' /\ 
    getValinR s' fp = getVinH s (getValinR s fp) /\
    getValinR s' sp = getValinR s sp + 3 /\
    getValinR s' ra = getVinH s (getValinR s fp - 1) /\
    getValinR s' s0 = getValinR s s0 /\
    HIntact TOP (getValinR s fp) s s' /\ 
    forall oth : assert, RfInsA oth -> star (Stk TOP) oth s -> star (Stk TOP) oth s').
(* fp-2 in dom **)
unfold P_epilog. unfold GoodFr. intros. simplRf H. splitland.
splitstar. inlh. stkindom.
(*P S -> P'S' *)
unfold P_epilog. unfold GoodFr. intros. induction s as [h rf].
inversion H0. simplRf H; simplRf'. intuition. repeat (intros; rfinsen).
(*P S -> G' S' S'' -> G S S'' *)
unfold G_fact. intros. induction s as [h rf]; induction s'' as [h'' rf''].
inversion H0. rewrite <- H3 in H1. unall H1. simplRf H1; simplRf'.
intuition. ungetVinH. foldhread. rewrite H8; trivial. 
apply H24; auto. repeat (intros; rfinsen). othhold H30 rf.
(*lw_ ra fp 1 **)
apply WF_lw_ with
  (fun s : state => GoodFr 3 s /\ star (Stk TOP) arbh s /\ getValinR s ra <> 0)
  (fun s s' : state => Gand ES0 ALLT s s' /\ 
    getValinR s' fp = getVinH s (getValinR s fp) /\
    getValinR s' sp = getValinR s sp + 3 /\
    getValinR s' ra = getValinR s ra /\
    getValinR s' s0 = getValinR s s0 /\
    HIntact TOP (getValinR s fp) s s' /\ 
    forall oth : assert, RfInsA oth -> star (Stk TOP) oth s -> star (Stk TOP) oth s').
(* fp-1 in dom **)
unfold P_epilog. unfold GoodFr. intros. simplRf H. splitland.
splitstar. inlh. stkindom.
(*P S -> P'S' *)
unfold P_epilog. unfold GoodFr. intros. induction s as [h rf].
inversion H0. simplRf H; simplRf'. intuition. repeat (intros; rfinsen).
ungetVinH' H11. unfold hread in H7. rewrite H7 in H11. apply H11; auto.
(*P S -> G' S' S'' -> G S S'' *)
unfold G_fact. intros. induction s as [h rf]; induction s'' as [h'' rf''].
inversion H0. rewrite <- H3 in H1. unall H1. simplRf H1; simplRf'.
intuition. ungetVinH. foldhread. rewrite H8; trivial. 
apply H24; auto. repeat (intros; rfinsen). othhold H30 rf. 
(*lw fp fp 0 *)
apply WF_lw with
  (fun s : state => star (Stk TOP) arbh s /\ getValinR s ra <> 0)
  (fun s s' : state => Gand S0S1S2 ALLT s s' /\ 
    getValinR s' fp = getValinR s fp /\
    getValinR s' sp = getValinR s sp + 3 /\
    getValinR s' ra = getValinR s ra /\
    HIntact TOP (getValinR s sp + 3) s s' /\ 
    forall oth : assert, RfInsA oth -> star (Stk TOP) oth s -> star (Stk TOP) oth s').
(* fp in dom *)
unfold GoodFr. intros. simpl0p. simplRf H.  splitland. splitstar. inlh.
stkindom. 
(**P S -> P' S' *)
unfold GoodFr. intros. induction s as [h rf].
inversion H0. simplRf H; simplRf'. intuition. repeat (intros; rfinsen).
(*P S -> G' S' S'' -> G S S'' *)
unfold GoodFr. intros. induction s as [h rf]; induction s'' as [h'' rf''].
inversion H0. rewrite <- H3 in H1. unall H1. simplRf H1; simplRf'; simplRf H.
intuition. simpl0p'. ungetVinH. foldhread. rewrite H8; trivial. rewrite H16; auto.
apply H25; auto. repeat (intros; rfinsen). othhold H34 rf. 
(**addiu sp sp 3*)
apply WF_addiu with
  (fun s : state => getValinR s ra <> 0)
  (fun s s' : state => Gand S0S1S2 ALLT s s' /\ 
    getValinR s' fp = getValinR s fp /\
    getValinR s' sp = getValinR s sp /\
    getValinR s' ra = getValinR s ra /\
    HIntact TOP (getValinR s sp) s s' /\ 
    forall oth : assert, RfInsA oth -> star (Stk TOP) oth s -> star (Stk TOP) oth s').
(*P S -> P' S' *)
intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'. splitland; auto. 
(*P S -> G' S' S'' -> G S S''*)
intros. unall H1. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. simplRf H1; simplRf'. intuition.
apply H21; auto. repeat (intros; rfinsen). othhold H30 rf.
(*jr ra **)
apply WF_jr_ra'.
intros. simplRf H. getland. unall H. getland. unfold HIntact. 
intros; auto. 
Save.

(***************************************************************)

Theorem WF_mul :
  Infer (itpx2 (itpx1 T_mul) psi_mul) C_mul.
Proof.
(** addu v0 s0 v0 **)
apply WF_addu with P_epilog G_epilog.
(* P S -> P' S' *)
unfold P_epilog. unfold GoodFr. intros.
induction s as [h rf]. inversion H0. simplRf H; simplRf'.
intuition. repeat (intros; rfinsen).
(* P S -> G S' S'' -> G S S'' *)
unfold G_fact. intros. unall H1. induction s as [h rf].
induction s'' as [h'' rf'']. inversion H0. rewrite <- H2 in H1.
simplRf H1; simplRf'. intuition. apply H23; auto.
repeat (intros; rfinsen). othhold H29 rf.
(** j L_epilog *)
apply WF_j with P_epilog G_epilog.
constructor. 
(*P => P' *)
intros; trivial.
(*P => G' => G **)
intros; trivial.
Save.

(***************************************************************)
Theorem WF_nonzero :
  Infer (itpx2 (itpx1 T_nonzero) psi_nonzero) C_nonzero.
Proof.
(**addiu s0 a0 0*)
apply WF_addiu with P_nonzero G_nonzero.
(* P S -> P' S' *)
unfold P_nonzero. unfold P_fact. unfold GoodFr.
intros. induction s as [h rf]. inversion H0.
simplRf H; simplRf'. intuition. repeat (intros; rfinsen).
(** P S -> G' S' S'' -> G S S''*)
unfold G_fact. intros. induction s as [h rf]; induction s'' as [h'' rf''].
inversion H0.  rewrite <- H2 in H1. unall H1. intuition.
apply H23; auto. repeat (intros; rfinsen). othhold H29 rf.
(** subiu a0 a0 1 **)
apply WF_subiu with 
  (fun s : state => GoodFr 3 s /\ getValinR s sp > ((getValinR s a0 + 1) * 3 )
    /\ star (Stk TOP) arbh s /\ getVinH s (getValinR s fp - 1) <> 0)
  G_nonzero.
(*P S -> P' S'*)
unfold P_nonzero. unfold P_fact. unfold GoodFr.
intros. induction s as [h rf]. inversion H0.
simplRf H; simplRf'. intuition. repeat (intros; rfinsen).
(** P S -> G' S' S'' -> G S S''*)
unfold G_fact. intros. induction s as [h rf]; induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. unall H1. intuition.
apply H23; auto. repeat (intros; rfinsen). othhold H29 rf.
(** jal L_prolog L_mul **)
apply WF_jal with P_prolog P_mul G_prolog G_mul.
constructor. constructor; [constructor|omega].
(** P -> P'{ra~>_} **)
unfold P_prolog. unfold GoodFr.  intros. simplRf H; simplRf'.
intuition. exists 3. intuition. repeat (intros; rfinsen).
(** P => G' => P''*)
unfold P_epilog. unfold G_prolog. unfold GoodFr.
intros. unall H0. induction s' as [h' rf'].
simplRf H; simplRf H0; simplRf'. intuition.
apply H7; auto. unfold RfInsA; unfold arbh. intros; auto.
repeat (intros; rfinsen). rewrite <- H13 in H27.
ungetVinH' H27; ungetVinH' H9. apply H9. 
unfold HIntact in H. ungetVinH' H. 
assert (h (lookupR rf fp - 1) = h' (lookupR rf fp - 1)).
apply H; omega. rewrite H29; auto. 
(** P => G' => G'' => G **)
unfold G_prolog. unfold G_fact. unfold GoodFr.
intros. unall H0; unall H1. induction s' as [h' rf'].
induction s'' as [h'' rf''].
simplRf H; simplRf H0; simplRf H1; simplRf'.
splitland. unfold HIntact in H0. ungetVinH' H0.
intuition. rewrite H1. rewrite <- H28. 
ungetVinH. assert (h (lookupR rf fp) = h' (lookupR rf fp)).
apply H0. intuition. rewrite H50; trivial.
rewrite H4. rewrite <- H28. ungetVinH. 
assert (h (lookupR rf fp-1) = h' (lookupR rf fp-1)).
apply H0. intuition. rewrite H50; trivial.
rewrite H5. rewrite <- H28. ungetVinH. 
assert (h (lookupR rf fp-2) = h' (lookupR rf fp-2)).
apply H0. intuition. rewrite H50; trivial.
rewrite <- H28 in H6. 
assert (HIntact TOP (lookupR rf sp) (h, rf) (h', rf')).
unfold HIntact. ungetVinH.  trivial.
apply HIntact_trans with (lookupR rf sp) h' rf'; auto.
omega. apply H8; auto. apply H26; auto.
repeat (intros; rfinsen). othhold H50 rf.
unfold G_prolog. intros. unall H. simplRf H. intuition.
Save.

(***************************************************************)
Theorem WF_fact :
  Infer (itpx2 (itpx1 T_fact) psi_fact) C_fact.
Proof.
apply WF_bgtz with P_nonzero P_epilog G_nonzero G_epilog.
constructor.
(* >0 -> P => P' *) Focus 2.
unfold P_nonzero. unfold P_fact. intros; intuition.
(* >0 -> P => G' => G **) Focus 2.
intros; auto.
(* <=0 -> P => P'' *) Focus 2.
unfold P_fact. unfold P_epilog. intros; intuition.
(* <=0 -> P => G'' => G **) Focus 2.
intros; auto.
(*addiu v0 zero 1*)
apply WF_addiu with P_epilog G_epilog.
(** P S -> P' S'**)
unfold P_epilog. unfold GoodFr. intros. 
induction s as [h rf]. inversion H0. simplRf H; simplRf'.
intuition. repeat (intros; rfinsen).
unfold G_fact. intros. induction s as [h rf].
induction s'' as [h'' rf'']. inversion H0. rewrite <- H2 in H1.
unall H1. intuition. apply H23; auto. repeat (intros; rfinsen).
othhold H29 rf.
(*j L_epilog*)
apply WF_j with P_epilog G_epilog.
constructor; [constructor|omega].
(**P => P'*)
intros; trivial.
(**P => G' => G*)
intros; trivial.
Save.


(***************************************************************)
Theorem WF_prolog :
  Infer (itpx2 (itpx1 T_prolog) psi_prolog) C_prolog.
Proof. unfold GoodFr. 
(** subiu sp sp 3 *)
apply WF_subiu with 
  (fun s : state => getValinR s ra <> 0 /\ getValinR s fp <= TOP /\
    getValinR s sp + 3 > ((getValinR s a0 + 1) * 3 ) /\ 
    getValinR s sp + 3 <= TOP /\ star (Stk TOP) arbh s)
  (fun s s' : state => Gand S0S1S2 ALLT s s' /\
    getValinR s' sp = getValinR s sp + 3 /\ 
    getValinR s' fp = getValinR s fp /\ getValinR s' ra = getValinR s ra /\
    HIntact TOP (getValinR s sp + 3) s s' /\ forall oth : assert, RfInsA oth ->
      star (Stk TOP) oth s -> star (Stk TOP) oth s').
(** P S -> P' S' *)
unfold P_prolog. intros. induction s as [h rf].
inversion H0. unfold GoodFr in H. simplRf H; simplRf'. splitland. splitex.
intuition. repeat (intros; rfinsen).
(** P S -> G' S' S'' -> G S S'' **)
unfold P_prolog. unfold G_prolog. intros.
induction s as [h rf]; induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. unall H1.
simplRf H; simplRf H1; simplRf'. intuition.
simplps'. apply plusex; auto. omega. apply H23; auto.
repeat (intros; rfinsen). othhold H32 rf.
(**sw sp 3 fp **)
apply WF_sw with
  (fun s : state => getValinR s ra <> 0 /\ getValinR s fp <= TOP /\
    getValinR s sp + 3 > ((getValinR s a0 + 1) * 3 ) /\ 
    getValinR s sp + 3 <= TOP /\ star (Stk TOP) arbh s)
  (fun s s' : state => Gand S0S1S2 ALLT s s' /\
    getValinR s' sp = getValinR s sp + 3 /\ getValinR s' ra = getValinR s ra /\
    getValinR s' fp = getVinH s (getValinR s sp + 3) /\ 
    HIntact TOP (getValinR s sp + 3) s s' /\ forall oth : assert, RfInsA oth ->
      star (Stk TOP) oth s -> star (Stk TOP) oth s').
(** sp+3 in dom **)
intros. simplRf H. splitland. splitstar. inlh. stkindom.
(** P S -> P' S' **)
intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'.
intuition. splitstar. writelh. stkindom. apply StkRet; auto.
stkindom. 
(** P S -> G' S' S'' -> G S S''*)
intros. induction s as [h rf]; induction s'' as [h'' rf''].
inversion H0. rewrite <- H3 in H1. unall H1. simplRf H; simplRf H1; simplRf'. 
intuition. rewrite H17. ungetVinH. simpleq. apply beq_eq_true; auto.
hintact'. apply H25; auto. splitstar. writelh. stkindom. apply StkRet; auto.
stkindom.
(** addiu fp sp + 3 **)
apply WF_addiu with
  (fun s : state => getValinR s ra <> 0 /\ GoodFr 3 s /\ 
    getValinR s fp > ((getValinR s a0 + 1) * 3 ) /\ star (Stk TOP) arbh s)
  (fun s s' : state => Gand S0S1S2 ALLT s s' /\
    getValinR s' sp = getValinR s sp + 3 /\ getValinR s' ra = getValinR s ra /\
    getValinR s' fp = getVinH s (getValinR s fp) /\ 
    HIntact TOP (getValinR s fp) s s' /\ forall oth : assert, RfInsA oth ->
      star (Stk TOP) oth s -> star (Stk TOP) oth s').
(*P S -> P' S'*)
intros. induction s as [h rf]. unfold GoodFr. 
inversion H0. simplRf H; simplRf'. intuition.
repeat (intros; rfinsen).
(*P S -> G' S' S'' -> G S S''*)
intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. unall H1. 
simplRf H; simplRf H1;simplRf'. intuition.
apply H24; auto. repeat (intros; rfinsen). othhold H33 rf.
(**sw_ fp 1 ra**)
apply WF_sw_ with
  (fun s : state => getVinH s (getValinR s fp - 1) <> 0 /\ GoodFr 3 s /\ 
    getValinR s fp > ((getValinR s a0 + 1) * 3 ) /\ star (Stk TOP) arbh s)
  (fun s s' : state => Gand S0S1S2 ALLT s s' /\
    getValinR s' sp = getValinR s sp + 3 /\ 
    getValinR s' ra = getVinH s (getValinR s fp - 1) /\
    getValinR s' fp = getVinH s (getValinR s fp) /\ 
    HIntact TOP (getValinR s fp) s s' /\ forall oth : assert, RfInsA oth ->
      star (Stk TOP) oth s -> star (Stk TOP) oth s').
(*fp-1 in dom **)
unfold GoodFr. intros. simplRf H; simplRf'. splitland; splitstar.
inlh. stkindom.
(*P S -> P' S'*)
unfold GoodFr. intros. induction s as [h rf]. 
inversion H0. simplRf H; simplRf'. split.
ungetVinH. simpleq. intuition. apply beq_eq_true; auto.
intuition. splitstar. writelh. stkindom. apply StkRet; auto.
stkindom. 
(*P S -> G' S' S'' -> G S S''*)
unfold GoodFr. intros. induction s as [h rf]; induction s'' as [h'' rf''].
inversion H0. rewrite <- H3 in H1. unall H1. simplRf H; simplRf H1; simplRf'.
intuition. rewrite H13. ungetVinH. simpleq. apply beq_eq_true; auto.
rewrite H18. ungetVinH. simplneq. apply beq_neq_false; auto.
omega. hintact'. left; omega. apply H26; auto. 
splitstar. writelh. stkindom. apply StkRet; auto. stkindom. 
(**sw_ fp 2 s0**)
apply WF_sw_ with
  (fun s : state => getVinH s (getValinR s fp - 1) <> 0 /\ GoodFr 3 s /\ 
    getValinR s fp > ((getValinR s a0 + 1) * 3 ) /\ star (Stk TOP) arbh s)
  (fun s s' : state => Gand ES0 ALLT s s' /\
    getValinR s' sp = getValinR s sp + 3 /\ 
    getValinR s' ra = getVinH s (getValinR s fp - 1) /\
    getValinR s' fp = getVinH s (getValinR s fp) /\ 
    getValinR s' s0 = getVinH s (getValinR s fp - 2) /\
    HIntact TOP (getValinR s fp) s s' /\ forall oth : assert, RfInsA oth ->
      star (Stk TOP) oth s -> star (Stk TOP) oth s').
(*fp-2 in dom **)
unfold GoodFr. intros. simplRf H; simplRf'. splitland; splitstar.
inlh. stkindom.
(*P S -> P' S'*)
unfold GoodFr. intros. induction s as [h rf]. 
inversion H0. simplRf H; simplRf'. split.
ungetVinH. simplneq. inversion_clear H. ungetVinH' H8. assumption.
apply beq_neq_false. omega. intuition.
splitstar. writelh. stkindom. apply StkRet; auto.
stkindom. 
(*P S -> G' S' S'' -> G S S''*)
unfold GoodFr. intros. induction s as [h rf]; induction s'' as [h'' rf''].
inversion H0. rewrite <- H3 in H1. unall H1. simplRf H; simplRf H1; simplRf'.
intuition. rewrite H22. ungetVinH. simpleq. apply beq_eq_true; auto.
rewrite H13. ungetVinH. simplneq. apply beq_neq_false; auto.
omega. rewrite H18. ungetVinH. simplneq. apply beq_neq_false; auto. omega.
hintact'. left; omega. apply H29; auto. 
splitstar. writelh. stkindom. apply StkRet; auto. stkindom. 
(* j L_fact*)
apply WF_j with P_fact G_fact. constructor.
(** P => P' *)
unfold P_fact. intros; intuition.
(** P => G' => G *)
unfold G_fact. intros; intuition.
Save.

