Require Import scap_rules.
Require Import scap.
Require Import cdsa.
Require Import heap2.
Require Import lib_lemma.
Require Import Omega.
Require Import properties.
Require Import properties2.

(*****************************************************) 
(***                  Malloc                       ***)
(*****************************************************)
Notation reqsize := a0.
Notation actsize := s0.
Notation retptr := v0.
(*Notation prev := s1.
Notation p := s2.
Parameter flist : addr.
Parameter TOP : addr.*)

Notation malloc_framesize := 4. (* 10 dwords *)

(*Notation header_size := 2.*)

Notation L_malloc_prolog := 20.
Notation L_malloc_init := 21.
Notation L_malloc_iter := 22.
Notation L_malloc_epilog := 23.
Notation L_malloc_comp := 24.
Notation L_malloc_retptr := 25.
Notation L_malloc_split := 26.
Notation L_malloc_next := 27.
Notation L_malloc_lprv := 28.


Definition C_malloc_prolog : iseq := seq (
  subiu sp sp malloc_framesize) (seq (
  sw sp malloc_framesize fp)    (seq (
  addiu fp sp malloc_framesize) (seq (
  sw_ fp 1 s0)                (seq (
  sw_ fp 2 s1)                (seq (
  sw_ fp 3 s2)                (
  j L_malloc_init)             ))))).

Definition C_malloc_init : iseq := seq (
  addiu actsize reqsize header_size) (seq (
  addiu prev zero pnil)              (seq (
  lw p zero flist)                   (seq (
  addiu retptr zero pnil)            (
  j L_malloc_iter)))).

Definition C_malloc_iter : iseq := seq (
  bgtz p L_malloc_comp)            (
  j L_malloc_epilog).

Definition C_malloc_comp : iseq := seq (
  lw t0 p sz)                      (seq (
  addiu t1 actsize header_size)    (seq (
  subu t1 t0 t1)                   (seq (
  bgtz t1 L_malloc_split)          (seq (
  subu t1 actsize t0)              (seq (
  bgtz t1 L_malloc_next)           (seq (
  lw t0 p nxt)                     (seq (
  bgtz prev L_malloc_lprv)         (seq (
  sw zero flist t0)                (
  j L_malloc_retptr))))))))).

Definition C_malloc_retptr : iseq := seq (
  addiu retptr p header_size)        (
  j L_malloc_epilog).

Definition C_malloc_split : iseq := seq (
  subu t0 t0 actsize)               (seq (
  sw p sz t0        )               (seq (
  addu p p t0)                      (seq (
  sw p sz actsize)                  (
  j L_malloc_retptr)))).

Definition C_malloc_next : iseq := seq (
  addu prev p zero)                (seq (
  lw p p nxt)                      (
  j L_malloc_iter)).

Definition C_malloc_lprv : iseq := seq (
  sw prev nxt t0)                  (
  j L_malloc_retptr).

Definition C_malloc_epilog : iseq := seq (
  lw_ s2 fp 3)                    (seq (
  lw_ s1 fp 2)                    (seq (
  lw_ s0 fp 1)                    (seq (
  lw fp fp 0)                     (seq (
  addiu sp sp malloc_framesize)     (
  jr ra)                          )))).

(*************************************************)
(*************Malloc Types ***********************)
(*************************************************)
Notation MALLOC_FRAME := FREE_FRAME.

Notation ET0T1 := (
  Gand (RIntact t3) (Gand (RIntact t4) (Gand (RIntact t5) (
  Gand (RIntact t6) (Gand (RIntact t7) (Gand (RIntact t8) (
  Gand (RIntact t2) (Gand (RIntact a0) (RIntact t9))))))))).

Definition MCPOST (oth : assert) : assert :=
  fun s : state => (exists fv : addr, exists lnk : addr, exists siz : nat, 
    star (pto flist fv) (star (MBlk (getValinR s retptr) lnk siz)
      (star (Good fv)(star (Stk TOP) oth))) s /\ siz >= getValinR s a0 + 2) \/
  (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s /\ 
    getValinR s retptr = pnil).

Definition P_malloc_prolog : assert :=
  fun s : state =>
    getValinR s sp < getValinR s fp /\ getValinR s sp > 5 /\ 
    getValinR s fp < TOP /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 /\
    exists fv : addr, 
      star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s.

Definition G_malloc_prolog : action :=
  fun s s' : state =>
    (Gand S0S1S2 (Gand ET0T1 Rqd)) s s' /\ HIntact TOP (getValinR s sp) s s' /\
    forall oth : assert, RfInsA oth -> (exists fv : addr, 
      star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
    MCPOST oth s'.

Notation T_malloc_prolog := (cdspec P_malloc_prolog G_malloc_prolog).

Definition P_malloc_init : assert :=
  fun s : state =>
    MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s.

Definition G_malloc_init : action :=
  fun s s' : state =>
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
      MCPOST oth s'.

Notation T_malloc_init := (cdspec P_malloc_init G_malloc_init).
     
Definition P_malloc_epilog : assert :=
  fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ MCPOST arbh s.

Definition G_malloc_epilog : action :=
  fun s s' : state =>
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      MCPOST oth s -> MCPOST oth s'.

Notation T_malloc_epilog := (cdspec P_malloc_epilog G_malloc_epilog).

Definition P_malloc_iter : assert := 
  fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s retptr = pnil /\
    exists fv : addr, star (pto flist fv) (star 
      (MidL fv (getValinR s prev) (getValinR s p)) (star (Stk TOP) arbh)) s.

Definition G_malloc_iter : action :=
  fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, star (pto flist fv) (star 
	(MidL fv (getValinR s prev) (getValinR s p)) (star (Stk TOP) oth)) s) ->
      MCPOST oth s'.

Notation T_malloc_iter := (cdspec P_malloc_iter G_malloc_iter).

Definition P_malloc_comp : assert :=
  fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s retptr = pnil 
    /\ getValinR s p > pnil /\
    exists fv : addr, star (pto flist fv) (star 
      (MidL fv (getValinR s prev) (getValinR s p)) (star (Stk TOP) arbh)) s.

Definition G_malloc_comp : action :=
  fun s s' : state =>
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, star (pto flist fv) (star 
	(MidL fv (getValinR s prev) (getValinR s p)) (star (Stk TOP) oth)) s) ->
      MCPOST oth s'.

Notation T_malloc_comp := (cdspec P_malloc_comp G_malloc_comp).
  
Definition P_malloc_split : assert :=
  fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil /\
    getValinR s t0 > getValinR s actsize + 2 /\
    exists fv : addr, exists q : addr, exists n : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil).

Definition G_malloc_split : action :=
  fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ 
    forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, 
	star (pto flist fv) (star 
	  (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil)) ->
      MCPOST oth s'.

Notation T_malloc_split := (cdspec P_malloc_split G_malloc_split).

Definition P_malloc_next : assert :=
  fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil 
    /\ getValinR s retptr = pnil /\
    exists fv : addr, exists q : addr, exists n : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil).

Notation G_malloc_next := G_malloc_split.
Notation T_malloc_next := (cdspec P_malloc_next G_malloc_next).

Definition P_malloc_lprv := 
  fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil
    /\ getValinR s prev > pnil /\
    exists fv : addr, exists q : addr, exists n : nat, exists siz : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q siz) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil) /\ 
      siz >= getValinR s actsize /\ (getValinR s t0 = q).

Definition G_malloc_lprv : action :=
  fun s s' : state => (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
    (exists fv : addr, exists q : addr, exists n : nat, exists siz : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q siz) (MBlkLst n q pnil)))
	(star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil) /\ 
      siz >= getValinR s actsize /\ getValinR s t0 = q) ->
    MCPOST oth s'.
Notation T_malloc_lprv := (cdspec P_malloc_lprv G_malloc_lprv).

Definition P_malloc_retptr : assert :=
  fun s : state => MALLOC_FRAME s /\ getValinR s a0 > 0
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s ra <> 0 /\
    exists fv : addr, exists siz : nat, exists lnk : addr, 
      star (pto flist fv) (star (star (MBlk (getValinR s p + 2) lnk siz) (Good fv))
	(star (Stk TOP) arbh)) s /\ siz >= getValinR s actsize.

Definition G_malloc_retptr : action :=
  fun s s' : state => (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists siz : nat, exists lnk : addr, 
	star (pto flist fv) (star (star (MBlk (getValinR s p + 2) lnk siz) (Good fv))
	  (star (Stk TOP) oth)) s /\ siz >= getValinR s actsize) ->
      MCPOST oth s'.

Notation T_malloc_retptr := (cdspec P_malloc_retptr G_malloc_retptr).


(************************************************************)
(***************** Proof of Codes of Malloc******************)
(***************** Xiang Sen : 2005-7-4 *********************)
(************************************************************)

Ltac mc_unall H1 := try (unfold S0S1S2 in H1; unfold S0S1S2; unfold POPS0S1S2 in H1;
  unfold RESRqd;unfold RESRqd in H1; unfold POPS0S1S2; unfold ES0S1S2 in H1; unfold ES0S1S2; 
    unGand H1; unRIntact H1).
Ltac simplps' :=
  match goal with
    | [ H1 : context [(?X -?N + ?N)] |- _] =>
      replace (X - N + N) with X in H1; auto
  end.

Ltac othhold H rf := try (unfold RfInsA in H; intros; apply H with rf; auto).

Notation psi_malloc_prolog := (
  consCT L_malloc_init T_malloc_init emptyCT).
Notation psi_malloc_init := (
  consCT L_malloc_iter T_malloc_iter emptyCT).
Notation psi_malloc_iter := (
  consCT L_malloc_comp T_malloc_comp(
  consCT L_malloc_epilog T_malloc_epilog emptyCT)).
Notation psi_malloc_comp := (
  consCT L_malloc_split T_malloc_split (
  consCT L_malloc_next T_malloc_next (
  consCT L_malloc_lprv T_malloc_lprv (
  consCT L_malloc_retptr T_malloc_retptr emptyCT)))).
Notation psi_malloc_retptr := (consCT L_malloc_epilog T_malloc_epilog emptyCT).
Notation psi_malloc_lprv := (consCT L_malloc_retptr T_malloc_retptr emptyCT).
Notation psi_malloc_next := (consCT L_malloc_iter T_malloc_iter emptyCT).
Notation psi_malloc_split := (consCT L_malloc_retptr T_malloc_retptr emptyCT).

(***************************************************************)

Theorem WF_malloc_split :
  Infer (itpx2 (itpx1 T_malloc_split) psi_malloc_split) C_malloc_split.
Proof.
unfold C_malloc_split.
(** subu t0 t0 actsize ***)
apply WF_subu with
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil /\
    getValinR s t0 > 2 /\ exists fv : addr, exists q : addr, exists n : nat, 
      exists siz : nat, star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q siz) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil) /\
      siz =  getValinR s actsize + getValinR s t0)
  (fun s s' : state => (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, exists siz : nat, 
	star (pto flist fv) (star (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q siz) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil) 
	/\ siz = getValinR s actsize + getValinR s t0) ->
      MCPOST oth s').
(**P S -> P' nextS **)
unfold P_malloc_split. unfold FREE_FRAME. intros. induction s as [h rf].
inversion H0. simplRf H. simplRf'. splitland; getland. omega.
splitex. exists x; exists x0; exists x1. exists (lookupR rf t0). splitland.
getland. repeat (intros; rfinsen). omega.
(**P S -> G' nextS S' -> G S S'**)
unfold P_malloc_split. unfold G_malloc_split. unfold RESRqd. 
intros. induction s as [h rf]. induction s'' as [h'' rf''].
mc_unall H1. inversion H0. rewrite <- H2 in H1. simplRf'; simplRf H; simplRf H1.
splitland; getland. clear H35. intros; apply H12; trivial. splitex.
exists x; exists x0; exists x1. exists (lookupR rf t0). splitland.
getland. repeat (intros; rfinsen). othhold H34 rf. omega.
(** sw p size t0 **)
apply WF_sw with
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil /\
    getValinR s t0 > 2 /\ exists fv : addr, exists q : addr, exists n : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (star (MBlk (getValinR s p + 2) q (getValinR s t0))
	    (ptoanynlist (getValinR s p + getValinR s t0) (getValinR s actsize)))
	  (MBlkLst n q pnil))) (star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil))
  (fun s s' : state => (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, 
	star (pto flist fv) (star (star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (star (MBlk (getValinR s p + 2) q (getValinR s t0))
	    (ptoanynlist (getValinR s p + getValinR s t0) (getValinR s actsize)))
	  (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil)) ->
      MCPOST oth s').
(** p->size in dom **)
intros. simplRf H. splitland. splitex. splitland. repeat splitstar.
inrh. inlh. inrh. inlh. mblkindom. 
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. 
simplRf H; simplRf'. splitland; getland. splitex. exists x; exists x0; exists x1.
splitland. getland. splitstar. writerh. repeat splitstar. inlh. inrh. inlh. mblkindom.
splitstar. writelh. clear H23. repeat splitstar. inrh. inlh. mblkindom.
clear H23. splitstar. writerh. splitstar. inlh. mblkindom. splitstar. 
writelh. mblkindom. apply Sep_MBlk with x2; auto. 
(** P S -> G' nextS S' -> G S S' **)
unfold FREE_FRAME. unfold RESRqd. intros. induction s as [h rf].
induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
mc_unall H1. simplRf'; simplRf H; simplRf H1. splitland. splitex.
splitland. splitstar. starassocl H43. splitstar. starassocr H46.
splitstar. starrev H47. starassocr H48. splitstar. starrev H50. splitstar.
getland. hintact'. notinstk. mblkindom. fromnotinstk. inversion_clear H56.
left; omega. right; auto. rewrite H14. ungetVinH. simplneq. 
apply beq_neq_false. apply disj_indom_neq with x11 x12; auto. mblkindom.
stkindom. rewrite H10. ungetVinH. simplneq. 
apply beq_neq_false. apply disj_indom_neq with x11 x12; auto. mblkindom.
stkindom. rewrite H17. ungetVinH. simplneq. 
apply beq_neq_false. apply disj_indom_neq with x11 x12; auto. mblkindom.
stkindom. rewrite H12. ungetVinH. simplneq. 
apply beq_neq_false. apply disj_indom_neq with x11 x12; auto. mblkindom.
stkindom. clalldisj. clallheq. clallha. clallh. intros; apply H13; trivial.
splitex. exists x3; exists x4; exists x6.
splitland. getland. splitstar. writerh. repeat splitstar. inlh. inrh. inlh. mblkindom.
splitstar. writelh. clear H49. repeat splitstar. inrh. inlh. mblkindom.
clear H49. splitstar. writerh. splitstar. inlh. mblkindom. splitstar. 
writelh. mblkindom. apply Sep_MBlk with x7; auto. 
(**addu p p t0 **)
apply WF_addu with 
    (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > 2 /\
    getValinR s t0 > 2 /\ exists fv : addr, exists q : addr, exists n : nat, 
      exists tp : addr, 
      star (pto flist fv) (star (star (EndL fv (getValinR s prev) tp)(
	  star (star (MBlk (tp + 2) q (getValinR s t0))
	    (ptoanynlist (getValinR s p) (getValinR s actsize)))
	  (MBlkLst n q pnil))) (star (Stk TOP) arbh)) s /\ (tp < q \/ q = pnil))
  (fun s s' : state => (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, exists tp : addr, 
	star (pto flist fv) (star (star (EndL fv (getValinR s prev) tp)(
	  star (star (MBlk (tp + 2) q (getValinR s t0))
	    (ptoanynlist (getValinR s p) (getValinR s actsize)))
	  (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (tp < q \/ q = pnil)) ->
      MCPOST oth s').
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0.
simplRf H; simplRf'. splitland. getland. omega. splitex. splitland.
exists x; exists x0; exists x1; exists (lookupR rf p). getland.
repeat (intros; rfinsen). 
(** P S -> G' nextS S' -> G S S' **)
unfold RESRqd. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. mc_unall H1. simplRf'; simplRf H; simplRf H1.
splitland; getland. intros; apply H12; trivial. clear H35.  splitex. splitland.
exists x; exists x0; exists x1; exists (lookupR rf p). getland.
repeat (intros; rfinsen). othhold H34 rf.
(** sw p next actsize **)
apply WF_sw with 
    (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > 2 /\
    getValinR s t0 > 2 /\ exists fv : addr, exists q : addr, exists n : nat, 
      exists tp : addr, exists lnk : addr, 
      star (pto flist fv) (star (star (EndL fv (getValinR s prev) tp)(
	  star (star (MBlk (tp + 2) q (getValinR s t0))
	    (MBlk (getValinR s p + 2) lnk (getValinR s actsize)))
	  (MBlkLst n q pnil))) (star (Stk TOP) arbh)) s /\ (tp < q \/ q = pnil))
  (fun s s' : state => (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, exists tp : addr, exists lnk : addr, 
	star (pto flist fv) (star (star (EndL fv (getValinR s prev) tp)(
	  star (star (MBlk (tp + 2) q (getValinR s t0))
	    (MBlk (getValinR s p + 2) lnk (getValinR s actsize)))
	  (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (tp < q \/ q = pnil)) ->
      MCPOST oth s').
(** p+1 in dom **)
intros. simplRf H. splitland. splitex. splitland. repeat splitstar.
inrh. inlh. inrh. inlh. inrh. pnalindom. intuition.
(** P S -> P' next **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0.
simplRf H; simplRf'. splitland; getland. splitex. splitland.
exists x; exists x0; exists x1; exists x2. repeat splitstar. 
exists (hread' x14 (lookupR rf p)). getland. move H14 after H35.
writerh. inlh. inrh. inlh. inrh. pnalindom. intuition. 
move H19 after H35. writelh. getstar'. inrh. inlh. inrh. pnalindom. intuition.
move H25 after H35. writerh. inlh. inrh. pnalindom. intuition.
move H28 after H35. writelh. inrh. pnalindom. intuition.
writerh. pnalindom. intuition. apply Pnal_MBlk; auto. omega. omega.
(** P S -> G' nextS S' -> G S S'*)
unfold FREE_FRAME. unfold RESRqd. intros. induction s as [h rf].
induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
mc_unall H1. simplRf H; simplRf H1; simplRf'. splitland. splitex.
splitland. splitstar. starassocl H43. splitstar. starassocl H46.
starassocl H45. splitstar. starrev H49. starassocl H48. splitstar.
starrev H52. starassocr H51. splitstar. splitstar. getland. 
hintact'. notinstk. pnalindom. intuition. fromnotinstk. inversion_clear H58.
left; omega. right; trivial. rewrite H14. ungetVinH. simplneq. 
apply beq_neq_false. apply disj_indom_neq with x13 x14; auto.
pnalindom. intuition. stkindom. rewrite H10. ungetVinH. simplneq. 
apply beq_neq_false. apply disj_indom_neq with x13 x14; auto.
pnalindom. intuition. stkindom. rewrite H17. ungetVinH. simplneq. 
apply beq_neq_false. apply disj_indom_neq with x13 x14; auto.
pnalindom. intuition. stkindom. rewrite H12. ungetVinH. simplneq. 
apply beq_neq_false. apply disj_indom_neq with x13 x14; auto.
pnalindom. intuition. stkindom. clalldisj. clallheq. clallha. clallh. 
intros; apply H13; trivial. splitex. splitland.
exists x3; exists x4; exists x6; exists x7. repeat splitstar. 
exists (hread' x21 (lookupR rf p)). getland. move H37 after H62.
writerh. inlh. inrh. inlh. inrh. pnalindom. intuition. 
move H44 after H62. writelh. getstar'. inrh. inlh. inrh. pnalindom. intuition.
move H51 after H62. writerh. inlh. inrh. pnalindom. intuition.
move H54 after H62. writelh. inrh. pnalindom. intuition.
writerh. pnalindom. intuition. apply Pnal_MBlk; auto. omega. omega.
(** j L_malloc_retptr **)
apply WF_j with P_malloc_retptr G_malloc_retptr.
(* psi(f) = theta**)
constructor. 
(* P => P' **)
unfold P_malloc_retptr. intros. splitland; getland. 
splitex. exists x; exists (getValinR s actsize); exists x3. splitland.
induction s as [h rf]. splitstar. split; [idtac| omega]. getstar'. 
splitstar. getstar'. clear H13. starassocr H11. starrev H12. starassocr H11.
starassocr H12. starrev H11. splitstar. getstar'. starassocl H15. 
starrev H14. splitstar. splitstar. 
assert (EndL x x2 x0 (happend x12 x13, rf)). 
apply Compose_EndL_MBlk_to_EndL with (getValinR (h, rf) prev) (getValinR (h, rf) t0);auto.
replace (2+x2) with (x2+2); [auto|omega]. 
assert (EndL x x2 x0 (x10, rf)). apply h_subst_EndL with (happend x12 x13); auto.
clear H20. clear H21. clear H19. elg.
(**P => G' => G **)
unfold G_malloc_retptr. intros. splitland; getland. clear H11.
intros; apply H5; trivial. splitex. 
exists x; exists (getValinR s actsize); exists x3. splitland.
induction s as [h rf]. splitstar. split; [idtac| omega]. getstar'. 
splitstar. getstar'. clear H19. starassocr H17. starrev H18. starassocr H17.
starassocr H18. starrev H17. splitstar. getstar'. starassocl H21. 
starrev H20. splitstar. splitstar. 
assert (EndL x x2 x0 (happend x12 x13, rf)). 
apply Compose_EndL_MBlk_to_EndL with (getValinR (h, rf) prev) (getValinR (h, rf) t0);auto.
replace (2+x2) with (x2+2); [auto|omega]. 
assert (EndL x x2 x0 (x10, rf)). apply h_subst_EndL with (happend x12 x13); auto.
clear H26. clear H27. clear H25. elg.
Save.


(***************************************************************)

Theorem WF_malloc_next :
  Infer (itpx2 (itpx1 T_malloc_next) psi_malloc_next) C_malloc_next.
Proof. 
unfold C_malloc_next. 
(** addu prev p zero **)
apply WF_addu with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s prev = getValinR s p /\
    getValinR s retptr = pnil /\
    exists fv : addr, exists q : addr, exists n : nat, exists tp : addr, 
      star (pto flist fv) (star 
	(star (EndL fv tp (getValinR s prev))(
	  star (MBlk (getValinR s prev + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s prev < q \/ q = pnil))
  (fun s s' : state => (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, exists tp : addr, 
      star (pto flist fv) (star (star (EndL fv tp (getValinR s prev))(
	  star (MBlk (getValinR s prev + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) oth)) s /\ (getValinR s prev < q \/ q = pnil)) ->
      MCPOST oth s').
(** P S -> P' nextS **)
unfold P_malloc_next. unfold FREE_FRAME. intros. induction s as [h rf].
inversion H0. simplRf H; simplRf'. intuition. simplz. fauto. splitex.
exists x; exists x0; exists x1. exists (lookupR rf prev).
splitland; getland. simplz. simpl0p. repeat (intros; rfinsen). fauto.
simplz. simpl0p. trivial. fauto.
(** P S -> G' nextS -> G S S'**)
unfold G_malloc_split. unfold P_malloc_next. unfold RESRqd. 
intros. mc_unall H1. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. simplRf H; simplRf H1; simplRf'. 
splitland. getland. clear H35. intros; apply H12; trivial. splitex.
exists x; exists x0; exists x1. exists (lookupR rf prev). simplz. simpl0p.
splitland; getland.  repeat (intros; rfinsen). othhold H34 rf. fauto.
(** lw p p next ***)
apply WF_lw with
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s retptr = pnil /\
    exists fv : addr, exists q : addr, exists n : nat, exists tp : addr, 
      star (pto flist fv) (star (star (EndL fv tp (getValinR s prev))(
	  star (MBlk (getValinR s prev + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s prev < q \/ q = pnil) /\ getValinR s p = q)
  (fun s s' : state => (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, exists tp : addr, 
	star (pto flist fv) (star (star (EndL fv tp (getValinR s prev))(
	  star (MBlk (getValinR s prev + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) oth)) s /\ (getValinR s prev < q \/ q = pnil) /\ getValinR s p = q) ->
      MCPOST oth s').
(**p->next in dom **)
intros. simpl0p. simplRf H. splitland. splitex. splitland. repeat splitstar.
rewrite <- H3. inrh. inlh. inrh. inlh. mblkindom.
(**P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. 
simplRf H; simplRf'. splitland; getland. splitex. splitland.
exists x0; exists x1; exists x2; exists x3. getland. repeat (intros; rfinsen).
simpl0p'. replace x with (hread' h (lookupR rf p)). unfold hread'.
repeat splitstar. rewrite <- H11. hread2. readrh. hread2. readlh. hread2. readrh. 
hread2. readlh. foldhread'. mblkchunk. fauto. mblkindom. inlh. mblkindom.
fauto. inrh. inlh. mblkindom. inlh. inrh. inlh. mblkindom. unfold hread'.
rewrite H7; trivial. 
(** P S -> P' nextS S' -> G S S' **)
unfold RESRqd. intros. mc_unall H1. induction s as [h rf].
induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
simplRf H; simplRf H1; simplRf'. splitland; getland. clear H36.
intros; apply H13; trivial. splitex. splitland.
exists x0; exists x1; exists x2; exists x3. getland. repeat (intros; rfinsen). 
othhold H35 rf. simpl0p'. replace x with (hread' h (lookupR rf p)). unfold hread'.
repeat splitstar. rewrite <- H33. hread2. readrh. hread2. readlh. hread2. readrh. 
hread2. readlh. foldhread'. mblkchunk. fauto. mblkindom. inlh. mblkindom.
fauto. inrh. inlh. mblkindom. inlh. inrh. inlh. mblkindom. unfold hread'.
rewrite H8; trivial.
(** j L_malloc_iter **)
apply WF_j with P_malloc_iter G_malloc_iter.
(** psi(f) = theta **)
constructor.
(** P => P' **)
unfold P_malloc_iter. intros. induction s as [h rf]. splitland; getland.
splitex. splitland. exists x. splitstar. getstar'. splitstar. getstar'.
rewrite H7. unfold MidL. exists x1. clear H13. starassocr H11. 
splitstar. getstar'. splitstar. apply h_subst_EndL with (happend x9 x10); auto.
apply Compose_EndL_MBlk_to_EndL with x2 (getValinR (h, rf) t0); auto.
replace (2 + getValinR (h, rf) prev) with (getValinR (h, rf) prev + 2);
[auto|omega].
(** P => G' => G **)
unfold G_malloc_iter. intros. splitland. getland.
intros; apply H5; trivial. clear H10. induction s as [h rf]. splitex.
splitland. exists x. splitstar. getstar'. splitstar. getstar'. rewrite H13.
unfold MidL. exists x1. clear H19. starassocr H17. splitstar. getstar'.
splitstar. apply h_subst_EndL with (happend x9 x10); auto.
apply Compose_EndL_MBlk_to_EndL with x2 (getValinR (h, rf) t0); auto.
replace (2 + getValinR (h, rf) prev) with (getValinR (h, rf) prev + 2);
[auto|omega].
Save.

(***************************************************************)

Theorem WF_malloc_lprv :
  Infer (itpx2 (itpx1 T_malloc_lprv) psi_malloc_lprv) C_malloc_lprv.
Proof.
unfold C_malloc_lprv. 
(**sw prev nxt t0 **)
apply WF_sw with P_malloc_retptr G_malloc_retptr.
(** prev->nxt in dom **) 
unfold P_malloc_lprv. intros. simpl0p. simplRf H. splitland. splitland.
splitex. splitland. repeat splitstar. inrh. inlh. inlh. endlindom. omega.
(** P => P' **)
unfold P_malloc_lprv. unfold P_malloc_retptr. unfold FREE_FRAME.
intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'.
intuition. splitex. exists x; exists x2; exists x0. splitland.
getland. simpl0p. splitstar. writerh. repeat splitstar. inlh. inlh.
endlindom. omega. splitstar. writelh. clear H24. splitstar. inlh. endlindom.
omega. clear H24. starrev H22. starassocl H18. splitstar. writerh. splitstar.
inrh. endlindom. omega. rewrite H19. splitstar. 
apply MBLkLst_EndL_Good with x9 x10 (lookupR rf p) x1; auto.
assert (lookupR rf prev < lookupR rf p). apply EndL_Po_gt_Pr with x10 rf x;auto.
omega.
(** P S -> G' nextS S' -> G S S' **)
unfold G_malloc_retptr; unfold G_malloc_lprv; unfold P_malloc_lprv.
unfold FREE_FRAME. unfold RESRqd. intros. induction s as [h rf].
induction s'' as [h'' rf'']. inversion H0. simpl0p'. rewrite <- H3 in H1.
mc_unall H1. simplRf'; simplRf H; simplRf H1. splitland. splitex.
splitland. splitstar. starrev H44. starassocr H43. splitstar. clear H47.
starrev H45. starassocr H46. splitstar. splitstar. getland. hintact'.
notinstk. endlindom. omega. fromnotinstk. inversion_clear H53. left; omega.
right; trivial. rewrite H14. ungetVinH. simplneq. apply beq_neq_false.
apply disj_indom_neq with x9 x10; auto. endlindom. omega. stkindom.
rewrite H10. ungetVinH. simplneq. apply beq_neq_false.
apply disj_indom_neq with x9 x10; auto. endlindom. omega. stkindom.
rewrite H17. ungetVinH. simplneq. apply beq_neq_false.
apply disj_indom_neq with x9 x10; auto. endlindom. omega. stkindom.
rewrite H12. ungetVinH. simplneq. apply beq_neq_false.
apply disj_indom_neq with x9 x10; auto. endlindom. omega. stkindom.
clalldisj. clallheq. clallha. clallh. intros; apply H13; trivial.
splitex. exists x3; exists x6; exists x4. splitland.
getland. splitstar. writerh. repeat splitstar. inlh. inlh.
endlindom. omega. splitstar. writelh. clear H51. splitstar. inlh. endlindom.
omega. clear H51. starrev H48. starassocl H44. splitstar. writerh. splitstar.
inrh. endlindom. omega. rewrite H45. splitstar. 
apply MBLkLst_EndL_Good with x14 x15 (lookupR rf p) x5; auto.
assert (lookupR rf prev < lookupR rf p). apply EndL_Po_gt_Pr with x15 rf x3;auto.
omega. 
(** j L_malloc_retptr **)
apply WF_j with P_malloc_retptr G_malloc_retptr.
(** psi (f)= theta*)
constructor.
(** P => P' **)
intros; trivial.
(** P => G' => G**)
intros; trivial.
Save.

(***************************************************************)

Theorem WF_malloc_retptr :
  Infer (itpx2 (itpx1 T_malloc_retptr) psi_malloc_retptr) C_malloc_retptr.
Proof.
unfold C_malloc_retptr. 
(**addiu retptr p 2 ***)
apply WF_addiu with P_malloc_epilog G_malloc_epilog.
(** P S -> P' next S **)
unfold P_malloc_retptr. unfold P_malloc_epilog. unfold FREE_FRAME.
unfold MCPOST. intros. induction s as [h rf]. inversion H0. 
simplRf H; simplRf'. splitland; getland. left. rewrite <- H8.
splitex. exists x; exists x1; exists x0. intuition.
splitstar. getstar'. starassocl H17. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S' **)
unfold P_malloc_retptr; unfold G_malloc_epilog. unfold G_malloc_retptr.
unfold MCPOST; unfold RESRqd. intros. mc_unall H1. induction s as [h rf].
induction s'' as [h'' rf'']. inversion H0. rewrite <- H2 in H1.
simplRf H; simplRf H1; simplRf'. splitland; getland. clear H33.
intros; apply H12; trivial. left. rewrite <- H30. splitex. 
exists x; exists x1; exists x0. intuition. splitstar. getstar'.
starassocl H38. repeat (intros; rfinsen). othhold H32 rf.
(** j L_malloc_epilog **)
apply WF_j with P_malloc_epilog G_malloc_epilog.
(** psi (f) = theta **)
constructor.
(** P => P' **)
intros; trivial.
(** P => G' => G **)
intros; trivial.
Save.

(***************************************************************)

Theorem WF_malloc_comp :
  Infer (itpx2 (itpx1 T_malloc_comp) psi_malloc_comp) C_malloc_comp.
Proof.
unfold C_malloc_comp.
(*** lw t0 p size ***)
apply WF_lw with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil /\
    getValinR s retptr = pnil /\
    exists fv : addr, exists q : addr, exists n : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil))
  (fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ 
    forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, 
	star (pto flist fv) (star 
	  (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil)) ->
      MCPOST oth s').
(** p+1 in dom **)
unfold P_malloc_comp. intros. simplRf H. splitland. splitex. repeat splitstar.
inrh. inlh. midlindom. omega.
(** P S -> P' nextS **)
unfold P_malloc_comp. unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. 
simplRf H; simplRf'. intuition. splitex. exists x0. repeat splitstar. unfold MidL in H20.
splitex. splitstar. inversion H28. clear H30. clear p. assert ((lookupR rf p) <> 0).
omega. contradiction. exists p'; exists m. split; [idtac|trivial]. getstar'. 
getstar'. getstar'. rfinsen. getstar';[rfinsen|rfinsen]. rfinsen. 
replace x with (hread' h (lookupR rf s2 + 1)); [idtac|unfold hread'; rewrite H7; trivial].
unfold hread'. hread2. readrh. hread2. readlh. hread2. readrh. hread2. readlh.
foldhread'. replace (hread' h2 (lookupR rf s2 + 1)) with n; [trivial|symmetry; mblkchunk].
fauto. mblkindom. inlh. mblkindom. fauto. inrh. inlh. mblkindom. inlh. inrh. inlh. 
mblkindom. getstar'. rfinsen. 
(** P S -> G' nextS S' -> G S S' **)
unfold G_malloc_comp. unfold RESRqd. intros. mc_unall H1. induction s as [h rf].
unfold P_malloc_comp in H. induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
simplRf H1; simplRf'; simplRf H. splitland; getland. clear H36.
intros; apply H13; trivial. splitex. exists x0.  repeat splitstar. 
unfold MidL in H41. splitex. splitstar. inversion H49. 
assert ((lookupR rf s2) <> 0). omega. contradiction. 
exists p'; exists m. split; [idtac|trivial]. getstar'.  getstar'. getstar'. 
rfinsen. getstar';[rfinsen|rfinsen]. rfinsen.  
replace x with (hread' h (lookupR rf s2 + 1)); [idtac|unfold hread'; rewrite H8; trivial].
unfold hread'. hread2. readrh. hread2. readlh. hread2. readrh. hread2. readlh.
foldhread'. replace (hread' h2 (lookupR rf s2 + 1)) with n; [trivial|symmetry; mblkchunk].
fauto. mblkindom. inlh. mblkindom. fauto. inrh. inlh. mblkindom. inlh. inrh. inlh. 
mblkindom. getstar'. rfinsen. othhold H35 rf.
apply WF_addiu with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil 
    /\ getValinR s t1 = getValinR s actsize + 2 /\ getValinR s retptr = pnil /\
    exists fv : addr, exists q : addr, exists n : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil))
  (fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ 
    forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, 
	star (pto flist fv) (star 
	  (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil)) ->
      MCPOST oth s').
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'.
splitland; getland. splitex. splitland. exists x; exists x0; exists x1. getland.
repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S' **)
unfold RESRqd. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. mc_unall H1. clear H. simplRf'; simplRf H1.
splitland; getland. intros; apply H11; trivial. splitex. splitland.
exists x; exists x0; exists x1. getland. repeat (intros; rfinsen). othhold H27 rf.
(**subu t1 t0 t1 **)
apply WF_subu with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil 
    /\ getValinR s t1 = getValinR s t0 - (getValinR s actsize + 2) /\
    getValinR s retptr = pnil /\
    exists fv : addr, exists q : addr, exists n : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil))
  (fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ 
    forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, 
	star (pto flist fv) (star 
	  (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil)) ->
      MCPOST oth s').
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. simplRf H.
simplRf'. splitland; getland. splitex. splitland. exists x; exists x0; exists x1.
getland. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S' **)
unfold RESRqd. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. mc_unall H1. clear H. simplRf'; simplRf H1.
splitland; getland. intros; apply H11; trivial. splitex. splitland.
exists x; exists x0; exists x1. getland. repeat (intros; rfinsen). othhold H27 rf.
(** bgtz t1 L_malloc_split **)
apply WF_bgtz with P_malloc_split
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil 
    /\ getValinR s t0 <= (getValinR s actsize + 2) /\ getValinR s retptr = pnil /\
    exists fv : addr, exists q : addr, exists n : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil))
  G_malloc_split
  (fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ 
    forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, 
	star (pto flist fv) (star 
	  (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil)) ->
      MCPOST oth s').
(** psi (f) = theta **)
constructor. 
(** t1 > 0 -> P => P' **) Focus 2.
unfold P_malloc_split. intros. simplRf'; simplRf H. intuition. 
(** t1 > 0 -> P => G' => G **) Focus 2.
unfold G_malloc_split. intros. simplRf H; simplRf H1; simplRf'.
intuition. 
(** t1 <= 0 -> P => P' **) Focus 2.
intros. simplRf'; simplRf H. intuition.
(** t1 <= 0 -> P => G' => G **) Focus 2.
intros. simplRf H; simplRf H1; simplRf'. intuition.
(** subu t1 actsize t0 **)
apply WF_subu with
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil 
    /\ getValinR s t0 <= (getValinR s actsize + 2) 
    /\ getValinR s t1 = getValinR s actsize - getValinR s t0 /\
    getValinR s retptr = pnil /\
    exists fv : addr, exists q : addr, exists n : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil))
  (fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ 
    forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, 
	star (pto flist fv) (star 
	  (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil)) ->
      MCPOST oth s').
(**P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. 
simplRf H. simplRf'. intuition. splitex. exists x; exists x0; exists x1.
splitland; getland. repeat (intros; rfinsen).
(**P S -> G' nextS S' -> G S S' **)
unfold RESRqd. intros. clear H. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H5 in H1. mc_unall H1. simplRf H1. simplRf'. intuition.
apply H17; trivial. splitex. exists x; exists x0; exists x1. splitland; getland.
repeat (intros; rfinsen). othhold H27 rf.
(**bgtz t1 L_malloc_next **)
apply WF_bgtz with P_malloc_next
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil 
    /\ getValinR s t0 <= (getValinR s actsize + 2) 
    /\ getValinR s actsize <= getValinR s t0 /\
    exists fv : addr, exists q : addr, exists n : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil))
  G_malloc_next
  (fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ 
    forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, 
	star (pto flist fv) (star 
	  (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q (getValinR s t0)) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil)) ->
      MCPOST oth s').
(** psi (f) = theta **)
constructor; [constructor| trivial].
(** t1 > 0 -> P => P' ***) Focus 2.
unfold P_malloc_next. intros. simplRf H; simplRf'. intuition.
(** t1 >0 -> P => G' => G **) Focus 2.
unfold G_malloc_next. intros. clear H. simplRf H; simplRf'. intuition.
(** t1 <= 0 -> P => P'' **) Focus 2.
intros. simplRf H; simplRf'. intuition.
(** t1 <= 0 -> P => G'' => G **) Focus 2.
intros. clear H. simplRf'; simplRf H1. intuition.
(** lw t0 p nxt **)
apply WF_lw with
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil /\
    exists fv : addr, exists q : addr, exists n : nat, exists siz : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q siz) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil) /\
      siz >= getValinR s actsize /\ getValinR s t0 = q)
  (fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ 
    forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, exists siz : nat, 
	star (pto flist fv) (star 
	  (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q siz) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil) /\
	siz >= getValinR s actsize /\ getValinR s t0 = q) ->
      MCPOST oth s').
(** p->next in dom **)
intros. simpl0p. simplRf H. splitland. splitex. splitland. repeat splitstar.
inrh. inlh. inrh. inlh. mblkindom. 
(** P S -> P' nextS *)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'.
intuition. splitex. exists x0; exists x1; exists x2; exists (lookupR rf t0).
splitland; getland. repeat (intros; rfinsen). simpl0p'.
replace x with (hread' h (lookupR rf p)); [idtac| unfold hread'; rewrite H7; trivial].
unfold hread'. repeat splitstar. hread2. readrh. hread2. readlh. hread2. readrh.
hread2. readlh. foldhread'. mblkchunk. fauto. mblkindom. inlh. mblkindom. fauto.
inrh. inlh. mblkindom. inlh. inrh. inlh. mblkindom.
(** P S -> G' nextS S' -> G S S' **)
unfold RESRqd. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H3 in H1. mc_unall H1. simplRf H; simplRf H1; simplRf'.
splitland; getland. clear H37. intros; apply H13; trivial. 
splitex. exists x0; exists x1; exists x2; exists (lookupR rf t0).
splitland; getland. repeat (intros; rfinsen). othhold H36 rf. simpl0p'.
replace x with (hread' h (lookupR rf p)); [idtac| unfold hread'; rewrite H8; trivial].
unfold hread'. repeat splitstar. hread2. readrh. hread2. readlh. hread2. readrh.
hread2. readlh. foldhread'. mblkchunk. fauto. mblkindom. inlh. mblkindom. fauto.
inrh. inlh. mblkindom. inlh. inrh. inlh. mblkindom.
(** bgtz prev L_malloc_lprv **)
apply WF_bgtz with P_malloc_lprv
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0 
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s p > pnil /\
    getValinR s prev = pnil /\
    exists fv : addr, exists q : addr, exists n : nat, exists siz : nat, 
      star (pto flist fv) (star 
	(star (EndL fv (getValinR s prev) (getValinR s p))(
	  star (MBlk (getValinR s p + 2) q siz) (MBlkLst n q pnil)))
	(star (Stk TOP) arbh)) s /\ (getValinR s p < q \/ q = pnil) /\
      siz >= getValinR s actsize /\ getValinR s t0 = q)
  G_malloc_lprv
  (fun s s' : state => 
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ 
    forall oth : assert, RfInsA oth ->
      (exists fv : addr, exists q : addr, exists n : nat, exists siz : nat, 
	star (pto flist fv) (star 
	  (star (EndL fv (getValinR s prev) (getValinR s p))(
	    star (MBlk (getValinR s p + 2) q siz) (MBlkLst n q pnil)))
	  (star (Stk TOP) oth)) s /\ (getValinR s p < q \/ q = pnil) /\
	siz >= getValinR s actsize /\ getValinR s t0 = q) ->
      MCPOST oth s').
(** psi (f) = theta **)
constructor; [idtac|omega]. constructor; [idtac|omega]. constructor.
(** prev<>0 -> P => P' ***) Focus 2.
unfold P_malloc_lprv. intros. intuition. 
(** prev > 0 -> P => G' => G **) Focus 2.
unfold G_malloc_lprv. intros. intuition.
(** prev = 0 -> P => P'' **) Focus 2.
intros. intuition.
(** prev = 0 -> P => G'' => G **) Focus 2.
intros. intuition.
(** sw zero flist t0 **)
apply WF_sw with P_malloc_retptr G_malloc_retptr.
(** flist in dom **)
intros. splitland. splitex. splitland. splitstar. simplz. simpl0p.
inlh. ptoindom. fauto.
(** P S ->  P' nextS **)
unfold P_malloc_retptr. intros. induction s as [h rf]. inversion H0.
simplRf H. simplRf'. intuition. splitex. exists x0. exists x2. exists x0.
splitland. getland. splitstar. writelh. splitstar. getstar'. clear H22.
splitstar. rewrite H12 in H22. fromendl0. assert (heap_eq (happend x7 x8) x8).
apply append_nullheap. apply heap_eq_sym; auto. revheq' H25.
assert (heap_eq x8 x5). heqtrans. apply h_subst_star with x8; auto.
splitstar. getstar'. unfold Good. exists x1; auto. simplz. simpl0p.
ptoindom. fauto. simplz. simpl0p. rewrite H17. ptopto. fauto. 
(** P S -> G' nextS S' -> G S S' **)
unfold G_malloc_retptr. unfold RESRqd. unfold FREE_FRAME. intros. 
induction s as [h rf]. induction s'' as [h'' rf''].  mc_unall H1.
inversion H0. simplz'. simpl0p'. rewrite <- H3 in H1. 
simplRf H; simplRf H1; simplRf'. splitland. splitex. splitland.
starrev H37. starassocl H40. splitstar. clear H42. starrev H44.
starassocr H42. splitstar. splitstar. getland. hintact'. notinstk.
ptoindom. fromnotinstk. inversion_clear H50. left; omega. right; trivial.
rewrite H14. ungetVinH. simplneq. apply beq_neq_false. 
apply disj_indom_neq with x7 x8; auto. ptoindom. stkindom.
rewrite H10. ungetVinH. simplneq. apply beq_neq_false. 
apply disj_indom_neq with x7 x8; auto. ptoindom. stkindom.
rewrite H17. ungetVinH. simplneq. apply beq_neq_false. 
apply disj_indom_neq with x7 x8; auto. ptoindom. stkindom.
rewrite H12. ungetVinH. simplneq. apply beq_neq_false. 
apply disj_indom_neq with x7 x8; auto. ptoindom. stkindom.
intros. apply H13; trivial. clalldisj. clallheq. clallha. clallh.
splitex. exists x4. exists x6. exists x4.
splitland. getland. splitstar. writelh. splitstar. getstar'. clear H50.
splitstar. rewrite H34 in H50. fromendl0. assert (heap_eq (happend x11 x12) x12).
apply append_nullheap. apply heap_eq_sym; auto. revheq' H53.
assert (heap_eq x12 x9). heqtrans. apply h_subst_star with x12; auto.
splitstar. getstar'. unfold Good. exists x5; auto. ptoindom. rewrite H44.
ptopto. fauto. 
(** j L_malloc_retptr **)
apply WF_j with P_malloc_retptr G_malloc_retptr.
repeat (constructor;[idtac|omega]). constructor.
(**P => P' **)
intros; trivial.
(** P => G' => G **)
intros; trivial.
Save.

(***************************************************************)

Theorem WF_malloc_iter: 
  Infer (itpx2 (itpx1 T_malloc_iter) psi_malloc_iter) C_malloc_iter.
Proof.
unfold C_malloc_iter. 
apply WF_bgtz with P_malloc_comp
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s a0 > 0
    /\ getValinR s actsize = getValinR s a0 + 2 /\ getValinR s retptr = pnil 
    /\ getValinR s p = pnil /\
    exists fv : addr, star (pto flist fv) (star 
      (MidL fv (getValinR s prev) (getValinR s p)) (star (Stk TOP) arbh)) s)
  G_malloc_comp G_malloc_comp.
(** psi (f) = theta **)
constructor.
(** p>0 -> P => P' **) Focus 2. 
unfold P_malloc_iter; unfold P_malloc_comp. intros. intuition.
(** p>0 -> P => G' => G **) Focus 2.
unfold G_malloc_comp; unfold G_malloc_iter. intros. intuition.
(** p=0 -> P => P' **) Focus 2.
unfold P_malloc_iter. intros. intuition. 
(** p=0 -> P => G'' => G **) Focus 2.
unfold G_malloc_comp; unfold G_malloc_iter. intros. intuition.
(** j L_malloc_epilog **)
apply WF_j with P_malloc_epilog G_malloc_epilog.
constructor. constructor. auto.
(** P => P' **)
unfold P_malloc_epilog. unfold MCPOST. intros. intuition. right.
splitex. exists x. intuition. induction s as [h rf]. splitstar.
getstar'. clear H6. splitstar. getstar'. simplRf H9.
apply MidL_Good with (lookupR rf prev) (lookupR rf p); auto.
(** P => G' => G **)
unfold G_malloc_epilog; unfold G_malloc_comp.
intros. intuition. clear H11. apply H9; trivial. unfold MCPOST.
right. splitex. exists x. intuition. induction s as [h rf].
splitstar. getstar'. clear H12. splitstar. getstar'. simplRf H15.
apply MidL_Good with (lookupR rf prev) (lookupR rf p); auto.
Save.

Check 0.

(***************************************************************)

Theorem WF_malloc_init :
  Infer (itpx2 (itpx1 T_malloc_init) psi_malloc_init) C_malloc_init.
Proof. 
unfold C_malloc_init. unfold P_malloc_init. unfold G_malloc_init.
(** addiu actsize reqsize header_size **)
apply WF_addiu with 
  (fun s : state =>
    MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s actsize > 2 /\ 
    getValinR s actsize = getValinR s reqsize + 2 /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s)
  (fun s s' : state =>
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
      MCPOST oth s').
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. 
simplRf H; simplRf'. intuition. splitex. exists x. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S' **)
unfold RESRqd. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. simplRf H1; simplRf'. splitland; getland.
intros; apply H12; auto. splitex. exists x. repeat (intros; rfinsen).
othhold H16 rf.
(**addiu prev zero 0 **)
apply WF_addiu with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s actsize > 2 /\ 
    getValinR s actsize = getValinR s reqsize + 2 /\ getValinR s prev = pnil /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s)
  (fun s s' : state =>
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
      MCPOST oth s').
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. 
simplRf H; simplRf'. intuition. simplz. fauto.
splitex. exists x. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S' **)
unfold RESRqd. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. simplRf H1; simplRf'. splitland; getland.
intros; apply H12; auto. splitex. exists x. repeat (intros; rfinsen).
othhold H17 rf.
(** lw p zero flist **)
apply WF_lw with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s actsize > 2 /\ 
    getValinR s actsize = getValinR s reqsize + 2 /\ getValinR s prev = pnil /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s
      /\ getValinR s p = fv)
  (fun s s' : state =>
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s
	/\ getValinR s p = fv) ->
      MCPOST oth s').
(** flist in dom **)
intros. simplRf H. splitland. simplz. simpl0p. splitex. repeat splitstar.
inlh. ptoindom. fauto. 
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. 
simplRf H; simplRf'. intuition. splitex. simplz'. simpl0p'. 
exists x0. split; repeat (intros; rfinsen). splitstar. unfold pto in H16.
inversion_clear H16. replace x with (hread' h flist). unfold getValinH in H17.
unfold hread'. hread2. readlh; fauto. rewrite H17; trivial. 
apply indom_some with x0. unfold hread in H17; trivial. unfold hread'; rewrite H7; trivial.
(** P S -> G' nextS S' -> G S S' **)
unfold RESRqd. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H3 in H1. mc_unall H1. simplRf H1; simplRf'. 
intuition. apply H23; trivial. clear H24. splitex. exists x0.
split; repeat (intros; rfinsen). othhold H34 rf. splitstar. unfold pto in H37.
inversion_clear H37. replace x with (hread' h flist). unfold getValinH in H38.
unfold hread'. hread2. readlh; fauto. rewrite H38; trivial. 
apply indom_some with x0. unfold hread in H38; trivial. 
simplz. simpl0p'. unfold hread'; rewrite H8; trivial. fauto.
(** addiu retptr zero 0**)
apply WF_addiu with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ getValinR s actsize > 2 /\ 
    getValinR s actsize = getValinR s reqsize + 2 /\ getValinR s prev = pnil /\
    getValinR s retptr = pnil /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s
      /\ getValinR s p = fv)
  (fun s s' : state =>
    (Gand ES0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    POPS0S1S2 s s' /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s
	/\ getValinR s p = fv) ->
      MCPOST oth s').
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0.
simplRf'; simplRf H. intuition. simplz; fauto. splitex. exists x.
intuition. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S' **)
unfold RESRqd. intros. mc_unall H1. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. simplRf H1; simplRf'. clear H. intuition.
apply H17; trivial. splitex. exists x. intuition. 
repeat (intros; rfinsen). othhold H27 rf.
(** j L_malloc_iter **)
apply WF_j with P_malloc_iter G_malloc_iter.
(** psi (f) = theta **)
constructor.
(** P => P' **)
unfold P_malloc_iter. intros. intuition. splitex. inversion_clear H5.
rewrite H3. exists x. induction s as [h rf]. splitstar. getstar'.
clear H5. splitstar. getstar'. rewrite H7. apply Good_MidL; auto.
(** P => G' => G **)
unfold G_malloc_iter. intros. intuition. clear H11. apply H9; auto.
splitex. exists x. inversion_clear H11. rewrite H7. 
induction s as [h rf]. splitstar. getstar'. clear H11. splitstar.
getstar'. rewrite H13. apply Good_MidL; auto.
Save.

Check 0.

(***************************************************************)

Theorem WF_malloc_epilog :
  Infer (itpx2 (itpx1 T_malloc_epilog) emptyCT) C_malloc_epilog.
Proof.
unfold C_malloc_epilog.
(** lw_ p fp 3 **)
apply WF_lw_ with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ MCPOST arbh s)
  (fun s s' : state => (Gand ES0S1 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    getValinR s' s1 = getVinH s (getValinR s fp - 2) /\ getValinR s' s0 = 
    getVinH s (getValinR s fp - 1) /\ RESRqd s s' /\ forall oth : assert, RfInsA oth ->
      MCPOST oth s -> MCPOST oth s').
(**fp-3 in dom**)
unfold P_malloc_epilog. unfold MCPOST. unfold FREE_FRAME. intros.
simplRf H. splitland. inversion_clear H2. splitex. inversion_clear H2. repeat splitstar.
inrh. inrh. inrh. inlh. stkindom. splitex. inversion_clear H2.  repeat splitstar.
inrh. inrh. inlh. stkindom.
(** P S -> P' nextS **)
unfold P_malloc_epilog. unfold FREE_FRAME. unfold MCPOST. 
intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'.
intuition. left. splitex. exists x0; exists x1; exists x2. intuition. 
repeat (intros; rfinsen). right. splitex. exists x0.  intuition. 
repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S' **)
unfold G_malloc_epilog. unfold RESRqd. unfold MCPOST. intros. induction s as [h rf].
induction s'' as [h'' rf'']. unfold ES0S1 in H1. mc_unall H1.
inversion H0. rewrite <- H3 in H1. simplRf H1; simplRf'. splitland.
getland. rewrite <- H24. ungetVinH. foldhread. rewrite H8; trivial.
 intros; apply H14; auto. inversion_clear H31. left. splitex. exists x0; exists x1; exists x2.
intuition. repeat (intros; rfinsen). othhold H29 rf. right. splitex. exists x0.  
intuition. repeat (intros; rfinsen). othhold H29 rf.
(** lw_ prev fp 2 **)
apply WF_lw_ with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ MCPOST arbh s)
  (fun s s' : state => (Gand ES0 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    getValinR s' s0 = getVinH s (getValinR s fp - 1) /\ RESRqd s s' /\
    forall oth : assert, RfInsA oth -> MCPOST oth s -> MCPOST oth s').
(** fp-2 in dom **)
unfold FREE_FRAME. unfold MCPOST. intros. simplRf H. splitland. inversion_clear H2.
splitex. inversion_clear H2. repeat splitstar. inrh. inrh. inrh. inlh. stkindom. splitex. 
inversion_clear H2. repeat splitstar. inrh. inrh. inlh. stkindom.
(** P S -> P' nextS **)
unfold FREE_FRAME. unfold MCPOST. intros. induction s as [h rf].
inversion H0. rewrite <- H2 in H0. simplRf H; simplRf'. intuition. left.
splitex. exists x0; exists x1; exists x2. intuition. repeat (intros; rfinsen). right.
splitex. exists x0. intuition. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S'**)
unfold FREE_FRAME. unfold RESRqd. unfold MCPOST; unfold ES0S1; unfold ES0;  intros. 
induction s as [h rf]. induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
mc_unall H1. simplRf'; simplRf H1. splitland. getland. rewrite <- H23. ungetVinH. foldhread.
rewrite H8; trivial. intros; apply H13; auto. inversion_clear H35. left. 
splitex. exists x0; exists x1; exists x2. splitland; getland. 
repeat (intros; rfinsen). othhold H33 rf. right. splitex. exists x0. 
splitland; getland. repeat (intros; rfinsen). othhold H33 rf.
(** lw_ actsize fp 1 **)
apply WF_lw_ with 
  (fun s : state => MALLOC_FRAME s /\ getValinR s ra <> 0 /\ MCPOST arbh s)
  (fun s s' : state => (Gand S0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s fp) s s' /\
    RESRqd s s' /\
    forall oth : assert, RfInsA oth -> MCPOST oth s -> MCPOST oth s').
(** fp-2 in dom **)
unfold FREE_FRAME. unfold MCPOST. intros. simplRf H. splitland. inversion_clear H2.
splitex.  inversion_clear H2. repeat splitstar. inrh. inrh. inrh. inlh. stkindom. splitex. 
inversion_clear H2. repeat splitstar. inrh. inrh. inlh. stkindom.
(** P S -> P' nextS **)
unfold FREE_FRAME. unfold MCPOST. intros. induction s as [h rf].
inversion H0. rewrite <- H2 in H0. simplRf H; simplRf'. intuition. left. 
splitex. exists x0; exists x1; exists x2. splitland; getland. 
repeat (intros; rfinsen). right. splitex. exists x0. intuition. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S'**)
unfold FREE_FRAME. unfold RESRqd. unfold MCPOST; unfold ES0;  intros. 
induction s as [h rf]. induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
mc_unall H1. simplRf'; simplRf H1. splitland. getland. rewrite <- H22. ungetVinH. foldhread.
rewrite H8; trivial. intros; apply H12; auto. inversion_clear H35. left. 
splitex. exists x0; exists x1; exists x2. splitland; getland.
repeat (intros; rfinsen). othhold H33 rf. right. splitex. exists x0. 
splitland; getland. repeat (intros; rfinsen). othhold H33 rf.
(** lw fp fp 0 **)
apply WF_lw with 
  (fun s : state => getValinR s ra <> 0 /\ MCPOST arbh s)
  (fun s s' : state => (Gand S0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s sp + 4) s s' /\
    getValinR s' fp = getValinR s fp /\ getValinR s' sp = getValinR s sp + 4 /\
    getValinR s' ra = getValinR s ra /\
    forall oth : assert, RfInsA oth -> MCPOST oth s -> MCPOST oth s').
(** fp in dom **)
unfold FREE_FRAME. unfold MCPOST. intros. simplRf H. splitland. inversion_clear H2.
splitex. inversion_clear H2. repeat splitstar. inrh. inrh. inrh. inlh. stkindom. splitex. 
inversion_clear H2. repeat splitstar. inrh. inrh. inlh. stkindom.
(** P S -> P' nextS **)
unfold FREE_FRAME. unfold MCPOST. intros. induction s as [h rf].
inversion H0. simplRf H; simplRf'. intuition. left.
splitex. exists x0; exists x1; exists x2. splitland; getland.  repeat (intros; rfinsen). right.
splitex. exists x0. intuition. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S'**)
unfold FREE_FRAME. unfold RESRqd. unfold MCPOST;  intros. 
induction s as [h rf]. induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
mc_unall H1. simplRf'; simplRf H1; simplRf H. splitland. getland.
rewrite <- H31; trivial. rewrite H10. ungetVinH. foldhread. simpl0p'.
rewrite H8; trivial. intros; apply H14; auto. inversion_clear H35. left. 
splitex. exists x0; exists x1; exists x2. splitland; getland. repeat (intros; rfinsen). 
othhold H33 rf. right. splitex. exists x0. splitland; getland. 
repeat (intros; rfinsen). othhold H33 rf.
apply WF_addiu with 
  (fun s : state => getValinR s ra <> 0 /\ MCPOST arbh s)
  (fun s s' : state => (Gand S0S1S2 ET0T1) s s' /\ HIntact TOP (getValinR s sp) s s'
    /\ Rqd s s' /\ forall oth : assert, RfInsA oth -> MCPOST oth s -> MCPOST oth s').
(** P S -> P' nextS **)
unfold MCPOST. intros. induction s as [h rf]. inversion H0.  simplRf'; simplRf H. 
splitland; getland. inversion_clear H8. left. splitex. exists x; exists x0; exists x1.
splitland; getland. repeat (intros; rfinsen). right. splitex.
exists x. intuition. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S'**)
unfold Rqd.  unfold MCPOST. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. mc_unall H1. simplRf'; simplRf H1. splitland; getland.
intros; apply H11; auto. inversion_clear H31. left; splitex. exists x; exists x0; exists x1.
splitland; getland. repeat (intros; rfinsen). othhold H rf. right; 
splitex. exists x. clear H30.  intuition. repeat (intros; rfinsen). othhold H rf.
(** jr ra **)
apply WF_jr_ra'. 
intros. unfold Rqd. mc_unall H.  intuition. unfold HIntact. intros; trivial. 
Save.


(****************************************************)

Theorem WF_malloc_prolog :
  Infer (itpx2 (itpx1 T_malloc_prolog) psi_malloc_prolog) C_malloc_prolog.
Proof.
unfold C_malloc_prolog. 
(** subiu sp sp 4 **)
apply WF_subiu with 
  (fun s : state => getValinR s ra <> 0 /\ getValinR s fp < TOP /\
    getValinR s sp + 4 < getValinR s fp /\ getValinR s sp > 1 /\ getValinR s a0 > 0 /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s)
  (fun s s' : state => (Gand S0S1S2 ET0T1) s s' /\ getValinR s' fp = getValinR s fp /\
    getValinR s' sp = getValinR s sp + 4 /\ getValinR s' ra = getValinR s ra /\
    HIntact TOP (getValinR s sp + 4) s s' /\ forall oth : assert, RfInsA oth -> 
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
      MCPOST oth s').
(** P S -> P' nextS **)
unfold P_malloc_prolog. intros. induction s as [h rf]. inversion H0. 
simplRf H; simplRf'. intuition. splitex. exists x. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S' **)
unfold P_malloc_prolog. unfold G_malloc_prolog. unfold MCPOST. intros. 
induction s as [h rf]. induction s'' as [h'' rf'']. inversion H0.  
rewrite <- H2 in H1. unfold Rqd. mc_unall H1. simplRf'; simplRf H1; simplRf H. 
intuition. replace (lookupR rf sp - 4 + 4) with (lookupR rf sp) in H20.  
apply HIntactInsen with (consR rf sp (lookupR rf sp - 4)); auto.
apply plusex; omega. apply H25; auto. clear H22. splitex. exists x. 
repeat (intros; rfinsen). othhold H33 rf.
(** sw sp 4 fp **)
apply WF_sw with 
  (fun s : state => getValinR s ra <> 0 /\ getValinR s sp + 4 < TOP /\ 
    getValinR s sp > 1 /\ getValinR s a0 > 0 /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s)
  (fun s s' : state => (Gand S0S1S2 ET0T1) s s' /\ 
    getValinR s' fp = getVinH s (getValinR s sp + 4) /\
    getValinR s' sp = getValinR s sp + 4 /\ getValinR s' ra = getValinR s ra /\
    HIntact TOP (getValinR s sp + 4) s s' /\ (forall oth : assert, RfInsA oth -> 
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
      MCPOST oth s')).
(** sp+4 in dom **)
intros.  simplRf H. splitland. splitex. repeat splitstar. inrh. inrh. inlh. stkindom.
(** P S -> P' nextS **)
intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'. 
intuition. splitex. exists x. splitstar. writerh. repeat splitstar. 
inrh. inlh. stkindom. splitstar. writerh. repeat splitstar. inlh. stkindom.
splitstar. writelh. stkindom. apply StkRet;auto. stkindom. 
(** P S -> G' nextS S' -> G S S'**)
unfold MCPOST. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H3 in H1. mc_unall H1. simplRf H; simplRf H1; simplRf'.
intuition. rewrite H. ungetVinH. simpleq. apply beq_eq_true; auto. hintact'.
apply H26; auto. clear H23; splitex. exists x. splitstar. writerh. repeat splitstar. 
inrh. inlh. stkindom. splitstar. writerh. repeat splitstar. inlh. stkindom.
splitstar. writelh. stkindom. apply StkRet;auto. stkindom. 
(** addiu fp sp 4 **)
apply WF_addiu with 
  (fun s : state => getValinR s ra <> 0 /\ MALLOC_FRAME s /\ getValinR s a0 > 0 /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s)
  (fun s s' : state => (Gand S0S1S2 ET0T1) s s' /\ RESRqd s s' /\
    HIntact TOP (getValinR s sp + 4) s s' /\ (forall oth : assert, RfInsA oth -> 
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
      MCPOST oth s')).
(* P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. 
simplRf H; simplRf'. intuition. splitex. exists x. repeat (intros; rfinsen).
(** P S -> G' nextS S' -> G S S' **)
unfold RESRqd. unfold MCPOST. intros. induction s as [h rf]. induction s'' as [h'' rf''].
inversion H0. rewrite <- H2 in H1. mc_unall H1. simplRf H1; simplRf H; simplRf'.
intuition. clear H19. apply H18; auto. splitex. exists x. repeat (intros; rfinsen).
othhold H32 rf.
(** sw_ fp 1 s0 ***)
apply WF_sw_ with
  (fun s : state => getValinR s ra <> 0 /\ MALLOC_FRAME s /\ getValinR s a0 > 0 /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s)
  (fun s s' : state => (Gand ES0 ET0T1) s s' /\ RESRqd s s' /\
    getValinR s' actsize = getVinH s (getValinR s fp - 1) /\
    HIntact TOP (getValinR s sp + 4) s s' /\ (forall oth : assert, RfInsA oth -> 
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
      MCPOST oth s')).
(** fp-1 in dom **)
unfold FREE_FRAME. intros. simplRf H. splitland. splitex. repeat splitstar.
inrh. inrh. inlh. stkindom. 
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'.
intuition. splitex. exists x. splitstar. writerh. repeat splitstar. inrh. inlh. 
stkindom. splitstar. writerh. splitstar. inlh. stkindom. splitstar. writelh. stkindom.
apply StkRet; auto. stkindom. 
(** P S -> G' nextS -> G S S' **)
unfold FREE_FRAME. unfold RESRqd. unfold ES0. intros. induction s as [h rf].
induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
mc_unall H1. simplRf H1; simplRf H; simplRf'. intuition. rewrite H.
ungetVinH. simpleq. apply beq_eq_true; trivial. rewrite H13. ungetVinH.
simplneq. apply beq_neq_false; omega. hintact'. left; omega. apply H25; trivial.
clear H16. splitex. exists x. splitstar. writerh. repeat splitstar. inrh. inlh. 
stkindom. splitstar. writerh. splitstar. inlh. stkindom. splitstar. writelh. stkindom.
apply StkRet; auto. stkindom.
(** sw_ fp 2 prev **)
apply WF_sw_ with
  (fun s : state => getValinR s ra <> 0 /\ MALLOC_FRAME s /\ getValinR s a0 > 0 /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s)
  (fun s s' : state => (Gand ES0S1 ET0T1) s s' /\ RESRqd s s' /\
    getValinR s' actsize = getVinH s (getValinR s fp - 1) /\
    getValinR s' prev = getVinH s (getValinR s fp - 2) /\
    HIntact TOP (getValinR s sp + 4) s s' /\ (forall oth : assert, RfInsA oth -> 
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
      MCPOST oth s')).
(** fp-1 in dom **)
unfold FREE_FRAME. intros. simplRf H. splitland. splitex. repeat splitstar.
inrh. inrh. inlh. stkindom. 
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'.
intuition. splitex. exists x. splitstar. writerh. repeat splitstar. inrh. inlh. 
stkindom. splitstar. writerh. splitstar. inlh. stkindom. splitstar. writelh. stkindom.
apply StkRet; auto. stkindom. 
(** P S -> G' nextS -> G S S' **)
unfold FREE_FRAME. unfold RESRqd. unfold ES0S1. unfold ES0. intros. 
induction s as [h rf]. induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
mc_unall H1. simplRf H1; simplRf H; simplRf'. intuition. rewrite H19.
ungetVinH. simpleq. apply beq_eq_true; trivial. rewrite H13. ungetVinH.
simplneq. apply beq_neq_false; omega. rewrite H. ungetVinH. simplneq. 
apply beq_neq_false; omega. hintact'. left; omega. apply H28; trivial.
clear H16. splitex. exists x. splitstar. writerh. repeat splitstar. inrh. inlh. 
stkindom. splitstar. writerh. splitstar. inlh. stkindom. splitstar. writelh. stkindom.
apply StkRet; auto. stkindom.
(** sw_ fp 3 p **)
apply WF_sw_ with
  (fun s : state => getValinR s ra <> 0 /\ MALLOC_FRAME s /\ getValinR s a0 > 0 /\
    exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) arbh)) s)
  (fun s s' : state => (Gand ES0S1S2 ET0T1) s s' /\ RESRqd s s' /\ POPS0S1S2 s s' /\
    HIntact TOP (getValinR s fp) s s' /\ (forall oth : assert, RfInsA oth -> 
      (exists fv : addr, star (pto flist fv) (star (Good fv) (star (Stk TOP) oth)) s) ->
      MCPOST oth s')).
(** fp-1 in dom **)
unfold FREE_FRAME. intros. simplRf H. splitland. splitex. repeat splitstar.
inrh. inrh. inlh. stkindom. 
(** P S -> P' nextS **)
unfold FREE_FRAME. intros. induction s as [h rf]. inversion H0. simplRf H; simplRf'.
intuition. splitex. exists x. splitstar. writerh. repeat splitstar. inrh. inlh. 
stkindom. splitstar. writerh. splitstar. inlh. stkindom. splitstar. writelh. stkindom.
apply StkRet; auto. stkindom. 
(** P S -> G' nextS -> G S S' **)
unfold FREE_FRAME. unfold RESRqd. unfold ES0S1. unfold ES0S1S2. intros. 
induction s as [h rf]. induction s'' as [h'' rf'']. inversion H0. rewrite <- H3 in H1.
mc_unall H1. simplRf H1; simplRf H; simplRf'. intuition. rewrite H29.
ungetVinH. simpleq. apply beq_eq_true; trivial. rewrite H13. ungetVinH.
simplneq. apply beq_neq_false; omega. rewrite H19. ungetVinH. simplneq. 
apply beq_neq_false; omega. rewrite H23. ungetVinH. simplneq. 
apply beq_neq_false; omega. rewrite H11. hintact'. left; omega. apply H26; trivial.
clear H16. splitex. exists x. splitstar. writerh. repeat splitstar. inrh. inlh. 
stkindom. splitstar. writerh. splitstar. inlh. stkindom. splitstar. writelh. stkindom.
apply StkRet; auto. stkindom.
(** j L_malloc_init **)
apply WF_j with P_malloc_init G_malloc_init.
(** psi (f) = theta **)
constructor.
(** P => P' **)
unfold P_malloc_init. intros. intuition.
(** P => G' => G **)
unfold G_malloc_init. intros; intuition.
Save.
