
(** ** Utilities for manipulating & reasoning natural numbers *)
(** By Nadeem A. Hamid *)

(*Require Import Div2.*)
Require Import EqNat.
Require Import Lt.
Require Import Le.
Require Import Omega.

(*blt_nat and related properties*)
Fixpoint blt_nat (n m:nat){struct n}:bool:=
  match n,m with
    |O,O=>false
      |O,S _=>true
	|S _,O=>false
	  |S n1,S n2=>blt_nat n1 n2
  end.

Lemma blt_S_false : forall a:nat, blt_nat (S a) a = false.
Proof.
simple induction a.
unfold blt_nat.
auto.
intros.
unfold blt_nat.
auto.
Qed.

Lemma blt_S_eq:
  forall (a b:nat)(t:bool),blt_nat a b = t -> blt_nat (S a) (S b) = t.
Proof.
intros.
unfold blt_nat in |- *.
fold blt_nat in |- *.
auto.
Qed.

Lemma blt_lt_true: forall a b:nat, a<b -> blt_nat a b = true.
Proof.
simple induction a.
intros.
inversion H.
unfold blt_nat in |- *.
auto.
unfold blt_nat in |- *.
auto.
intros.
inversion H0.
apply blt_S_eq.
apply H.
apply lt_n_Sn.
rewrite <- H2 in H0.
unfold blt_nat.
fold blt_nat.
apply H.
apply lt_S_n.
auto.
Qed.

Lemma blt_Sm_false :
  forall a b:nat, blt_nat a b = false -> blt_nat (S a) b = false.
Proof.
double induction a b.
intros.
unfold blt_nat;auto.
intros.
unfold blt_nat in H0.
absurd (true=false).
discriminate.
auto.
intros.
unfold blt_nat.
auto.
intros.
unfold blt_nat in H1.
fold blt_nat in H1.
cut (blt_nat (S n0) n = false);intros.
apply blt_S_eq.
auto.
apply H0.
auto.
Qed.

Lemma blt_false : forall a b : nat, a < b -> blt_nat b a = false.
Proof.
intros.
induction H as [| m H HrecH].
apply blt_S_false.
apply blt_Sm_false.
auto.
Qed.

Lemma blt_eq_false : forall a:nat, blt_nat a a = false.
Proof.
simple induction a.
unfold blt_nat;auto.
intros.
unfold blt_nat.
auto.
Qed.

Lemma le_blt_false : forall a b : nat, a <= b -> blt_nat b a = false.
Proof.
intros.
induction H as [| m H HrecH].
apply blt_eq_false.
apply blt_false.
unfold lt.
apply le_n_S.
auto.
Qed.

(** Properties of natural numbers *)
Lemma beq_neq_false : forall a b : nat, a <> b -> false = beq_nat a b.
Proof.
double induction a b.
intros.
absurd (0<>0);auto.
intros.
unfold beq_nat.
auto.
intros.
unfold beq_nat.
auto.
intros.
unfold beq_nat.
fold beq_nat.
apply H0.
auto.
Qed.

Lemma beq_eq_true : forall a b : nat, a = b -> true = beq_nat a b.
Proof.
double induction a b.
intros.
unfold beq_nat; auto.
intros.
absurd (0 = S n); auto.
intros.
absurd (S n = 0); auto.
intros.
unfold beq_nat.
fold beq_nat.
apply H0.
auto.
Save.

Lemma beq_true_eq : forall a b : nat, true = beq_nat a b -> a = b.
Proof.
double induction a b;simpl in |-.
reflexivity .
intros; discriminate H0.
intros; discriminate H0.
intros; case (H0 _ H1); reflexivity.
Save.

Lemma eq_neq_Sl : forall a b : nat, a = b -> S a <> b.
Proof.
double induction a b.
intros.
auto.
intros.
absurd (0 = S n).
discriminate.
auto.
intros.
discriminate.
intros.
apply not_eq_S.
auto.
Qed.

Theorem eqorneq_nat : forall a a' : nat, a = a' \/ a <> a'.
Proof.
intros.
compare a a'.
auto.
auto.
Qed.


Theorem le_or_gt : forall m n : nat, n <= m \/ n > m.
Proof.
double induction m n.
left;auto.
intros.
right.
red.
red.
apply le_n_S.
apply le_O_n.
intros.
left.
apply le_O_n.
intros.
cut (n <= n1 \/ n > n1).
intros.
inversion_clear H1.
left.
apply le_n_S.
auto.
right.
unfold gt.
unfold gt in H2.
apply lt_n_S.
auto.
apply H0.
Qed.

Theorem gt_or_le : forall m n : nat, n > m \/ n <= m.
Proof.
intros.
cut (n <= m \/ n > m);intros.
inversion_clear H.
right;auto.
left;auto.
apply le_or_gt.
Qed.

Lemma lt_neq : forall m n : nat, n < m -> n <> m.
Proof.
simple induction m.
intros.
inversion H.
intros.
induction n0 as [| n0 Hrecn0].
auto.
cut (n0 < n);intros.
cut (n0 <> n);intros.
auto.
apply H;auto.
apply lt_S_n.
auto.
Qed.
Lemma lt_S_neq : forall n x : nat, n < S x -> x <> n -> n < x.
Proof.
double induction n x.
intros.
absurd (0 <> 0).
auto.
auto.
intros.
apply lt_O_Sn.
intros.
unfold lt in H0.
cut (S n0 <= 0).
intros.
absurd (S n0 <= 0).
apply le_Sn_O.
auto.
apply le_S_n.
auto.
intros.
apply lt_n_S.
cut (n1 = n0 \/ n1 <> n0).
intros.
inversion_clear H3.
absurd (S n1 <> S n1).
auto.
rewrite <- H4 in H2.
auto.
cut (n1 < S n0).
intros.
apply (H0 n0 H3).
auto.
apply lt_S_n.
auto.
apply eqorneq_nat.
Qed.

Lemma lt_S_eq_or_lt : forall n x : nat, n < S x -> x = n \/ n < x.
Proof.
intros.
cut (x = n \/ x <> n).
intros.
inversion_clear H0.
left; auto.
right; apply lt_S_neq.
auto.
auto.
apply eqorneq_nat.
Qed.

Lemma plus_n_O_n : forall n : nat, n + 0 = n.
Proof.
simple induction n.
unfold plus in |- *.
auto.
intros.
unfold plus in |- *.
fold plus in |- *.
rewrite H.
auto.
Qed.

Lemma plus_S_neq : forall n m : nat, n <> n + S m.
Proof.
simple induction n.
intros.
unfold plus in |- *.
auto.
intros.
unfold plus in |- *.
fold plus in |- *.
auto.
Qed.

Lemma gt_S_nat : forall (m n : nat), S m > S n -> m > n.
Proof.
intros.
omega.
Qed.

Lemma gt_n_O : forall (m n : nat), m > n -> m > 0.
Proof.
intros.
omega.
Qed.

Lemma gt_neq : forall (m n : nat), m > n -> m <> n.
Proof.
intros.
omega.
Qed.

Lemma plusex : forall (m n : nat), m > n -> m = m - n + n.
Proof.
double induction m n .
simpl; reflexivity .
intros; omega.
auto.
intros; omega.
Qed.

Lemma subneq : forall (m : nat), m > 0 -> m - 1 <> m.
Proof.
intros.
omega .
Qed.