Require Import heap2.
Require Import scap.
Require Import scap_rules.
Require Import cdsa.
Require Import Omega.


Axiom zero_r : forall rf : rfile, 0 = lookupR rf zero.

Hint Resolve zero_r : fsl.
Hint Resolve disj_sym : fsl.
Hint Resolve beq_neq_false : fsl.

(** fsl hints database and user define tactics ***)
Ltac fauto := try (auto with fsl).
Ltac splitland := 
  repeat match goal with 
	   | H : ?X1 /\ ?X2 |- ?X3 => inversion_clear H
	 end.
Ltac splitex :=
  repeat match goal with
	   | H : ex _ |- _ => inversion_clear H
	 end.

Ltac getstar h1 h2 :=
  match goal with
    | [ |- (star ?X1 ?X2) (?X3, ?H4) ] => unfold star at 1; exists h1; exists h2
  end.

Ltac getland :=
  repeat match goal with
	   | |- ?X1 /\ ?X2 => split; auto
	 end.

Ltac getstar' :=
  match goal with
    | [ H : hdisj ?h1 ?h2, H1 : heap_eq (happend ?h1 ?h2) ?h
      |- star ?P1 ?P2 (?h, _) ] => 
      unfold star at 1; exists h1; exists h2; getland
  end.

Ltac getstar1 :=
  repeat match goal with
	    | |- star1 _ _ _ _ => unfold star1; [getland]
	  end.

Ltac splitstar1 :=
  match goal with
    | H : star1 _ _ _ _ |- _ => unfold star1 in H; splitland
  end.

Ltac splitstar :=
  match goal with
    | H : star _ _ _ |- _ => unfold star at 1 in H; splitex; splitland
  end.
Ltac splitstarH H :=
  match goal with
    | H : star _ _ _ |- _ => unfold star at 1 in H; splitex; splitland
  end.

Ltac unGand H := try (unfold Gand; unfold Gand in H).
Ltac unRIntact H := try (unfold RIntact; unfold RIntact in H).
Ltac unHIntact H := try (unfold HIntact; unfold HIntact in H).
Ltac getvalinr H := try (unfold getValinR; unfold getValinR in H).
Ltac ungetValinH H := try (unfold getValinH; unfold getValinH in H).

Ltac ungetVinH := try (unfold getVinH;[unfold getValinH;[unfold hread;[unfold hwrite]]]).


Ltac tind h := 
  match goal with
    | |- hindom ?X1 ?X2 => assert (hindom X1 h)
  end.
(** one constructor for star *)
Ltac starc h1 h2 := try (splitstar; [getstar h1 h2; getland]).

Ltac vinreq := try (unfold updateR; [unfold lookupR;
  unfold beq_reg; fold lookupR; auto]).

(** clear all h : heap in hypo **)
Ltac clallha :=
  repeat match goal with
	   | [H : MBlk _ _ _ _ |- _] => clear H
	   | [H : Stk _ _ |- _] => clear H
	   | [H : arbh  _ |- _] => clear H
	   | [H : MidL _ _ _ _ |- _] => clear H
	   | [H : pto _ _ _ |- _] => clear H
	 end.

Ltac clalldisj :=
  repeat match goal with
	   | [H : hdisj _ _ |- _] => clear H
	 end.

Ltac clallheq :=
  repeat match goal with
	   | [H : heap_eq _ _ |- _] => clear H
	 end.

Ltac clallh :=
  repeat match goal with
	   | [h : heap |- _] => clear h
	 end.

Ltac simplRf H := try (unfold getValinR in H; [unfold updateR in H; unfold lookupR in H;
  unfold beq_reg in H; fold lookupR in H]).

Ltac simplRf' := try (unfold getValinR; [unfold updateR; unfold lookupR;
  unfold beq_reg; fold lookupR]).

Ltac simplHp := try (unfold getVinH; unfold hwrite; unfold getValinH; unfold hread).
Ltac simpleq := 
  match goal with
    | [ |- context [(EqNat.beq_nat ?X1 ?X1)] ] => 
      replace (EqNat.beq_nat X1 X1) with true; auto
  end.
Ltac simplneq :=
  match goal with
    | [ |- context [(EqNat.beq_nat ?X1 ?X2)] ] =>
      replace (EqNat.beq_nat X1 X2) with false; auto
  end.
(** simplify plus and add **)
Ltac simplps :=
  match goal with
    | [ |- context [(?X -?N + ?N)] ] =>
      replace (X - N + N) with X; auto
  end.

Ltac simplS :=
  match goal with
    | [ |- context [S ?i]] =>
      replace (S i) with (i + 1); [auto | auto]
    | [ H : context [S ?i] |- _] =>
      replace (S i) with (i + 1) in H; [auto | auto]
  end.
(** simplify zero **)
Ltac simplz :=
  match goal with
    | [ |- context [lookupR ?rf zero] ] =>
      try (replace (lookupR rf zero) with 0; auto)
    | [ H : context [lookupR ?rf zero] |- _ ] =>
      try (replace (lookupR rf zero) with 0 in H; auto)
  end.

Ltac simplz' :=
  match goal with
    | [ H : context [lookupR ?rf zero] |- _] =>
      replace (lookupR rf zero) with 0 in H; [idtac | fauto]
  end.

Ltac simpl0p :=
  match goal with
    | [ |- context [0 + ?X] ] =>
      replace (0 + X) with X; [idtac | auto]
    | [ |- context [?X + 0] ] =>
      replace (X + 0) with X; [idtac | auto]
  end.

Ltac simpl0p' :=
  match goal with
    | [ H : context [0 + ?X] |- _] =>
      replace (0 + X) with X in H; [idtac | auto]
    | [ H : context [?X + 0] |- _] =>
      replace (X + 0) with X in H; [idtac | auto]
  end.

Ltac hread2 :=
  match goal with
    | [ H : heap_eq (happend ?h1 ?h2) ?h |- context [hread ?h ?x] ] =>
      replace (hread h x) with (hread (happend h1 h2) x); 
	[idtac | apply heap_eq_veq; auto]
    | [ H : heap_eq (happend ?h1 ?h2) ?h, H1 : context [hread ?h ?x] |- _ ] =>
      replace (hread h x) with (hread (happend h1 h2) x) in H1; 
	[idtac | apply heap_eq_veq; auto]
  end.

Ltac getnotin h :=
  match goal with
    | [|- hnotindom _ _] => apply disj_inr_notinl with h; auto
  end.

Ltac readrh :=
  match goal with
    | [ H : hdisj ?h1 ?h2 |- context [hread (happend ?h1 ?h2) ?x] ] =>
      replace (hread (happend h1 h2) x) with (hread h2 x);
  	[idtac | symmetry; apply append_preserve_read_R; getnotin h2]
  end.
Ltac readlh :=
  match goal with
    | [ H : hdisj ?h1 ?h2 |- context [hread (happend ?h1 ?h2) ?x] ] =>
      replace (hread (happend h1 h2) x) with (hread h1 x);
  	[idtac | symmetry; apply append_preserve_read_L; getnotin h1]
    | [ H : hdisj ?h1 ?h2, H1 : context [hread (happend ?h1 ?h2) ?x] |- _ ] =>
      replace (hread (happend h1 h2) x) with (hread h1 x) in H1;
  	[idtac | symmetry; apply append_preserve_read_L; getnotin h1]
  end.

Ltac notinstk :=
  match goal with
    | [ H : Stk ?top (?h, _), Hdisj : hdisj ?hl ?h |- context [?i > ?top]] => 
      assert (hnotindom i h); [apply disj_inl_notinr with hl; auto|idtac]
  end.

Ltac foldhread' :=
  match goal with
    | [ |- context [match hread ?x ?i with | Some n => n | None => 0 end] ] =>
      replace (match hread x i with | Some n => n | None => 0 end) with (hread' x i);
	[idtac | unfold hread'; auto]
    | [ H : context [match hread ?x ?i with | Some n => n | None => 0 end] |- _] =>
      replace (match hread x i with | Some n => n | None => 0 end) with (hread' x i) in H;
	[idtac | unfold hread'; auto]
  end.

Ltac foldhread :=
  match goal with
    | [ |- context [?h ?i] ] => replace (h i) with (hread h i); 
      [idtac | unfold hread; auto]
  end.

Ltac readsize :=
  match goal with
    | [ H : context [hread' ?h (?hptr + 1)], 
      H1 : context [MBlk (?hptr + 2) ?lnk ?siz] |- _] =>
    replace (hread' h (hptr + 1)) with siz in H
    | [H1 : context [MBlk (?hptr + 2) ?lnk ?siz] |- context [hread' ?h (?hptr + 1)] ] =>
    replace (hread' h (hptr + 1)) with siz
  end.

Hint Resolve plusex : fsl.

Ltac veq H1 H2 := try (rewrite <- H1; [inversion H2; vinreq]).
Ltac veq' H1 H2 := try (rewrite H1; [inversion H2; vinreq]).

(**reverse the star assertions **)
Ltac revstar H := 
  match goal with
    | H : star ?X1 ?X2 ?X3 |- _ => assert (star X2 X1 X3)
  end.
Ltac assocstar H := 
  match goal with
    | H : star (star ?X1 ?X2) ?X3 ?X4 |- _ => assert (star X1 (star X2 X3) X4)
  end.

Ltac starfrom H :=
  match goal with 
    | H : star ?P1 ?P2 (?H1, ?RF) |- star ?P3 ?P4 (?H2, ?RF) =>
      assert (star P1 P2 (H2, RF))
  end.

Ltac assocstarR H := 
  match goal with
    | H : star ?X1 (star ?X2 ?X3) ?X4 |- _ => assert (star (star X1 X2) X3 X4)
  end.

Ltac revheq :=
  match goal with
    | [H : heap_eq ?h1 ?h2 |- _] => assert (heap_eq h2 h1);
      [apply heap_eq_sym; trivial | clear H]
  end.

Ltac heqindom :=
  match goal with
    | [H : heap_eq (?h1 ?h2) |- hindom ?i ?h2] => assert (hindom i h1);
      [idtac | apply heap_eq_preserve_indom with h1; auto]
  end.


  
(*** functions and some predicates general properties ***)
Lemma lt_trans : forall (m n l : nat), m < n -> n < l -> m < l.
Proof.
intuition.
Save.


(*** register insensitive functions **)
(*** register insensitive assertions and predicates **)
(*Lemma *)    
Lemma ptoInsen :
  forall (hd : addr) (v : nat) (h : heap) (rf rf' : rfile),
    (pto flist v) (h, rf) -> (pto flist v) (h, rf').
Proof.
unfold pto; unfold getValinH; auto.
Save.

Lemma beq_nat_neq :
  forall (m n : nat),
    m <> n -> false = EqNat.beq_nat m n.
Proof.
double induction m n;simpl;intuition.
Save.

Lemma MBlkUP2 : forall (up : addr) (lnk : addr) (siz : nat) (s : state),
  MBlk up lnk siz s -> up > 2.
Proof.
intros.
unfold MBlk in H.
inversion_clear H; auto.
Save.

Ltac mblkup2 := 
  match goal with
    | [H : MBlk ?u ?l ?siz ?s |- ?u > 2] => apply MBlkUP2 with l siz s; auto
  end.

Lemma MBlkChunk : forall (hptr : addr) (lnk : addr) (siz : nat) 
  (h : heap) (rf : rfile), MBlk (hptr + 2) lnk siz (h, rf) -> 
  (hread' h hptr = lnk /\ hread' h (hptr + 1) = siz).
Proof.
intros.
inversion_clear H; inversion_clear H1.
splitstar1; split.
unfold getValinH in H1; unfold hread'.
cut (hread h hptr = Some lnk).
intros; rewrite H3; auto.
rewrite <- H1.
cut (hptr = hptr + 2 - 2 ).
intros; auto.
omega.
unfold getValinH in H2; unfold hread'.
cut (hread h (hptr + 1)  = Some siz).
intros; rewrite H3; auto.
rewrite <- H2; simpl.
unfold hread; unfold hfree.
cut (EqNat.beq_nat (hptr + 2 - 2 )(hptr + 2 -1)=false).
intros; rewrite H3.
cut (hptr + 1 = hptr + 2 - 1).
intros;auto.
omega.
symmetry .
apply beq_neq_false.
omega.
Save.

Ltac mblkchunk :=
  match goal with
    | [H : MBlk (?hptr + 2) ?lnk ?siz (?h, ?rf) |- hread' ?h ?hptr = ?lnk] =>
      assert (hread' h hptr = lnk /\ hread' h (hptr + 1) = siz); 
	[apply MBlkChunk with rf; auto | intuition]
    | [H : MBlk (?hptr + 2) ?lnk ?siz (?h, ?rf) |- hread' ?h (?hptr + 1) = ?siz] =>
      assert (hread' h hptr = lnk /\ hread' h (hptr + 1) = siz); 
	[apply MBlkChunk with rf; auto | intuition]
  end.  
  
Lemma pnalInsen :
  forall (sz : nat),
  forall (h : heap) (rf rf' : rfile) (p : addr) ,
    ptoanynlist p sz (h, rf) -> ptoanynlist p sz (h, rf').
Proof.
induction sz.
intros.
unfold ptoanynlist; unfold ptoanynlist in H.
unfold emp; unfold emp in H.
unfold getValinH; unfold getValinH in H.
auto.
intros.
unfold ptoanynlist in H; unfold ptoanynlist.
fold ptoanynlist in H; fold ptoanynlist.
splitstar.
getstar x x0.
getland.
apply IHsz with rf; auto.
Save.

Lemma StkInsen :
  forall (btm : nat) (h : heap) (rf rf' : rfile), Stk btm (h, rf) -> Stk btm (h, rf').
Proof.
intros.
unfold Stk in H; unfold Stk.
apply pnalInsen with rf.
auto.
Save.

Lemma hwrite_append :
  forall (h1 h2 h : heap) (i : addr) (v : nat),
    hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
    hindom i h1 -> heap_eq (happend (hwrite h1 i v) h2) (hwrite h i v).
Proof.
intros.
constructor;intros.
transitivity (hwrite (happend h1 h2) i v i0).
unfold happend;unfold hwrite.
compare i i0;intros.
assert (true=EqNat.beq_nat i i0);[rewrite e;apply EqNat.beq_nat_refl|idtac].
rewrite <- H2;simpl;auto.
assert (false=EqNat.beq_nat i i0);[apply beq_nat_neq;auto|idtac].
rewrite <- H2;simpl;auto.
inversion_clear H0.
unfold hwrite.
compare i i0;intros.
assert (true=EqNat.beq_nat i i0);[rewrite e;apply EqNat.beq_nat_refl|idtac].
rewrite <- H2;simpl;auto.
assert (false=EqNat.beq_nat i i0);[apply beq_nat_neq;auto|idtac].
rewrite <- H2;simpl;auto.
Save.


(** assertion stk retain ***)
Lemma pnalRet :
  forall (sz :addr),
    forall (i : addr) (v : nat) (h : heap) (rf : rfile) (p : addr),
      hindom i h -> ptoanynlist p sz (h, rf) -> ptoanynlist p sz (hwrite h i v, rf).
Proof.
induction sz.
intros.
inversion_clear H.
unfold ptoanynlist in H0;unfold emp in H0;unfold getValinH in H0;unfold hread in H0.
assert (h i = None).
apply H0.
rewrite H in H1;discriminate.
intros.
unfold ptoanynlist in H0;unfold ptoanynlist.
fold ptoanynlist in H0;fold ptoanynlist.
splitstar.
compare p i;intros.
assert (hindom i x).
inversion_clear H2.
inversion_clear H3.
unfold getValinH in H2;unfold hread in H2.
rewrite e in H2.
apply indom_some with x1;auto.
getstar (hwrite x i v) x0.
getland.
apply notin_write_preserve_disj_L;auto.
apply disj_inl_notinr with x;auto.
apply hwrite_append;auto.
rewrite e;unfold ptoany.
exists v.
unfold pto;simpl.
unfold hread;unfold hwrite;simpl.
getland.
assert (true = EqNat.beq_nat i i);[apply EqNat.beq_nat_refl|idtac].
rewrite <- H5;simpl;auto.
intros.
assert (false = EqNat.beq_nat i j);[apply beq_nat_neq;auto|idtac].
rewrite <- H6;simpl;auto.
inversion_clear H2.
inversion_clear H7.
simpl in H8;unfold hread in H8;rewrite e in H8.
apply H8;auto.
getstar x (hwrite x0 i v).
inversion  H2.
inversion_clear H3.
simpl in H6;unfold hread in H6.
assert (hnotindom i x).
apply notindom_none;apply H6;auto.
assert (hindom i x0).
apply indom_part_R with h x;auto.
getland.
apply notin_write_preserve_disj_R;auto.
apply heap_eq_trans with (happend (hwrite x0 i v) x);[apply append_sym|apply hwrite_append;auto].
apply notin_write_preserve_disj_R;auto.
apply disj_sym;auto.
apply heap_eq_trans with (happend x x0);[apply append_sym;auto|auto].
apply disj_sym;auto.
Save.

Lemma StkRet :
  forall (i : addr) (v : addr) (h : heap) (rf : rfile) (bottom : addr),
    hindom i h -> Stk bottom (h, rf) -> Stk bottom (hwrite h i v, rf).
Proof.
unfold Stk;intros.
apply pnalRet;auto.
Save.

Lemma pnalPtrIn :
  forall (siz :nat), 
    forall(h :heap) (rf :rfile) (hptr po :addr),
      ptoanynlist hptr siz (h, rf) ->
      po >= hptr -> po < hptr + siz -> hindom po h.
Proof.
induction siz.
 intros.
   assert False; [ omega | tauto ].
 intros.
   unfold ptoanynlist in H; fold ptoanynlist in H.
   splitstar.
   inversion H0.
  assert (hindom po x).
   inversion_clear H3.
     inversion_clear H6.
     simpl in H3; unfold hread in H3; rewrite <- H4; apply indom_some with x1;
      auto.
   apply heap_eq_preserve_indom with (happend x x0);
    [ auto | apply append_preserve_indom_L; auto ].
  rewrite H6.
    assert (po >= S hptr); [ omega | idtac ].
    assert (hindom po x0).
   apply IHsiz with rf (S hptr); auto.
     omega.
   apply heap_eq_preserve_indom with (happend x x0);
    [ auto | apply append_perserve_indom_R; auto ].
Save.

(** not in stk domain means = 0 or > TOP **)
Lemma NotinStk :
  forall (i : addr) (h : heap) (rf : rfile) (bottom : addr),
    Stk bottom (h, rf) -> hnotindom i h -> i = 0 \/ i > bottom.
Proof.
intros.
assert (i <> 0 /\ i <= bottom -> False).
intros.
assert (hindom i h).
unfold Stk in H.
apply pnalPtrIn with bottom rf 1; [auto | omega | omega].
inversion_clear H2; inversion_clear H0; rewrite H3 in H2; discriminate.
omega.
Save.

Ltac fromnotinstk :=
  match goal with
    | [H1 : Stk ?bot (?h, ?rf), H2 : hnotindom ?i ?h |- _] =>
      assert (i = 0 \/ i > bot); [apply NotinStk with h rf;auto|idtac]
  end.

Lemma MBlkInsen :
  forall (h : heap) (rf rf' : rfile) (p : addr) (lnk : addr) (sz : nat),
    MBlk p lnk sz (h, rf) -> MBlk p lnk sz (h, rf').
Proof.
intros.
unfold MBlk in H; unfold MBlk.
splitland; getland.
splitstar1.
getstar1.
unfold removeS in H4; unfold removeS.
apply pnalInsen with rf; auto.
Save.

Lemma MBlkLstInsen :
  forall (n : nat),
    forall (h : heap) (rf rf' : rfile)(p q : addr),
      MBlkLst n p q (h, rf) -> MBlkLst n p q (h, rf').
Proof.
induction n.
intros.
inversion_clear H.
constructor.
unfold emp; unfold emp in H0; [unfold getValinH; unfold getValinH in H0; auto].
intros.
inversion_clear H.
apply mbl_S with h1 h2 n0 p'; auto; [auto | auto].
apply MBlkInsen with rf; auto.
apply IHn with rf; auto.
Save.

Lemma GoodInsen :
  forall (h : heap) (rf rf' : rfile) (p : addr),
    Good p (h, rf) -> Good p (h, rf').
Proof.
intros.
unfold Good in H; unfold Good.
splitex.
exists x.
apply MBlkLstInsen with rf; auto.
Save.

Lemma HIntactInsen :
  forall (h : heap) (rf rf' : rfile) (s'' : state) (top bottom : addr),
   HIntact top bottom (h, rf) s'' -> HIntact top bottom (h, rf') s''.
Proof.
intros.
unHIntact H.
intros.
replace (getValinH (h, rf') i) with (getValinH (h, rf) i).
apply H; auto.
unfold getValinH; auto.
Save.

Lemma MidLRfInsen :
  forall (h : heap) (rf rf' : rfile) (fl : addr) (pr po : addr),
    MidL fl pr po (h, rf) -> MidL fl pr po (h, rf').
Proof.
unfold MidL; unfold star.
intros.
splitex.
exists x; exists x0; exists x1.
splitland; getland.
unfold EndL in H1; unfold EndL.
splitland; getland.
intros.
apply MBlkLstInsen with rf.
apply H2; auto.
intros; elim H4.
intros; getland.
splitex.
exists x2; exists x3.
splitstar; unfold star.
exists x4; exists x5.
getland.
apply MBlkLstInsen with rf; auto.
apply MBlkInsen with rf; auto.
auto.
apply MBlkLstInsen with rf; auto.
Save.


Lemma EndLRfInsen :
  forall (h : heap) (rf rf' : rfile) (fl : addr) (pr po : addr),
    EndL fl pr po (h, rf) -> EndL fl pr po (h, rf').
Proof.
unfold EndL.
intros.
getland; splitland.
intros.
apply MBlkLstInsen with rf.
apply H0; auto.
intros.
elim H1.
intros.
getland; auto.
splitex; exists x; exists x0.
splitstar; unfold star.
exists x1; exists x2.
getland.
apply MBlkLstInsen with rf; auto.
apply MBlkInsen with rf; auto.
auto.
Save.

(***register insensitivity is distributive on star ***)
Lemma StarRfInsen :
  forall (h : heap) (rf rf' : rfile) (P Q : assert),
    (forall (h' : heap), P (h', rf) -> P (h', rf')) ->
    (forall (h' : heap), Q (h', rf) -> Q (h', rf')) ->
    ((star P Q) (h, rf) -> (star P Q) (h, rf')).
Proof.
intros.
unfold star at 1 in H1.
splitex; [splitland].
getstar x x0.
getland.
Save.

(** rf can be auto determined , but **)
Ltac rfinsen  := 
  match goal with
    | [H : star _ _ (_, ?rf) |- star _ _ (_, _)] => 
      apply StarRfInsen with rf; auto
    | [H : MBlk _ _ _ (_, ?rf) |- MBlk _ _ _ (_, _)] =>
      apply MBlkInsen with rf; auto
    | [H : pto _ _ (_, ?rf) |- pto _ _ (_, _)] =>
      apply ptoInsen with rf; auto
    | [H : Good _ (_, ?rf) |- Good _ (_, _)] =>
      apply GoodInsen with rf; auto
    | [H : Stk _ (_, ?rf) |- Stk _ (_, _)] =>
      apply StkInsen with rf; auto
    | [H : MidL _ _ _ (_, ?rf) |- MidL _ _ _ (_, _)] =>
      apply MidLRfInsen with rf; auto
    | [H : EndL _ _ _ (_, ?rf) |- EndL _ _ _ (_, _)] =>
      apply EndLRfInsen with rf; auto
    | [H : MBlkLst _ _ _ (_, ?rf) |- MBlkLst _ _ _ (_, _)] =>
      apply MBlkLstInsen with rf; auto
    | [H : ptoanynlist _ _ (_, ?rf) |- ptoanynlist _ _ (_, _)] =>
      apply pnalInsen with rf; auto
  end.

(** one constructor for indom **)
Lemma InDomC :
  forall (h h1 h2 : heap) (x : addr), 
    hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
    hindom x h1 \/ hindom x h2 -> hindom x h.
Proof.
intuition.
inversion_clear H2.
apply indom_some with v.
inversion_clear H0.
cut (happend h1 h2 x = h x);[intuition|apply H2].
rewrite <- H1;rewrite <- H0.
unfold happend;rewrite H1;auto.
inversion_clear H2.
apply indom_some with v.
inversion_clear H0.
cut (happend h1 h2 x = h x);[intuition|apply H2].
rewrite <- H1;rewrite <- H0.
cut (hnotindom x h1 \/ hnotindom x h2);[intuition|inversion H;auto].
inversion H4.
unfold happend;rewrite H3;rewrite H1;auto.
inversion H4.
rewrite H1 in H3;discriminate.
Save.

(** addr is in the left heap of star **)
Ltac inlh := 
  match goal with
    | [HP : hdisj ?h1 ?h2, Heq : heap_eq (happend ?h1 ?h2) ?h |- hindom ?x ?h ] => 
      apply InDomC with h1 h2; auto; [left]
  end.

(** addr in right heap **)
Ltac inrh := 
  match goal with
    | [HP : hdisj ?h1 ?h2, Heq : heap_eq (happend ?h1 ?h2) ?h |- hindom ?x ?h ] => 
      apply InDomC with h1 h2; auto; [right]
  end.

Lemma ListInDom :
  forall (len : nat)(m : addr)(h : heap)(x : addr)(rf : rfile),
    ptoanynlist m len (h,rf) -> x < len + m /\ x >= m  -> hindom x h.
Proof.
induction len.
intros.
cut False;intuition.
intros.
unfold ptoanynlist in H.
fold ptoanynlist in H.
assert (((x < len + S m) /\ x >= S m) \/ (x = m)).
omega.
splitstar.
inversion_clear H1.
assert (hindom x x1).
apply IHlen with (S m) rf;auto.
apply heap_eq_preserve_indom with (happend x0 x1);auto.
apply append_perserve_indom_R;auto.
assert (hindom x x0).
inversion_clear H3.
rewrite H0.
inversion_clear H1.
unfold getValinH in H3.
unfold hread in H3.
apply indom_some with x2;auto.
apply heap_eq_preserve_indom with (happend x0 x1);auto.
apply append_preserve_indom_L;auto.
Save.

(** addr in the dom of Stack ***)
Lemma ptoInDom :
  forall (x : addr) (v : nat) (h : heap) (rf : rfile), 
    pto x v (h, rf) -> hindom x h.
Proof.
intros.
inversion_clear H.
unfold getValinH in H0.
unfold hread in  H0.
apply indom_some with v.
auto.
Save.

Ltac ptoindom :=
  match goal with
    | [H : pto ?X ?v (?h, ?rf) |- hindom ?X ?h] =>
      apply ptoInDom with v rf; auto
  end.

Lemma MblkPtrIn :forall (h : heap) (rf :rfile) (hptr lnk po : addr) (siz :nat),
   MBlk (hptr + 2) lnk siz (h, rf) ->
   po >= hptr -> po < hptr + siz -> hindom po h.
Proof.
intros.
inversion_clear H.
inversion_clear H3; splitstar1.
simpl in H3; unfold hread in H3; simpl in H4; unfold hread in H4.
assert (hptr + 2 - 2 = hptr);
 [ omega | rewrite H5 in H3; rewrite H5 in H6; rewrite H5 in H4; clear H5 ].
clear H2.
assert (hptr + 2 - 1 = S hptr);
 [ omega | rewrite H2 in H4; rewrite H2 in H6; clear H2 ].
unfold removeS in H6.
compare hptr po; intuition.
 apply indom_some with lnk.
   rewrite <- e; auto.
 compare (S hptr) po; intuition.
  apply indom_some with siz.
    rewrite <- e; unfold hfree in H4.
    assert (false = EqNat.beq_nat hptr (S hptr));
     [ apply beq_nat_neq; omega | rewrite <- H2 in H4; simpl in H4; auto ].
  assert (po >= hptr + 2); [ omega | clear n; clear H0 ].
    assert (hindom po (hfree (hfree h hptr) (S hptr))).
   apply pnalPtrIn with (siz - 2) rf (hptr + 2); auto.
     omega.
   inversion_clear H0.
     unfold hfree in H5.
     assert (false = EqNat.beq_nat (S hptr) po);
      [ apply beq_nat_neq; omega
      | rewrite <- H0 in H5; simpl in |- *; clear H0 ].
     assert (false = EqNat.beq_nat hptr po);
      [ apply beq_nat_neq; omega
      | rewrite <- H0 in H5; simpl in |- *; clear H0 ].
     apply indom_some with v; auto.
Save.

(*Another Proof of MBlkInDom
Lemma MBlkInDom : forall (hptr : addr) (lnk : addr) (siz : nat) (h : heap) (rf : rfile),
  MBlk (hptr + 2) lnk siz (h, rf) -> (hindom (hptr + 0) h /\ hindom (hptr + 1) h).
Proof.
intuition.
 apply MblkPtrIn with rf hptr lnk siz; try omega; auto.
   inversion_clear H; inversion_clear H1; omega.
 apply MblkPtrIn with rf hptr lnk siz; try omega; auto.
   inversion_clear H; inversion_clear H1; omega.
Save.
*)

Lemma MBlkInDom : forall (hptr : addr) (lnk : addr) (siz : nat) (h : heap) (rf : rfile),
  MBlk (hptr + 2) lnk siz (h, rf) -> (hindom (hptr + 0) h /\ hindom (hptr + 1) h).
Proof.
intros.
inversion_clear H.
inversion_clear H1.
splitstar1.
split.
unfold getValinH in H1.
unfold hread in H1.
apply indom_some with lnk.
rewrite <- H1.
cut (hptr + 2 - 2 = hptr + 0).
intros; auto.
omega.
apply indom_some with siz.
rewrite <- H2.
simpl in |- *.
unfold hread; unfold hfree.
cut (EqNat.beq_nat (hptr +2 -2)(hptr + 2 - 1 ) = false).
intros.
rewrite H3.
cut (hptr + 1 = hptr + 2 - 1).
intros; auto.
omega.
symmetry .
apply beq_neq_false.
omega.
Save.



Ltac mblkindom :=
  match goal with
    | [H : MBlk (?p + 2) ?lnk ?siz (?h, ?rf) |- hindom (?p  + 0) ?h] =>
      assert (hindom (p + 0) h /\ hindom (p + 1) h); [apply MBlkInDom with lnk siz rf;trivial |
	intuition]
    | [H : MBlk (?p + 2) ?lnk ?siz (?h, ?rf) |- hindom ?p ?h] =>
      replace p with (p + 0); [
      assert (hindom (p + 0) h /\ hindom (p + 1) h); [apply MBlkInDom with lnk siz rf;trivial |
	intuition] | auto]

    | [H : MBlk (?p + 2) ?lnk ?siz (?h, ?rf) |- hindom (?p  + 1) ?h] =>
      assert (hindom (p + 0) h /\ hindom (p + 1) h); [apply MBlkInDom with lnk siz rf;trivial |
	intuition]
  end.

Lemma MidLInDom : forall (fl pr po : addr) (h : heap) (rf : rfile),
  MidL fl pr po (h, rf) -> po <> pnil -> (hindom po h /\ hindom (po + 1) h).
Proof.
intros.
inversion_clear H.
splitstar.
inversion H4.
 rewrite H7 in H0.
   assert False; [ auto | tauto ].
 constructor.
  assert (hindom po h1).
   apply MblkPtrIn with rf po p' n; auto.
     inversion_clear H12; inversion_clear H14; omega.
   apply heap_eq_preserve_indom with (happend x0 x1); auto.
     apply append_perserve_indom_R; auto.
     apply heap_eq_preserve_indom with (happend h1 h2); auto.
     apply append_preserve_indom_L; auto.
  assert (hindom (po + 1) h1).
   apply MblkPtrIn with rf po p' n; auto.
    omega.
    inversion_clear H12; inversion_clear H14; omega.
   apply heap_eq_preserve_indom with (happend x0 x1); auto.
     apply append_perserve_indom_R; auto.
     apply heap_eq_preserve_indom with (happend h1 h2); auto.
     apply append_preserve_indom_L; auto.
Save.


Ltac midlindom :=
  match goal with
    | [H : MidL ?fl ?pr ?po (?h, ?rf) |- hindom ?po ?h] =>
      assert (hindom po h /\ hindom (po + 1) h);
	[apply MidLInDom with fl pr rf; auto | intuition]
    | [H : MidL ?fl ?pr ?po (?h, ?rf) |- hindom (?po + 1) ?h] =>
      assert (hindom po h /\ hindom (po + 1) h);
	[apply MidLInDom with fl pr rf; auto | intuition]
  end.

Lemma MidLInDom0 : forall (fl pr po : addr) (h : heap) (rf : rfile),
  MidL fl pr po (h, rf) -> pr <> pnil -> (hindom pr h /\ hindom (pr + 1) h).
Proof.
intros.
inversion_clear H.
splitstar.
inversion_clear H2; clear H3.
assert
 ((pr < po \/ po = 0) /\
  (exists n : nat,
     (exists m : nat, star (MBlkLst m fl pr) (MBlk (pr + 2) po n) (x0, rf))));
 [ apply H5; auto | inversion_clear H2; clear H5 ].
inversion_clear H6; inversion_clear H2; splitstar.
constructor.
 assert (hindom pr x5).
  apply MblkPtrIn with rf pr po x2; auto.
    inversion_clear H8; inversion_clear H9; omega.
  apply heap_eq_preserve_indom with (happend x0 x1); auto.
    apply append_preserve_indom_L.
    apply heap_eq_preserve_indom with (happend x4 x5); auto.
    apply append_perserve_indom_R; auto.
 assert (hindom (pr + 1) x5).
  apply MblkPtrIn with rf pr po x2; auto.
   omega.
   inversion_clear H8; inversion_clear H7; omega.
  apply heap_eq_preserve_indom with (happend x0 x1); auto.
    apply append_preserve_indom_L.
    apply heap_eq_preserve_indom with (happend x4 x5); auto.
    apply append_perserve_indom_R; auto.
Save.



Ltac midlindom0 :=
  match goal with
    | [H : MidL ?fl ?pr ?po (?h, ?rf) |- hindom ?pr ?h] =>
      assert (hindom pr h /\ hindom (pr + 1) h);
	[apply MidLInDom0 with fl po rf; auto | intuition]
    | [H : MidL ?fl ?pr ?po (?h, ?rf) |- hindom ?pr ?h] =>
      assert (hindom pr h /\ hindom (pr + 1) h);
	[apply MidLInDom0 with fl po rf; auto | intuition]
	
  end.


Lemma ptoHRead :
  forall (x : addr) (v : nat) (h : heap) (rf : rfile),
    (pto x v) (h, rf) -> (Some v = hread h x).
Proof.
intros.
inversion_clear H.
unfold getValinH in H0.
auto.
Save.

Hint Resolve ptoHRead : fsl.

Ltac ptoread :=
  match goal with
    | [ H : context [pto ?x ?v] |- context [hread ?h ?x] ] =>
      replace (hread h x) with (Some v); auto
  end.

Lemma StkInDom :
  forall (top :addr),
    forall (h : heap) (x : addr) (rf : rfile),
      Stk top (h, rf) -> x <= top /\ x > 0 -> hindom x h.
Proof.
unfold Stk.
intros.
apply ListInDom with top 1 rf;auto.
omega.
Save.

Ltac stkindom :=
  match goal with
    | [H : Stk ?TOP (?h, ?rf) |- hindom _ ?h] =>
      apply StkInDom with TOP rf; auto;[intuition]
  end.

(** semetry of star **)
Lemma StarSym : forall (a1 a2 : assert) (s : state),
  (star a1 a2) s -> ((star a2 a1) s).
Proof.
induction s;unfold star;intros.
splitex;intuition.
exists x0;exists x;intuition.
apply heap_eq_trans with (happend x x0);
[apply append_sym;apply disj_sym;auto|auto].
Save.

Ltac starsym :=
  match goal with
    | |- star _ _ _ => apply StarSym; auto
  end.


(**left association of star***)
Lemma StarAssocL :
  forall (a1 a2 a3 : assert) (s : state),
    star (star a1 a2) a3 s -> star a1 (star a2 a3) s.
Proof.
unfold star.
intuition.
induction s.
inversion_clear H.
inversion_clear H0.
intuition.
inversion_clear H1.
inversion_clear H2.
intuition.
exists x1;exists (happend x2 x0).
intuition.
apply disj_append_L;intuition.
apply disj_sym.
cut(hdisj x0 x1 /\ hdisj x0 x2);[intuition|apply disj_append_R].
apply disj_heap_eq with x;[apply disj_sym;auto|apply heap_eq_sym;auto].
apply heap_eq_trans with (happend x x0);intuition.
apply heap_eq_trans with (happend (happend x1 x2) x0).
apply append_assoc;auto.
apply disj_sym.
cut(hdisj x0 x1 /\ hdisj x0 x2);[intuition|apply disj_append_R].
apply disj_heap_eq with x;[apply disj_sym;auto|apply heap_eq_sym;auto].
apply disj_sym.
cut(hdisj x0 x1 /\ hdisj x0 x2);[intuition|apply disj_append_R].
apply disj_heap_eq with x;[apply disj_sym;auto|apply heap_eq_sym;auto].
apply heap_eq_append;auto.
exists x2;exists x0;intuition.
apply disj_sym.
cut(hdisj x0 x1 /\ hdisj x0 x2);[intuition|apply disj_append_R].
apply disj_heap_eq with x;[apply disj_sym;auto|apply heap_eq_sym;auto].
constructor;auto.
Save.

Lemma StarAssocR :
  forall (a1 a2 a3 : assert) (s : state),
    star a1 (star a2 a3) s -> star (star a1 a2) a3 s.
Proof.
induction s.
unfold star;intuition.
inversion_clear H.
inversion_clear H0;intuition.
inversion_clear H3.
inversion_clear H2;intuition.
exists (happend x x1);exists x2;intuition.
apply disj_sym;apply disj_append_L;intuition;apply disj_sym;auto.
cut(hdisj x x1 /\ hdisj x x2);
[intuition|apply disj_append_R;apply disj_heap_eq with x0;auto].
apply heap_eq_sym;auto.
apply heap_eq_trans with (happend x x0);auto.
apply heap_eq_trans with (happend x (happend x1 x2)).
apply heap_eq_sym;apply append_assoc.
cut(hdisj x x1 /\ hdisj x x2);
[intuition|apply disj_append_R;apply disj_heap_eq with x0;auto].
apply heap_eq_sym;auto.
cut(hdisj x x1 /\ hdisj x x2);
[intuition|apply disj_append_R;apply disj_heap_eq with x0;auto].
apply heap_eq_sym;auto.
auto.
apply heap_eq_trans with (happend x0 x);[intros|apply append_sym;auto].
apply heap_eq_trans with (happend (happend x1 x2) x).
apply append_sym;auto.
apply disj_heap_eq with x0;[auto|apply heap_eq_sym;auto].
apply heap_eq_append;auto.
apply disj_sym;auto.
exists x;exists x1;intuition.
cut(hdisj x x1 /\ hdisj x x2);
[intuition|apply disj_append_R;apply disj_heap_eq with x0;auto].
apply heap_eq_sym;auto.
constructor;auto.
Save.

Ltac starrev H :=
  match goal with
    | H : star ?a1 ?a2 ?s |- _ => assert (star a2 a1 s); 
      [apply StarSym; trivial | clear H]
  end.

Ltac starrev' :=
  match goal with
    | [ |- star ?a1 ?a2 ?s ] => assert (star a2 a1 s);
      [idtac | apply StarSym; trivial]
  end.

Ltac starassocl H :=
  match goal with
    | [H : star (star ?a1 ?a2) ?a3 ?s |- _] => assert (star a1 (star a2 a3) s);
      [apply StarAssocL; trivial | clear H]
  end.

Ltac starassocl' :=
  match goal with
    | [H : star (star ?a1 ?a2) ?a3 ?s |- _] => assert (star a1 (star a2 a3) s);
      [apply StarAssocL; trivial | clear H]
  end.

Ltac assoclgoal :=
  match goal with
    | [ |- star (star ?a1 ?a2) ?a3 ?s ] => assert (star a1 (star a2 a3) s);
      [ idtac | apply StarAssocR; trivial]
  end.
Ltac starassocr H :=
  match goal with
    | [H : star ?a1 (star ?a2 ?a3) ?s |- _] => assert (star (star a1 a2) a3 s);
      [apply StarAssocR; trivial | clear H]
  end.

Ltac starassocr' :=
  match goal with
    | [H : star ?a1 (star ?a2 ?a3) ?s |- _] => assert (star (star a1 a2) a3 s);
      [apply StarAssocR; trivial | clear H]
  end.

Ltac assocrgoal :=
  match goal with
    | [|- star ?a1 (star ?a2 ?a3) ?s ] => assert (star (star a1 a2) a3 s);
      [idtac | apply StarAssocL; trivial]
  end.

Lemma SWPartLH :
  forall (h1 h2 h : heap) (rf : rfile) (P Q : assert) (i : addr) (v : nat),
    hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
    Q (h2, rf) -> hindom i h1 ->
    P (hwrite h1 i v, rf) -> star P Q (hwrite h i v, rf).
Proof.
intros.
unfold star.
exists (hwrite h1 i v);exists h2;intuition.
constructor;intros.
inversion_clear H.
assert (hnotindom a h1 \/ hnotindom a h2);[apply H4|intuition].
left.
constructor.
unfold hwrite.
inversion_clear H5;inversion_clear H2;intuition.
assert (false = EqNat.beq_nat i a).
assert (i<>a);intuition.
rewrite H2 in H5;rewrite H in H5;discriminate.
rewrite <- H2;simpl;auto.
apply hwrite_append;auto.
Save.

Lemma SWPartRH :
  forall (h1 h2 h : heap) (rf : rfile) (P Q : assert) (i : addr) (v : nat),
    hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
    P (h1, rf) -> hindom i h2 ->
    Q (hwrite h2 i v, rf) -> star P Q (hwrite h i v, rf).
Proof.
intros.
unfold star.
exists h1;exists (hwrite h2 i v);intuition.
constructor;intros.
inversion_clear H.
assert (hnotindom a h1 \/ hnotindom a h2);[apply H4|intuition].
right.
constructor.
unfold hwrite.
inversion_clear H5;inversion_clear H2;intuition.
assert (false = EqNat.beq_nat i a).
assert (i<>a);intuition.
rewrite H2 in H5;rewrite H in H5;discriminate.
rewrite <- H2;simpl;auto.
apply heap_eq_trans with (happend (hwrite h2 i v) h1);[apply append_sym|apply hwrite_append;auto].
apply notin_write_preserve_disj_R;auto.
apply disj_inr_notinl with h2;auto.
apply disj_sym;auto.
apply heap_eq_trans with (happend h1 h2);[apply append_sym;auto|auto].
apply disj_sym;auto.
Save.


Ltac swpartrh h1 h2 := try (apply SWPartRH with h1 h2; auto).
Ltac swpartlh h1 h2 := try (apply SWPartLH with h1 h2; auto).

Ltac writerh :=
  match goal with
    | [H : hdisj ?h1 ?h2 |- _] => swpartrh h1 h2; clear H
  end.
Ltac writelh :=
  match goal with
    | [H : hdisj ?h1 ?h2 |- _] => swpartlh h1 h2; clear H
  end.

(** write other slot don't affect HIntact **)
Lemma HIntactRe : 
  forall (h : heap) (top bot : addr) (i : addr) (v : nat) (s'' : state) (rf : rfile),
    HIntact top bot (hwrite h i v, rf) s'' ->
    i <= bot \/ i > top -> HIntact top bot (h, rf) s''.
Proof.
unfold HIntact.
induction s''.
unfold getValinH. 
intuition.
assert (hread (hwrite h i v) i0 = hread h i0).
assert (i0 <> i);try omega.
unfold hread.
unfold hwrite.
assert (false = EqNat.beq_nat i i0);try apply beq_nat_neq;auto.
rewrite <- H4;auto.
rewrite <- H0;apply H;auto.
assert (hread (hwrite h i v) i0 = hread h i0).
assert (i0 <> i);try omega.
unfold hread.
unfold hwrite.
assert (false = EqNat.beq_nat i i0);try apply beq_nat_neq;auto.
rewrite <- H4;auto.
rewrite <- H0;apply H;auto.
Save.

Ltac hintact' := 
  match goal with
    | [H : HIntact _ _ (hwrite _ ?i ?v, _)(_, _) |- _] =>
      apply HIntactRe with i v; auto
  end.
Ltac hintact i v:= try (apply HIntactRe with i v; auto).

(* Many things on eq heaps *)
Lemma h_subst_emp :
  forall (h h' : heap) (rf : rfile),
    (emp (h, rf)) -> (heap_eq h h') -> (emp (h', rf)).
Proof.
unfold emp; unfold getValinH.
intros; replace (hread h' i) with (hread h i); auto.
unfold hread;inversion_clear H0;auto.
Save.

(* star on eq heaps *)
Lemma h_subst_star : 
  forall (h h':heap) ( rf : rfile) (a1 a2 : assert),
    (star a1 a2 (h, rf)) ->
    (heap_eq h h') ->
    (star a1 a2 (h', rf)).
Proof.
intros.
splitstar.
getstar x x0.
getland;auto.
apply heap_eq_trans with h; auto.
Save.


Lemma h_subst_ptoanynlist : 
  forall (n : nat) ( h h' : heap) ( rf : rfile) ( p : addr),
    (ptoanynlist p n (h, rf)) -> (heap_eq h h') -> (ptoanynlist p n (h', rf)).
Proof.
induction n.
intros until p; unfold ptoanynlist.
unfold emp; unfold getValinH.
intros; replace (hread h' i) with (hread h i); auto.
inversion H0;unfold hread;auto.
intros until p.
unfold ptoanynlist; fold ptoanynlist.
apply h_subst_star.
Save.

(*not used*)
Lemma h_subst_MBlk: 
  forall (h h' : heap) (rf : rfile) (p q : addr) (siz : nat),
    (MBlk p q siz (h, rf)) -> (heap_eq h h') -> (MBlk p q siz (h', rf)).
Proof.
unfold MBlk.
intros.
inversion_clear H0.
splitland; getland.
splitstar1; unfold star1.
getland.
rewrite <- H2; unfold getValinH; unfold hread; auto.
rewrite <- H3; unfold getValinH; unfold hread.
simpl; unfold hfree.
assert (false = EqNat.beq_nat (p - 2)(p - 1)).
apply beq_neq_false; omega.
rewrite <- H4; auto.
unfold removeS; unfold removeS in H5.
apply h_subst_ptoanynlist with (hfree (hfree h (p-2))(p-1)); auto.
apply heap_eq_free.
apply heap_eq_free.
apply all_same; auto.
Save.

Lemma h_subst_MBlkLst : 
  forall(h h' : heap) (rf : rfile) (p q : addr) (n : nat),
    (MBlkLst n p q (h, rf)) -> (heap_eq h h') -> (MBlkLst n p q (h', rf)).
Proof.
induction n; intros.
inversion_clear H;apply mbl_O.
apply h_subst_emp with h; auto.
inversion_clear H.
apply mbl_S with h1 h2 n0 p' ; auto.
apply heap_eq_trans with h; auto.
Save.


Lemma h_subst_EndL : 
  forall(h h' : heap) (rf : rfile) (p q fl : addr),
    (EndL fl p q (h, rf)) -> (heap_eq h h') -> (EndL fl p q (h', rf)).
Proof.
induction p; intros.
 inversion_clear H.
   unfold EndL in |- *; getland; intros.
  apply h_subst_MBlkLst with h; auto.
  assert False; [ auto | tauto ].
 inversion_clear H.
   assert
    ((S p < q \/ q = 0) /\
     (exists n : nat,
        (exists m : nat,
           star (MBlkLst m fl (S p)) (MBlk (S p + 2) q n) (h, rf))));
    [ apply H2; omega | clear H2; inversion_clear H ].
   inversion_clear H3.
   inversion_clear H.
   splitstar.
   unfold EndL in |- *; getland; intros; auto.
  discriminate.
  constructor; auto.
    exists x; exists x0.
    getstar x1 x2.
    getland; auto.
    apply heap_eq_trans with h; auto.
Save.


(*not used*)
Lemma h_subst_Good : 
  forall (h h' : heap) (rf : rfile) (fl : addr),
    (Good fl (h, rf)) -> (heap_eq h h') -> (Good fl (h', rf)).
Proof.
unfold Good.
intros.
inversion_clear H.
exists x.
apply h_subst_MBlkLst with h; auto.
Save.

Lemma Compose_Blk : 
  forall (n1 n2 : nat) (s : state) (fl p q :addr) (siz : nat),
    ((star
      (star
	(MBlkLst n1 fl p)
	(MBlk (2 + p) q siz)
      )
      (MBlkLst n2 q 0)
    ) s)
    -> ((lt p q) \/ (q = 0))
    -> ((MBlkLst (n1 + S n2) fl 0) s).
Proof.
induction n1; intros.
cut ((star (MBlkLst (0) fl p)
     (star (MBlk (plus (2) p) q siz) (MBlkLst n2 q (0)))) s).
clear H; intro.
induction s; unfold star in H.
inversion_clear H.
inversion_clear H1.
inversion_clear H.
inversion_clear H2.
inversion_clear H3.
inversion_clear H4.
inversion_clear H3.
inversion_clear H4.
inversion_clear H5.
inversion_clear H6.
unfold plus.
inversion_clear H2.
apply mbl_S with x1 x2 siz q; auto.
apply heap_eq_trans with x0; auto.
apply heap_eq_trans with (happend x x0); auto.
apply heap_eq_sym.
unfold emp in H6;unfold getValinH in H6;unfold hread in H6;simpl.
constructor;intros.
unfold happend.
assert (x i = None);[apply H6|rewrite H2;auto].
assert (2 + p = p + 2);[omega|rewrite H2 in H5;auto].
apply StarAssocL; auto.
cut (star (MBlkLst (S n1) fl p) (
   star (MBlk (plus (2) p) q siz) (MBlkLst n2 q (0))) s).
intro.
replace (plus (S n1) (S n2)) with (S (plus n1 (S n2)));auto with arith.
induction s.
splitstar.
inversion_clear H3.
apply mbl_S with h1 (happend h2 x0) n p'; auto.
apply disj_append_L; auto.
getland.
assert (hdisj x0 h1 /\ hdisj x0 h2);try apply disj_append_R;auto.
apply disj_heap_eq with x;[apply disj_sym;auto|apply heap_eq_sym;auto].
apply disj_sym;inversion_clear H3;auto.
apply heap_eq_sym; auto.
apply heap_eq_trans with (happend (happend h1 h2) x0).
apply heap_eq_trans with (happend x x0); auto.
apply heap_eq_sym;auto.
apply heap_eq_append;apply heap_eq_sym;auto.
apply heap_eq_sym;apply append_assoc;auto.
assert (hdisj x0 h1 /\ hdisj x0 h2);try apply disj_append_R;auto.
apply disj_heap_eq with x;[apply disj_sym;auto|apply heap_eq_sym;auto].
apply disj_sym;inversion_clear H3;auto.
assert (hdisj x0 h1 /\ hdisj x0 h2);try apply disj_append_R;auto.
apply disj_heap_eq with x;[apply disj_sym;auto|apply heap_eq_sym;auto].
apply disj_sym;inversion_clear H3;auto.
apply IHn1 with p q siz; auto.
splitstar.
getstar (happend h2 x1) x2.
getland.
apply disj_sym;apply disj_append_L;auto.
getland.
assert (hdisj x0 h1 /\ hdisj x0 h2);[apply disj_append_R;auto|idtac]. 
apply disj_heap_eq with x;auto.
apply disj_sym; auto.
apply heap_eq_sym;auto.
assert(hdisj h2 x1 /\ hdisj h2 x2);[apply disj_append_R;auto|idtac].
apply disj_heap_eq with x0;auto.
apply disj_sym;inversion_clear H11;auto.
apply heap_eq_sym;auto.
apply disj_sym;inversion_clear H13;auto.
apply disj_sym;auto.
apply heap_eq_trans with  (happend h2 (happend x1 x2)).
assert (hdisj x0 h1 /\ hdisj x0 h2);[apply disj_append_R;auto|idtac]. 
apply disj_heap_eq with x;auto.
apply disj_sym; auto.
apply heap_eq_sym;auto.
assert(hdisj h2 x1 /\ hdisj h2 x2);[apply disj_append_R;auto|idtac].
apply disj_heap_eq with x0;auto.
apply disj_sym;inversion_clear H11;auto.
apply heap_eq_sym;auto.
apply heap_eq_sym;apply append_assoc;auto.
inversion H13;auto.
inversion H13;auto.
apply heap_eq_trans with (happend x0 h2);try apply append_sym;auto.
apply heap_eq_trans with (happend (happend x1 x2) h2);try apply append_sym;auto.
apply disj_heap_eq with x0;auto.
assert(hdisj x0 h1 /\ hdisj x0 h2);[apply disj_append_R;auto|idtac].
apply disj_heap_eq with x;auto.
apply disj_sym;auto.
apply heap_eq_sym;auto.
inversion H11;apply disj_sym;auto.
apply heap_eq_sym;auto.
apply heap_eq_append;auto.
assert(hdisj x0 h1 /\ hdisj x0 h2);[apply disj_append_R;auto|idtac].
apply disj_heap_eq with x;[apply disj_sym;auto|apply heap_eq_sym;auto].
inversion H11;auto.
getstar h2 x1.
constructor.
assert(hdisj h2 x1 /\ hdisj h2 x2);[apply disj_append_R;auto|idtac].
apply disj_heap_eq with x0;[apply disj_sym;auto|apply heap_eq_sym;auto].
assert (hdisj x0 h1 /\ hdisj x0 h2);[apply disj_append_R;auto|idtac].
apply disj_heap_eq with x;[apply disj_sym;auto|apply heap_eq_sym;auto].
inversion_clear H11;auto.
inversion_clear H11;auto.
constructor.
apply heap_eq_refl; auto.
constructor; auto.
auto.
apply StarAssocL; auto.
Save.

Lemma Compose_MBlk_MBlkLst_to_MBlkLst : 
  forall (s : state) ( p p' q : addr)( n siz : nat),
    (star (MBlk (plus (2) p) p' siz) (MBlkLst n p' q) s)
    -> ((lt p p') \/ (p'=pnil))
    -> (MBlkLst (S n) p q s).
Proof.
intros.
induction s; inversion_clear H.
inversion_clear H1.
inversion_clear H.
inversion_clear H2.
inversion_clear H3.
apply mbl_S with x x0 siz p'; auto.
assert (2 + p = p + 2);[omega|rewrite H3 in H2;auto].
Save.

Lemma Compose_MBlkLst_MBlk_to_MBlkLst : 
  forall (n : nat) ( s : state) ( fl p q : addr) ( siz : nat),
    (star (MBlkLst n fl p) (MBlk (plus (2) p) q siz) s)
    -> ((lt p q) \/ (q=pnil))
    -> (MBlkLst (S n) fl q s).
Proof.
induction n; induction s; intros.
 inversion_clear H.
   inversion_clear H1.
   inversion_clear H.
   inversion_clear H2.
   inversion_clear H3.
   inversion_clear H2.
   apply mbl_S with x0 x siz q; auto.
  apply disj_sym; auto.
  apply heap_eq_trans with (happend x x0); auto.
    apply append_sym; apply disj_sym; auto.
  assert (2 + p = p + 2); [ omega | rewrite H2 in H4; auto ].
  apply mbl_O; auto.
 inversion_clear H.
   inversion_clear H1.
   inversion_clear H.
   inversion_clear H2.
   inversion_clear H3.
   inversion_clear H2.
   apply mbl_S with h1 (happend h2 x0) n0 p'; auto.
  apply disj_append_L; auto.
    getland.
    assert (hdisj x0 h1 /\ hdisj x0 h2);
     [ apply disj_append_R; auto | idtac ].
   apply disj_heap_eq with x;
    [ apply disj_sym; auto | apply heap_eq_sym; auto ].
   inversion H2; apply disj_sym; auto.
  apply heap_eq_trans with (happend (happend h1 h2) x0).
   assert (hdisj x0 h1 /\ hdisj x0 h2); [ apply disj_append_R; auto | idtac ].
    apply disj_heap_eq with x;
     [ apply disj_sym; auto | apply heap_eq_sym; auto ].
    apply append_assoc; auto.
     inversion H2; apply disj_sym; auto.
     inversion H2; apply disj_sym; auto.
   apply heap_eq_trans with (happend x x0); auto.
     apply heap_eq_append; auto.
  apply IHn with p siz; auto.
    getstar h2 x0.
    getland; auto.
   assert (hdisj x0 h1 /\ hdisj x0 h2); [ apply disj_append_R; auto | idtac ].
    apply disj_heap_eq with x;
     [ apply disj_sym; auto | apply heap_eq_sym; auto ].
    inversion H2; apply disj_sym; auto.
   apply heap_eq_refl.
Save.


Lemma Compose_EndL_MBlk_to_EndL :
 forall (h1 h2 : heap ) ( rf : rfile) ( fl p q q' : addr) (siz : nat),
 (EndL fl p q (h1,rf)) -> (MBlk (plus (2) q) q' siz (h2,rf))
  -> ((lt q q')\/(q'=pnil))
  -> (hdisj h1 h2)
  -> (EndL fl q q' ((happend h1 h2),rf)).
Proof.
unfold EndL.
intros.
constructor.
cut (gt (plus (2) q) (2)); intros.
rewrite H4 in H3; unfold plus in H3;inversion H3.
inversion H6.
inversion H8.
inversion_clear H0;auto.
intro; constructor; auto.
inversion_clear H.
induction p.
exists siz; exists 0.
unfold star.
cut (MBlkLst (0) fl q (h1,rf));[intro EMP;inversion_clear EMP|apply H4;auto].
exists h1; exists h2.
constructor;auto.
constructor.
apply heap_eq_refl.
constructor; auto.
apply mbl_O; auto.
assert (q + 2 = 2 + q);[omega|rewrite H6;auto].
cut (
     (S p < q \/ q = 0) /\
       (exists n : nat,
          (exists m : nat,
             star (MBlkLst m fl (S p)) (MBlk (S p + 2) q n) (h1, rf)))
); auto.
clear H5; intros.
inversion_clear H.
inversion_clear H6.
inversion_clear H.
exists siz;exists (S x0).
unfold star.
exists h1; exists h2.
constructor; auto.
constructor.
apply heap_eq_refl.
constructor; auto.
apply Compose_MBlkLst_MBlk_to_MBlkLst with (S p) x; auto.
assert (S p + 2 = 2 + S p);[omega|rewrite <- H;auto].
assert (q + 2 = 2 + q);[omega|rewrite H;auto].
Save.


(*** assertion transformation **)

Lemma MidL_Good : forall (s : state) (fl p q : addr),
  (MidL fl p q s) -> (Good fl s).
Proof.
induction p.
unfold MidL;unfold Good;intuition.
inversion_clear H.
exists x.
splitstar.
inversion_clear H1.
assert (MBlkLst 0 fl q (x0, b));auto.
inversion_clear H1.
unfold emp in H5;unfold getValinH in H5;unfold hread in H5;simpl.
assert (heap_eq (happend x0 x1) x1).
constructor;intros.
assert (x0 i = None);auto.
unfold happend;rewrite H1;auto.
assert (heap_eq x1 a).
apply heap_eq_trans with (happend x0 x1);[apply heap_eq_sym;auto|auto].
apply h_subst_MBlkLst with x1;auto.
induction s.
unfold MidL;unfold Good;unfold MidL in IHp;unfold Good in IHp;intros.
inversion_clear H.
splitstar.
inversion_clear H1.
assert ((S p < q \/ q = 0) /\
       (exists n : nat,
          (exists m : nat,
             star (MBlkLst m fl (S p)) (MBlk (S p + 2) q n) (x0, b)))).
apply H4;omega.
inversion_clear H1.
inversion_clear H6.
inversion_clear H1.
exists (x3 + S x).
apply Compose_Blk with (S p) q x2;auto.
getstar x0 x1.
getland.
assert (S p + 2 = 2 + S p).
omega.
rewrite H1 in H6;auto.
Save.


Lemma Good_MidL : forall (s : state) (fl : addr),
  Good fl s -> MidL fl 0 fl s.
Proof.
induction s;unfold Good;unfold MidL;intros.
inversion_clear H.
exists x.
getstar hempty a.
getland.
constructor;intros.
left.
constructor;unfold hempty;simpl;auto.
apply append_emp.
unfold EndL.
getland;intros.
apply mbl_O.
unfold emp;unfold getValinH;unfold hempty;simpl;auto.
assert False;auto.
tauto.
Save.

Lemma MBlkLst_pnil_pnil_emp : 
  forall (n : nat) ( s : state),
    (MBlkLst n pnil pnil s) -> (emp s).
Proof.
intros.
inversion_clear H;auto.
inversion_clear H3.
unfold plus in H.
inversion_clear H.
inversion_clear H3.
inversion_clear H.
Save.


Lemma MidL_EndL : forall (s : state) (fl : addr) (prev : addr),
  MidL fl prev pnil s -> EndL fl prev pnil s.
Proof.
intros.
inversion_clear H.
induction s; splitstar.
apply h_subst_EndL with x0; auto.
apply heap_eq_trans with (happend x1 x0).
assert ( emp (x1, b));[apply MBlkLst_pnil_pnil_emp with x;auto|unfold emp in H2].
unfold getValinH in H2;unfold hread in H2.
constructor;intros.
unfold happend.
assert ( x1 i = None);[apply H2|rewrite H4;simpl;auto].
apply heap_eq_trans with (happend x0 x1);[apply append_sym|auto].
apply disj_sym;auto.
Save.

Lemma Mov_MidL : forall (h : heap) (rf : rfile) (fl pr po : addr),
  po <> pnil -> MidL fl pr po (h, rf) -> 
  MidL fl po (hread' h po) (h, rf).
Proof.
unfold hread';unfold MidL;intuition.
inversion_clear H0;splitstar.
inversion H4.
intuition.
exists m.
getstar (happend x0 h1) h2.
inversion H12.
inversion_clear H15;splitstar1.
simpl in H15.
assert (po + 2 -2 = po);[omega|rewrite H18 in H15].
assert (hread (happend h1 h2) po = hread h1 po).
apply append_preserve_read_L.
assert (hindom po h1);[apply indom_some with p';inversion H15;auto|idtac].
apply disj_inl_notinr with h1;auto. 
assert (hread x1 po = hread (happend h1 h2) po).
apply heap_eq_veq;apply heap_eq_sym;auto.
rewrite H20 in H21;clear H20.
assert (hread (happend x1 x0) po = hread x1 po).
apply append_preserve_read_L.
rewrite H15 in H21.
assert (hindom po x1);[apply indom_some with p';inversion H21;auto|idtac].
apply disj_inr_notinl with x1;auto. 
assert (hread h po = hread (happend x1 x0) po).
apply heap_eq_veq.
apply heap_eq_trans with (happend x0 x1);[auto|apply append_sym;auto].
apply heap_eq_sym;auto.
rewrite H15 in H21;rewrite H21 in H20;rewrite H20 in H22;rewrite H22;simpl. 
getland.
apply disj_sym;apply disj_append_L;getland.
assert (hdisj x0 h1 /\ hdisj x0 h2);[apply disj_append_R;auto|inversion H23;apply disj_sym;auto].
apply disj_heap_eq with x1;[auto|apply heap_eq_sym;auto].
apply disj_sym;auto.
apply heap_eq_trans with (happend x0 (happend h1 h2));auto.
apply heap_eq_sym;apply append_assoc;auto.
assert (hdisj x0 h1 /\ hdisj x0 h2);[apply disj_append_R;auto|inversion H23;apply disj_sym;auto].
apply disj_heap_eq with x1;[auto|apply heap_eq_sym;auto].
apply disj_sym;auto.
assert (hdisj x0 h1 /\ hdisj x0 h2);[apply disj_append_R;auto|inversion H23;apply disj_sym;auto].
apply disj_heap_eq with x1;[auto|apply heap_eq_sym;auto].
apply disj_sym;auto.
apply heap_eq_trans with (happend x0 x1);auto.
apply heap_eq_trans with (happend x1 x0);[idtac|apply append_sym;apply disj_sym;auto].
apply heap_eq_trans with (happend (happend h1 h2) x0).
apply append_sym.
apply disj_heap_eq with x1;[auto|apply heap_eq_sym;auto].
apply heap_eq_append;auto.
apply Compose_EndL_MBlk_to_EndL with pr n;auto.
assert ( 2+po = po+2);[omega|rewrite H23;auto].
assert (hdisj x0 h1 /\ hdisj x0 h2); [ apply disj_append_R; auto | idtac ].
apply disj_heap_eq with x1;[auto|apply heap_eq_sym;auto].
intuition.
Save.
  


Ltac movmidl :=
  match goal with
    | [H : MidL _ ?pr _ (_, _) |- MidL _ _ _ (_, _)] =>
      apply Mov_MidL with pr; auto
  end.

Lemma MidLOrder : forall (fl pr po : addr) (h : heap) (rf : rfile),
  MidL fl pr po (h, rf) -> po <> pnil -> (po < hread' h po \/ hread' h po = pnil).
Proof.
intros.
assert (MidL fl po (hread' h po)(h, rf)).
movmidl.
inversion_clear H1.
splitstar.
inversion_clear H3.
elim H6; [intros; auto | auto].
Save.

Ltac midlorder :=
  match goal with
    | [H : MidL ?fl ?pr ?po (?h, ?rf) |- _] =>
      apply MidLOrder with fl pr rf; auto
  end.

Lemma Write_MBlk_lnk : 
  forall (h : heap) ( rf : rfile) ( p q q':addr)( siz : nat),
    (MBlk (plus (2) p) q siz (h,rf))
    -> (MBlk (plus (2) p) q' siz ((hwrite h p q'),rf)).
Proof.
unfold MBlk;intros.
intuition.
simpl in H2;simpl.
assert (p - 0 = p);[omega|rewrite H1 in H2;rewrite H1;clear H1].
splitstar1.
simpl in H1;simpl in H2;unfold hread in H1;unfold hread in H2.
getstar1;simpl;unfold hread;auto.
unfold hwrite;rewrite H1;auto.
assert (true=EqNat.beq_nat p p);[apply EqNat.beq_nat_refl|rewrite <-H3;auto]. 
unfold hfree;unfold hwrite;simpl.
unfold hfree in H2.
assert (false=EqNat.beq_nat p (S p));[apply beq_nat_neq;auto|rewrite <-H3;rewrite <- H3 in H2;auto]. 
unfold removeS in H4.
apply h_subst_ptoanynlist with (hfree (hfree h p) (S p));auto.
constructor;intros.
unfold hfree;unfold hwrite.
compare (S p) i;intros.
assert (true= EqNat.beq_nat (S p) i);[apply beq_eq_true;auto|rewrite <- H3;simpl;auto].
assert (false= EqNat.beq_nat (S p) i);[apply beq_nat_neq;auto|rewrite <- H3;simpl;auto].
compare p i;intros.
assert (true= EqNat.beq_nat p i);[apply beq_eq_true;auto|rewrite <- H5;simpl;auto].
assert (false= EqNat.beq_nat p i);[apply beq_nat_neq;auto|rewrite <- H5;simpl;auto].
Save.


Lemma MBlk_MBlkLst : forall (h : heap) (rf : rfile) (hptr : addr) 
  (lnk : addr) (siz : nat),
  MBlk (hptr + 2) lnk siz (h, rf) -> 
  MBlkLst 1 hptr pnil (hwrite h (hptr + 0) pnil, rf).
Proof.
intros.
assert (hptr + 0 = hptr);[omega|rewrite H0;clear H0].
apply mbl_S with (hwrite h hptr 0) hempty siz 0;auto.
apply disj_sym;apply disj_emp.
apply heap_eq_trans with (happend hempty (hwrite h hptr 0));
[apply append_sym|apply append_emp].
apply disj_sym;apply disj_emp.
assert (hptr +2 = 2 + hptr);[omega|rewrite H0;clear H0].
apply Write_MBlk_lnk with lnk;auto.
assert (hptr +2 = 2 + hptr);[omega|rewrite <- H0;clear H0;auto].
apply mbl_O.
unfold emp;auto.
Save.

Ltac mblkmblklst :=
  match goal with
    | [ H : MBlk (?hptr + 2) ?lnk ?siz (?h, ?rf) |- _ ] =>
    apply MBlk_MBlkLst with lnk siz; auto
  end.


Lemma MblkHiNotOverlap : forall (h1 h2 : heap) (rf : rfile) (hptr lnk : addr)
  (fl : addr) (pr po : addr) (siz : nat), 
  po <> pnil -> hptr < po -> pr < hptr ->
  hdisj h1 h2 -> MBlk (hptr + 2) lnk siz (h1, rf) ->
  MidL fl pr po (h2, rf) -> po >= hptr + siz. 
Proof.
intros.
inversion_clear H4.
splitstar.
inversion H8.
 rewrite H11 in H; assert False; [ auto | tauto ].
 assert (hdisj h1 x0 /\ hdisj h1 x1).
  apply disj_append_R.
    apply disj_heap_eq with h2; [ auto | apply heap_eq_sym; auto ].
  inversion_clear H18.
    assert (hdisj h1 h0 /\ hdisj h1 h3); [ apply disj_append_R | idtac ].
   apply disj_heap_eq with x1; [ auto | apply heap_eq_sym; auto ].
   inversion_clear H18.
     assert (hindom po h0).
    inversion_clear H16.
      inversion_clear H23.
      splitstar1.
      simpl in H23; unfold hread in H23.
      assert (po + 2 - 2 = po); [ omega | rewrite H25 in H23; clear H25 ].
      apply indom_some with p'; auto.
    assert (hnotindom po h1); [ apply disj_inr_notinl with h0; auto | idtac ].
      assert (po < hptr + siz -> False); [ intros | omega ].
      assert (hindom po h1).
     apply MblkPtrIn with rf hptr lnk siz; auto.
       omega.
     inversion_clear H23; inversion_clear H25.
       rewrite H26 in H23; discriminate.
Save.




Ltac mblkhinotoverlap :=
  match goal with
    | [H1 : MBlk (?hptr + 2) ?lnk ?siz (?h1, ?rf), 
      H2 : MidL ?fl ?pr ?po (?h2, ?rf) |- _] =>
    apply MblkHiNotOverlap with h1 h2 rf lnk fl pr; auto
  end.

Hint Resolve Good_MidL : fsl.
Hint Resolve MidL_EndL : fsl.