Require Import heap2.
Require Import scap.
Require Import scap_rules.
Require Import cdsa.
Require Import Omega.
Require Import properties.


(**connect MBlk with MBlkLst into a new MBlkLst **)
Lemma EndLIndom : forall (h : heap) (rf : rfile) (fl : addr) (pr po : addr),
  pr <> pnil -> EndL fl pr po (h, rf) -> hindom pr h.
Proof.
intros.
inversion_clear H0.
elim H2; [ intros; clear H1; clear H2 | exact H ].
inversion_clear H3; inversion_clear H1.
splitstar.
apply h_subst_indom with (happend x1 x2);
 [ exact H2 | apply append_perserve_indom_R; [ idtac | exact H1 ] ].
apply MblkPtrIn with rf pr po x; auto with arith.
inversion_clear H5; inversion_clear H6; omega.
Save.

Ltac endlindom :=
  match goal with
    | [ H1 : EndL ?fl ?pr ?po (?h, ?rf) |- hindom ?pr ?h] =>
      apply EndLIndom with rf fl po; auto
  end.

Lemma PtoPto : forall (h : heap) (rf : rfile) (fl : addr) (v v' : nat),
  pto fl v (h, rf) -> pto fl v' (hwrite h fl v', rf).
Proof.
unfold pto.
intros.
unfold getValinH; unfold hread; unfold hwrite.
splitland; getland.
simpleq; apply beq_eq_true; auto.
intros.
simplneq.
unfold getValinH in H1; unfold hread in H1.
apply H1; auto.
apply beq_neq_false; auto.
Save.

Ltac ptopto :=
  match goal with
    | [ H1 : pto ?fl ?v (?h, ?rf) |- pto ?fl ?v' (hwrite ?h ?fl ?v', ?rf) ] =>
      apply PtoPto with v; auto
  end.

Lemma CMBlkMBlkLst : forall (h1 h2 h : heap) (hptr lnk lh lt : addr)
  (siz len : nat) (rf : rfile),
  hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
  MBlk (hptr + 2) lnk siz (h1, rf) ->
  MBlkLst len lh lt (h2, rf) ->
  (hptr < lh \/ lh = pnil) ->
  MBlkLst (S len) hptr lt (hwrite h hptr lh, rf).
Proof.
intros.
assert (MBlk (hptr + 2) lh siz (hwrite h1 hptr lh, rf)).
 replace (hptr + 2) with (2 + hptr); [ idtac | omega ].
   apply Write_MBlk_lnk with lnk.
   replace (2 + hptr) with (hptr + 2); [ exact H1 | omega ].
 apply Compose_MBlk_MBlkLst_to_MBlkLst with lh siz; [ idtac | exact H3 ].
   replace (2 + hptr) with (hptr + 2); [ idtac | omega ].
   getstar (hwrite h1 hptr lh) h2; getland.
  apply in_write_preserve_disj_L; [ exact H | idtac ].
    apply MblkPtrIn with rf hptr lnk siz; auto.
    inversion_clear H1; inversion_clear H5; omega.
  apply hwrite_append; auto.
    apply MblkPtrIn with rf hptr lnk siz; auto.
    inversion_clear H1; inversion_clear H5; omega.
Save.



(*Axiom CMBlkMBlkLst' : forall (h1 h2 : heap) (hptr lnk lh lt : addr)
  (siz len : nat) (rf : rfile) ->
  hdisj h1 h2 -> MBlk (hptr + 2) lnk siz (h1, rf) ->
  MBlkLst len lh lt (h2, rf) ->
  (hptr < lh \/ lh = pnil) ->
  MBlkLst (S len) hptr lt (hwrite h1 hptr lh, rf).*)

Ltac cblkblklst :=
  match goal with 
    | [H1 : hdisj ?h1 ?h2, H2 : heap_eq (happend ?h1 ?h2) ?h, H3 : MBlk _ ?lnk ?siz _
      |-
	MBlkLst (S ?len) ?hptr ?lt (hwrite ?h ?hptr ?lh, ?rf)] =>
    apply CMBlkMBlkLst with h1 h2 lnk siz; auto
    | [H1 : hdisj ?h1 ?h2, H2 : MBlk _ ?lnk ?siz _
      |-
	MBlkLst (S ?len) ?htpr ?lt (hwrite (happend ?h1 ?h2) ?hptr ?lh, ?rf)] =>
    apply CMBlkMBlkLst with h1 h2 lnk siz; auto
  end.

(** Should there be more primitive Lemma named MergeMBlk? ***)
(*** unfold MBlkLst then merge two MBlk together **)



Lemma MBlk_ptoanynlist : 
  forall (s : state) (p q : addr) (siz : nat),
    (MBlk (p + 2) q siz s) -> (ptoanynlist p siz s).
Proof.
intros.
inversion_clear H; inversion_clear H1; splitstar1.
assert (ptoanynlist p (S (S (siz - 2))) s);
 [ idtac
 | cut (S (S (siz - 2)) = siz); [ intro E; rewrite <- E; auto | omega ] ].
unfold ptoanynlist in |- *; fold ptoanynlist in |- *.
cut (p + 2 - 2 = p);
 [ intro e; rewrite e in H4; rewrite e in H1; rewrite e in H2; clear e
 | omega ].
cut (p + 2 - 1 = S p);
 [ intro e; rewrite e in H4; rewrite e in H2; clear e | omega ].
cut (S (S p) = p + 2); [ intro e; rewrite e; clear e | omega ].
induction s.
getstar (hsingleton p q) (hfree a p).
unfold hsingleton in |- *; unfold hfree in |- *; getland.
 constructor; intros.
   compare a0 p; intros.
  right.
    apply notindom_none.
    cut (true = EqNat.beq_nat p a0);
     [ intro E; rewrite <- E; clear E; auto | apply beq_eq_true; omega ].
  left.
    apply notindom_none.
    cut (false = EqNat.beq_nat p a0);
     [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; omega ].
 constructor; intros.
   unfold happend in |- *.
   compare p i; intros.
  cut (true = EqNat.beq_nat p i);
   [ intro E; rewrite <- E; clear E; auto | apply beq_eq_true; omega ].
    inversion_clear H1.
    unfold hread in |- *; rewrite e; auto.
  cut (false = EqNat.beq_nat p i);
   [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; omega ].
 unfold ptoany in |- *; unfold pto in |- *.
   exists q; simpl in |- *; unfold hread in |- *.
   cut (true = EqNat.beq_nat p p);
    [ intro E; rewrite <- E; clear E; auto | apply EqNat.beq_nat_refl ].
   constructor; auto.
   intros.
   cut (false = EqNat.beq_nat p j);
    [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; omega ].
 getstar (hsingleton (S p) siz) (hfree (hfree a p) (S p)).
   getland; unfold hsingleton in |- *; unfold hfree in |- *.
  constructor; intros.
    compare a0 (S p); intros.
   right.
     apply notindom_none.
     cut (true = EqNat.beq_nat (S p) a0);
      [ intro E; rewrite <- E; clear E; auto | apply beq_eq_true; auto ].
   left.
     apply notindom_none.
     cut (false = EqNat.beq_nat (S p) a0);
      [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; auto ].
  constructor; intros.
    unfold happend in |- *.
    compare (S p) i; intros.
   cut (true = EqNat.beq_nat (S p) i);
    [ intro E; rewrite <- E; clear E; auto | apply beq_eq_true; auto ].
     inversion_clear H2.
     unfold hread in |- *; rewrite e; auto.
   cut (false = EqNat.beq_nat (S p) i);
    [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; omega ].
  unfold ptoany in |- *; unfold pto in |- *.
    exists siz; simpl in |- *; unfold hread in |- *.
    cut (true = EqNat.beq_nat p p);
     [ intro E; rewrite <- E; clear E; auto | apply EqNat.beq_nat_refl ].
    constructor; auto.
    intros.
    induction j; auto.
    cut (false = EqNat.beq_nat p j);
     [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; auto ].
Save.

Lemma Compose_ptoanynlist : 
  forall (siz1 siz2 : nat) (h1 h2 : heap) (rf:rfile) (p : addr),
    hdisj h1 h2 ->
    (ptoanynlist p siz1 (h1,rf)) -> (ptoanynlist (p + siz1) siz2 (h2,rf)) ->
    (ptoanynlist p (siz1 + siz2) ((happend h1 h2),rf)).
Proof.
induction siz1; intros.
 cut (p + 0 = p); [ intro e; rewrite e in H1; clear e | auto ].
   simpl in |- *.
   unfold ptoanynlist in H0.
   unfold emp in H0; unfold getValinH in H0; unfold hread in H0.
   assert (heap_eq h2 (happend h1 h2)).
  constructor; intros.
    unfold happend in |- *.
    cut (h1 i = None); [ intro e; rewrite e; auto | apply H0 ].
  apply h_subst_ptoanynlist with h2; auto.
 unfold ptoanynlist in H0; fold ptoanynlist in H0.
   splitstar.
   assert (hdisj h2 x /\ hdisj h2 x0);
    [ apply disj_append_R; auto | inversion_clear H4 ].
  apply disj_heap_eq with h1;
   [ apply disj_sym; auto | apply heap_eq_sym; auto ].
  assert (ptoanynlist (S p) (siz1 + siz2) (happend x0 h2, rf)).
   apply IHsiz1; auto.
    apply disj_sym; auto.
    cut (S p + siz1 = p + S siz1); [ intro e; rewrite e; auto | omega ].
   cut (S siz1 + siz2 = S (siz1 + siz2));
    [ intro e; rewrite e; clear e | omega ].
     unfold ptoanynlist in |- *; fold ptoanynlist in |- *;
      getstar x (happend x0 h2); getland.
    apply disj_append_L; constructor; auto.
      apply disj_sym; auto.
    apply heap_eq_trans with (happend (happend x x0) h2);
     [ apply append_assoc; auto | apply heap_eq_append; auto ].
     apply disj_sym; auto.
     apply disj_sym; auto.
Save.


Lemma compose_MBlk_MBlk_MBlk: forall (h1 h2 h:heap) (rf :rfile) (siz1 siz2 :nat)
  (pr1 lnk1 pr2 lnk2 : addr),
  hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
  pr1 < pr2 -> pr2 = pr1 + siz1 ->
  MBlk (pr1 + 2) lnk1 siz1 (h1,rf) ->
  MBlk (pr2 + 2) lnk2 siz2 (h2,rf) ->
  MBlk (pr1 + 2) lnk1 (hread' h1 (pr1 + 1) + hread' h2 (pr2 + 1)) 
  (hwrite h (pr1 + 1) (hread' h1 (pr1 + 1) + hread' h2 (pr2 + 1)), rf).
Proof.
intros.
assert (hread' h1 (pr1 + 1) = siz1).
 inversion_clear H3; inversion_clear H6; splitstar1.
   cut (pr1 + 2 - 2 = pr1); [ intro e; rewrite e in H7; clear e | omega ].
   cut (pr1 + 2 - 1 = pr1 + 1); [ intro e; rewrite e in H7; clear e | omega ].
   unfold removeS in |- *; simpl in H7.
   assert (hread h1 (pr1 + 1) = Some siz1).
  unfold hfree in H7.
    rewrite <- H7; unfold hread in |- *; simpl in |- *.
    cut (false = EqNat.beq_nat pr1 (pr1 + 1));
     [ intro e; rewrite <- e; clear e; auto | apply beq_nat_neq; omega ].
  unfold hread' in |- *; rewrite H8; auto.
 assert (hread' h2 (pr2 + 1) = siz2).
  inversion_clear H4; inversion_clear H7; splitstar1.
    cut (pr2 + 2 - 2 = pr2); [ intro e; rewrite e in H8; clear e | omega ].
    cut (pr2 + 2 - 1 = pr2 + 1);
     [ intro e; rewrite e in H8; clear e | omega ].
    unfold removeS in |- *; simpl in H8.
    assert (hread h2 (pr2 + 1) = Some siz2).
   unfold hfree in H8.
     rewrite <- H8; unfold hread in |- *; simpl in |- *.
     cut (false = EqNat.beq_nat pr2 (pr2 + 1));
      [ intro e; rewrite <- e; clear e; auto | apply beq_nat_neq; omega ].
   unfold hread' in |- *; rewrite H9; auto.
  rewrite H5; clear H5; rewrite H6; clear H6.
    assert (ptoanynlist pr2 siz2 (h2, rf));
     [ apply MBlk_ptoanynlist with lnk2; auto | idtac ].
    inversion H3; inversion_clear H7; splitstar1.
    cut (pr1 + 2 - 2 = pr1);
     [ intro E; rewrite E in H7; rewrite E in H9; rewrite E in H11; clear E;
        auto
     | omega ].
    cut (pr1 + 2 - 1 = pr1 + 1);
     [ intro E; rewrite E in H9; rewrite E in H11; clear E; auto | omega ].
    unfold removeS in H11; simpl in H11.
    rewrite H2 in H5.
    assert
     (ptoanynlist (pr1 + 2) (siz1 - 2 + siz2)
        (happend (hfree (hfree h1 pr1) (pr1 + 1)) h2, rf)).
   apply Compose_ptoanynlist; auto.
    inversion_clear H.
      constructor; intros.
      elim H10 with a; intros; inversion_clear H; [ left | right ].
     constructor; unfold hfree in |- *.
       compare (pr1 + 1) a; intros.
      cut (true = EqNat.beq_nat (pr1 + 1) a);
       [ intro E; rewrite <- E; clear E; auto | apply beq_eq_true; auto ].
      cut (false = EqNat.beq_nat (pr1 + 1) a);
       [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; auto ].
        compare pr1 a; intros.
       cut (true = EqNat.beq_nat pr1 a);
        [ intro E; rewrite <- E; clear E; auto | apply beq_eq_true; auto ].
       cut (false = EqNat.beq_nat pr1 a);
        [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; auto ].
     constructor; auto.
    cut (pr1 + 2 + (siz1 - 2) = pr1 + siz1);
     [ intro E; rewrite E; clear E; auto | omega ].
   assert
    (ptoanynlist (pr1 + 2) (siz1 - 2 + siz2)
       (hfree (hfree h pr1) (pr1 + 1), rf)).
    apply
     h_subst_ptoanynlist with (happend (hfree (hfree h1 pr1) (pr1 + 1)) h2);
     auto.
      constructor; intros.
      unfold hfree in |- *; unfold happend in |- *.
      compare (pr1 + 1) i; intros.
     cut (true = EqNat.beq_nat (pr1 + 1) i);
      [ intro E; rewrite <- E; clear E; auto | apply beq_eq_true; auto ].
       assert (hnotindom i h2);
        [ apply disj_inl_notinr with h1; auto | inversion_clear H12; auto ].
       apply MblkPtrIn with rf pr1 lnk1 siz1; try omega; auto.
     cut (false = EqNat.beq_nat (pr1 + 1) i);
      [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; auto ].
       compare pr1 i; intros.
      cut (true = EqNat.beq_nat pr1 i);
       [ intro E; rewrite <- E; clear E; auto | apply beq_eq_true; auto ].
        assert (hnotindom i h2);
         [ apply disj_inl_notinr with h1; auto | inversion_clear H12; auto ].
        apply MblkPtrIn with rf pr1 lnk1 siz1; try omega; auto.
      cut (false = EqNat.beq_nat pr1 i);
       [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; auto ].
        inversion_clear H0; unfold happend in H12; apply H12.
    unfold MBlk in |- *; getland.
     omega.
     cut (pr1 + 2 - 2 = pr1); [ intro E; rewrite E; clear E; auto | auto ].
      cut (pr1 + 2 - 1 = pr1 + 1);
       [ intro E; rewrite E; clear E; auto | omega ].
        cut (siz1 + siz2 - 2 = siz1 - 2 + siz2);
         [ intro E; rewrite E; clear E; auto | auto ].
       getstar1.
        simpl in |- *; unfold hwrite in |- *; unfold hread in |- *.
          cut (false = EqNat.beq_nat (pr1 + 1) pr1);
           [ intro E; rewrite <- E; simpl in H7; unfold hread in H7; auto
           | apply beq_nat_neq; auto ].
         inversion_clear H0.
           elim H13 with pr1; unfold happend in |- *; rewrite H7; auto.
         omega.
        simpl in |- *.
          unfold hread in |- *; unfold hfree in |- *; unfold hwrite in |- *.
          cut (false = EqNat.beq_nat pr1 (pr1 + 1));
           [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; auto ].
         cut (true = EqNat.beq_nat (pr1 + 1) (pr1 + 1));
          [ intro E; rewrite <- E; auto | apply EqNat.beq_nat_refl; auto ].
         omega.
        simpl in |- *.
          apply h_subst_ptoanynlist with (hfree (hfree h pr1) (pr1 + 1));
           auto.
          constructor; intros.
          unfold hfree in |- *; unfold hwrite in |- *; compare (pr1 + 1) i;
           intros.
         cut (true = EqNat.beq_nat (pr1 + 1) i);
          [ intro E; rewrite <- E; clear E; auto | apply beq_eq_true; auto ].
         cut (false = EqNat.beq_nat (pr1 + 1) i);
          [ intro E; rewrite <- E; clear E; auto | apply beq_nat_neq; auto ].
       omega.
      omega.
Save.

Lemma append_cross : forall (h1 h2 h3 h4 :heap),
  hdisj h1 h2 -> hdisj h1 h3 -> hdisj h1 h4 -> hdisj h2 h3 ->hdisj h2 h4->hdisj h3 h4->
  heap_eq (happend  h1 (happend h4 (happend h2 h3))) 
          (happend (happend h1 (happend h3 h4))h2).
Proof.
unfold happend in |- *; intuition.
constructor; intros.
inversion_clear H; inversion_clear H0; inversion_clear H2; inversion_clear H1;
 inversion_clear H3; inversion_clear H4.
assert (hnotindom i h1 \/ hnotindom i h2); [ apply H5 | clear H5 ].
assert (hnotindom i h1 \/ hnotindom i h3); [ apply H | clear H ].
assert (hnotindom i h2 \/ hnotindom i h3); [ apply H0 | clear H0 ].
assert (hnotindom i h1 \/ hnotindom i h4); [ apply H2 | clear H2 ].
assert (hnotindom i h2 \/ hnotindom i h4); [ apply H1 | clear H1 ].
assert (hnotindom i h3 \/ hnotindom i h4); [ apply H3 | clear H3 ].
intuition; inversion_clear H4; inversion_clear H5; inversion_clear H;
 inversion_clear H0; inversion_clear H2; inversion_clear H1; 
 induction (h1 i); induction (h2 i); induction (h3 i); 
 induction (h4 i); try discriminate; auto.
Save.


Lemma h_subst_read :
  forall (h1 h2 : heap) (i : addr), heap_eq h1 h2 -> hread h1 i = hread h2 i.
Proof.
unfold hread;intros.
inversion_clear H;apply H0.
Save.

Lemma h_subst_read' :
  forall (h1 h2 : heap) (i : addr), heap_eq h1 h2 -> hread' h1 i = hread' h2 i.
Proof.
unfold hread';unfold hread;intros.
inversion_clear H.
assert (h1 i = h2 i);[apply H0|clear H0].
rewrite H;induction (h2 i);auto.
Save.

Lemma h_subst_write : forall(h1 h2:heap) (i:addr) ( n:hval),
 (heap_eq h1 h2) -> (heap_eq (hwrite h1 i n) (hwrite h2 i n)).
Proof.
intros.
constructor;intros.
inversion_clear H.
unfold hwrite.
compare i i0;intros.
cut (true=EqNat.beq_nat i i0);[intro E;rewrite <- E;auto|apply beq_eq_true;auto].
cut (false=EqNat.beq_nat i i0);[intro E;rewrite <- E;auto|apply beq_nat_neq;auto].
Save.


Lemma append_preserve_read'_R :
  forall (h1 h2 : heap) (i : addr),
    hnotindom i h1 -> hread' (happend h1 h2) i = hread' h2 i.
Proof.
intros.
cut (hread (happend h1 h2) i = hread h2 i);[intro e;unfold hread';rewrite e;auto|apply append_preserve_read_R;auto].
Save.

Lemma append_preserve_read'_L :
  forall (h1 h2 : heap) (i : addr),
    hnotindom i h2 -> hread' (happend h1 h2) i = hread' h1 i.
Proof.
intros.
cut (hread (happend h1 h2) i = hread h1 i);[intro e;unfold hread';rewrite e;auto|apply append_preserve_read_L;auto].
Save.


Lemma append_disj_write : forall (h1 h2 h: heap) (a : addr) (v :nat),
  hdisj h1 h2 -> heap_eq (happend h1 h2) h -> hindom a h2-> 
  heap_eq (hwrite h a v) (happend h1 (hwrite h2 a v)).
Proof.
intuition.
constructor;intros.
inversion_clear H0 .
assert (happend h1 h2 i = h i);[apply H2|clear H2].
unfold happend in H0;unfold happend.
assert (hnotindom a h1);[apply disj_inr_notinl with h2;auto|inversion_clear H2].
unfold hwrite.
compare a i;intros.
cut (true=EqNat.beq_nat a i);[intro E;rewrite <- E;simpl;clear E|apply beq_eq_true;auto].
rewrite e in H3;rewrite H3;auto.
cut (false=EqNat.beq_nat a i);[intro E;rewrite <- E;simpl;clear E|apply beq_nat_neq;auto].
induction (h1 i);auto.
Save.

Lemma MergHi : forall (h1 h2 h : heap) (rf : rfile) (bsiz tlen : nat) 
  (fl pr po hptr lnk : addr),
  hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
  hptr < po -> pr < hptr -> po = hptr + bsiz ->
  MBlk (hptr + 2) lnk bsiz (h2, rf) -> 
  star (EndL fl pr po) (MBlkLst tlen po pnil) (h1, rf) ->
  star (star (EndL fl pr po) 
             (MBlkLst (tlen - 1) (hread' h1 po) pnil)
       )
       (MBlk (hptr + 2) lnk (hread' h1 (po + 1) + hread' h2 (hptr + 1)))
       (hwrite h (hptr + 1) (hread' h1 (po + 1) + hread' h2 (hptr + 1)), rf).
Proof.
intros.
splitstar.
inversion H9.
 rewrite H12 in H1; inversion H1.
 assert
  (MBlk (hptr + 2) lnk (hread' h2 (hptr + 1) + hread' h3 (po + 1))
     (hwrite (happend h2 h3) (hptr + 1)
        (hread' h2 (hptr + 1) + hread' h3 (po + 1)), rf)).
  apply compose_MBlk_MBlk_MBlk with bsiz n p'; auto.
   assert (hdisj h2 x /\ hdisj h2 x0);
    [ apply disj_append_R; auto | inversion_clear H19 ].
    apply disj_heap_eq with h1;
     [ apply disj_sym; auto | apply heap_eq_sym; auto ].
    assert (hdisj h2 h3 /\ hdisj h2 h4);
     [ apply disj_append_R; auto | inversion_clear H19; auto ].
      apply disj_heap_eq with x0; [ auto | apply heap_eq_sym; auto ].
   apply heap_eq_refl.
  assert
   (star (MBlkLst m p' 0)
      (MBlk (hptr + 2) lnk (hread' h2 (hptr + 1) + hread' h3 (po + 1)))
      (happend h4
         (hwrite (happend h2 h3) (hptr + 1)
            (hread' h2 (hptr + 1) + hread' h3 (po + 1))), rf)).
   getstar h4
    (hwrite (happend h2 h3) (hptr + 1)
       (hread' h2 (hptr + 1) + hread' h3 (po + 1))); getland.
    assert (hdisj h2 h4).
     assert (hdisj h2 x /\ hdisj h2 x0);
      [ apply disj_append_R; auto | inversion_clear H20 ].
      apply disj_heap_eq with h1;
       [ apply disj_sym; auto | apply heap_eq_sym; auto ].
      assert (hdisj h2 h3 /\ hdisj h2 h4);
       [ apply disj_append_R; auto | inversion_clear H20; auto ].
        apply disj_heap_eq with x0; [ auto | apply heap_eq_sym; auto ].
     apply notin_write_preserve_disj_R.
      apply disj_append_L; constructor; apply disj_sym; auto.
      apply disj_inl_notinr with h2; auto.
        apply MblkPtrIn with rf hptr lnk bsiz; auto.
       omega.
       inversion H4; inversion_clear H22; omega.
    apply heap_eq_refl.
   apply StarAssocR.
     cut (S m - 1 = m); [ intro Eq; rewrite Eq; clear Eq | omega ].
     inversion H17.
     inversion_clear H22.
     splitstar1.
     cut (po + 2 - 2 = po); [ intro Eq; rewrite Eq in H22; clear Eq | omega ].
     simpl in H22.
     clear H26; clear H24; clear H23; clear H21.
     assert (hread (happend h3 h4) po = hread h3 po).
    apply append_preserve_read_L; auto.
      apply disj_inl_notinr with h3; auto.
      apply indom_some with p'; unfold hread in H22; auto.
    rewrite H22 in H21.
      cut (hread (happend h3 h4) po = hread x0 po);
       [ intro E; rewrite E in H21; clear E
       | inversion H13; unfold hread in |- *; apply H23 ].
      assert (hread (happend x x0) po = hread x0 po).
     apply append_preserve_read_R; auto.
       apply disj_inl_notinr with x0; auto.
      apply disj_sym; auto.
      inversion H21; apply indom_some with p'; auto.
     assert (hread h1 po = Some p').
      rewrite H21 in H23; inversion_clear H5.
        inversion_clear H23; unfold hread in |- *.
        symmetry  in |- *; apply H24; auto.
      cut (hread' h1 po = p');
       [ intro e; rewrite e; clear e
       | unfold hread' in |- *; rewrite H24; auto ].
        assert (hread' h1 (po + 1) = hread' h3 (po + 1)).
       assert (hread' h1 (po + 1) = hread' (happend x x0) (po + 1)).
        apply h_subst_read'; apply heap_eq_sym; auto.
        rewrite H25; clear H25.
          assert (hread' (happend x x0) (po + 1) = hread' x0 (po + 1)).
         apply append_preserve_read'_R.
           assert (hindom (po + 1) x0).
          apply h_subst_indom with (happend h3 h4); auto.
            apply append_preserve_indom_L.
            apply MblkPtrIn with rf po p' n; auto.
           omega.
           inversion H17; inversion_clear H25; omega.
          apply disj_inr_notinl with x0; auto.
         rewrite H25.
           assert (hread' (happend h3 h4) (po + 1) = hread' h3 (po + 1)).
          apply append_preserve_read'_L.
            assert (hindom (po + 1) h3).
           apply MblkPtrIn with rf po p' n; auto.
            omega.
            inversion H17; inversion_clear H26; omega.
           apply disj_inr_notinl with h3; [ apply disj_sym; auto | auto ].
          rewrite <- H26; apply h_subst_read'; apply heap_eq_sym; auto.
       rewrite H25.
         assert
          (hread' h2 (hptr + 1) + hread' h3 (po + 1) =
           hread' h3 (po + 1) + hread' h2 (hptr + 1));
          [ omega | rewrite <- H26 ].
         clear H21; clear H22; clear H23; clear H24; clear H25; clear H26.
         assert
          (heap_eq
             (hwrite (happend h4 (happend h2 h3)) (hptr + 1)
                (hread' h2 (hptr + 1) + hread' h3 (po + 1)))
             (happend h4
                (hwrite (happend h2 h3) (hptr + 1)
                   (hread' h2 (hptr + 1) + hread' h3 (po + 1))))).
        apply append_disj_write; auto.
         apply disj_append_L; constructor; apply disj_sym; auto.
           assert (hdisj h2 x /\ hdisj h2 x0);
            [ apply disj_append_R | inversion H21 ].
          apply disj_heap_eq with h1;
           [ apply disj_sym; auto | apply heap_eq_sym; auto ].
          assert (hdisj h2 h3 /\ hdisj h2 h4);
           [ apply disj_append_R | inversion H24; auto ].
            apply disj_heap_eq with x0; [ auto | apply heap_eq_sym; auto ].
         apply heap_eq_refl.
         apply append_preserve_indom_L.
           apply MblkPtrIn with rf hptr lnk bsiz; auto.
          omega.
          inversion H4; inversion_clear H21; omega.
        assert
         (star (MBlkLst m p' 0)
            (MBlk (hptr + 2) lnk (hread' h2 (hptr + 1) + hread' h3 (po + 1)))
            (hwrite (happend h4 (happend h2 h3)) (hptr + 1)
               (hread' h2 (hptr + 1) + hread' h3 (po + 1)), rf)).
         apply
          h_subst_star
           with
             (happend h4
                (hwrite (happend h2 h3) (hptr + 1)
                   (hread' h2 (hptr + 1) + hread' h3 (po + 1)))); 
          auto.
           apply heap_eq_sym; auto.
         getstar x
          (hwrite (happend h4 (happend h2 h3)) (hptr + 1)
             (hread' h2 (hptr + 1) + hread' h3 (po + 1))).
           clear H19; clear H20; clear H21.
           getland.
          assert (hdisj x h3 /\ hdisj x h4);
           [ apply disj_append_R; auto | inversion_clear H19 ].
           apply disj_heap_eq with x0; [ auto | apply heap_eq_sym; auto ].
           assert (hdisj h2 x /\ hdisj h2 x0);
            [ apply disj_append_R; auto | inversion_clear H19 ].
            apply disj_heap_eq with h1;
             [ apply disj_sym; auto | apply heap_eq_sym; auto ].
            apply notin_write_preserve_disj_R.
             apply disj_append_L; constructor; auto.
               apply disj_append_L; constructor;
                [ apply disj_sym; auto | auto ].
             assert (hindom (hptr + 1) h2).
              apply MblkPtrIn with rf hptr lnk bsiz; auto.
               omega.
               inversion H4; inversion_clear H21; omega.
              apply disj_inl_notinr with h2; auto.
          apply
           heap_eq_trans
            with
              (hwrite (happend x (happend h4 (happend h2 h3))) 
                 (hptr + 1) (hread' h2 (hptr + 1) + hread' h3 (po + 1))).
           apply heap_eq_sym; apply append_disj_write.
            assert (hdisj x h3 /\ hdisj x h4);
             [ apply disj_append_R; auto | inversion_clear H19 ].
             apply disj_heap_eq with x0; [ auto | apply heap_eq_sym; auto ].
             assert (hdisj h2 x /\ hdisj h2 x0);
              [ apply disj_append_R; auto | inversion_clear H19 ].
              apply disj_heap_eq with h1;
               [ apply disj_sym; auto | apply heap_eq_sym; auto ].
              apply disj_append_L; constructor; auto.
                apply disj_append_L; constructor;
                 [ apply disj_sym; auto | auto ].
            apply heap_eq_refl.
            apply append_perserve_indom_R.
             apply append_preserve_indom_L.
               apply MblkPtrIn with rf hptr lnk bsiz; auto.
              omega.
              inversion H4; inversion_clear H20; omega.
             apply disj_append_L; constructor; apply disj_sym; auto.
               assert (hdisj h2 x /\ hdisj h2 x0);
                [ apply disj_append_R; auto | inversion_clear H19 ].
              apply disj_heap_eq with h1;
               [ apply disj_sym; auto | apply heap_eq_sym; auto ].
              assert (hdisj h2 h3 /\ hdisj h2 h4);
               [ apply disj_append_R; auto | inversion_clear H19; auto ].
                apply disj_heap_eq with x0;
                 [ auto | apply heap_eq_sym; auto ].
           apply h_subst_write.
             assert (heap_eq (happend h1 h2) h);
              [ auto | apply heap_eq_trans with (happend h1 h2); auto ].
             assert (heap_eq (happend (happend x x0) h2) (happend h1 h2));
              [ apply heap_eq_append; auto
              | apply heap_eq_trans with (happend (happend x x0) h2); auto ].
             assert
              (heap_eq (happend (happend x (happend h3 h4)) h2)
                 (happend (happend x x0) h2));
              [ apply heap_eq_append; auto
              | apply
                 heap_eq_trans with (happend (happend x (happend h3 h4)) h2);
                 auto ].
            apply heap_eq_trans with (happend x0 x);
             [ auto | apply append_sym; apply disj_sym; auto ].
              apply heap_eq_trans with (happend (happend h3 h4) x);
               [ apply append_sym; apply disj_sym; auto | auto ].
             apply disj_sym; apply disj_heap_eq with x0; auto.
               apply heap_eq_sym; auto.
             apply heap_eq_append; auto.
            assert (hdisj h2 x /\ hdisj h2 x0);
             [ apply disj_append_R; auto | inversion_clear H23 ].
             apply disj_heap_eq with h1;
              [ apply disj_sym; auto | apply heap_eq_sym; auto ].
             assert (hdisj h2 h3 /\ hdisj h2 h4);
              [ apply disj_append_R; auto | inversion_clear H23 ].
              apply disj_heap_eq with x0; [ auto | apply heap_eq_sym; auto ].
              assert (hdisj x h3 /\ hdisj x h4);
               [ apply disj_append_R; auto | inversion_clear H23 ].
               apply disj_heap_eq with x0; [ auto | apply heap_eq_sym; auto ].
               apply append_cross; auto.
                 apply disj_sym; auto.
Save.


Lemma AppendGood : forall (h1 h2 h : heap) (rf : rfile) (alen asiz : nat)
  (fl tp lnk : addr) (gp : addr) (glen : nat),
  tp < gp -> tp <> pnil -> 
  hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
  MBlkLst (S glen) gp pnil (h1, rf) ->
  star (MBlkLst alen fl tp) (MBlk (tp + 2) lnk asiz) (h2, rf) ->
  Good fl (hwrite h tp gp, rf).
Proof.   
intros.
assert (EndL fl tp gp (hwrite h2 tp gp, rf)).
 constructor; intros.
  rewrite H5 in H0; elim H0; auto.
  getland.
    exists asiz; exists alen.
    splitstar.
    getstar x (hwrite x0 tp gp); getland.
   apply notin_write_preserve_disj_R; auto.
     apply disj_inr_notinl with x0; auto.
     apply MblkPtrIn with rf tp lnk asiz; try omega; auto.
     inversion_clear H9; inversion_clear H8; omega.
   constructor; intros.
     unfold hwrite in |- *; unfold happend in |- *.
     compare tp i; intros.
    cut (true = EqNat.beq_nat tp i);
     [ intro E; rewrite <- E; clear E | apply beq_eq_true; auto ].
      assert (hnotindom tp x);
       [ apply disj_inr_notinl with x0; auto | inversion_clear H8 ].
     apply MblkPtrIn with rf tp lnk asiz; try omega; auto.
       inversion_clear H9; inversion_clear H8; omega.
     rewrite e in H10; rewrite H10; auto.
    cut (false = EqNat.beq_nat tp i);
     [ intro E; rewrite <- E; clear E | apply beq_nat_neq; auto ].
      inversion_clear H4.
      elim H8 with i; unfold happend in |- *; tauto.
   replace (tp + 2) with (2 + tp); [ apply Write_MBlk_lnk with lnk | omega ].
     replace (2 + tp) with (tp + 2); [ auto | omega ].
 apply MidL_Good with tp gp.
   unfold MidL in |- *.
   exists (S glen).
   getstar (hwrite h2 tp gp) h1; getland.
  apply disj_sym; apply notin_write_preserve_disj_R; auto.
    apply disj_inr_notinl with h2; auto.
    splitstar.
    apply h_subst_indom with (happend x x0); auto.
    apply append_perserve_indom_R; auto.
    apply MblkPtrIn with rf tp lnk asiz; auto.
    inversion_clear H9; inversion_clear H8; intuition.
  constructor; intros.
    unfold happend in |- *; unfold hwrite in |- *; unfold hread in |- *.
    assert (heap_eq (happend h2 h1) h).
   apply heap_eq_trans with (happend h1 h2);
    [ apply append_sym; auto | auto ].
     apply disj_sym; auto.
   inversion_clear H6; elim H7 with i.
     compare tp i; intros.
    rewrite e; simpleq.
      apply EqNat.beq_nat_refl.
    simplneq.
      apply beq_nat_neq; auto.
Save.

Ltac writeindom :=
  match goal with
    | [ |- hindom ?j (hwrite ?h ?i ?v)] =>
      apply write_preserver_indom; auto
  end.

(* lll *)
Lemma hdisj_split_L :
  forall (h1 h2 : heap) (h3 h4 : heap),
    hdisj h1 h2 -> hdisj h3 h4 ->
    heap_eq (happend h3 h4) h1 -> (hdisj h2 h3 /\ hdisj h2 h4).
Proof.
intros.
apply disj_append_R.
apply disj_heap_eq with h1;
 [ apply disj_sym; auto | apply heap_eq_sym; auto ].
Save.


Ltac revdisj :=
  match goal with
    | [ |- hdisj ?h1 ?h2 ] =>
      assert (hdisj h2 h1); [idtac| apply disj_sym; assumption ]
  end.

Ltac nephewhdisj :=
  match goal with
    | [H1 : hdisj ?h1 ?h2, H2 : hdisj ?h3 ?h4, 
      H3 : heap_eq (happend ?h3 ?h4) ?h1 |- hdisj ?h2 ?h3] =>
    assert (hdisj h2 h3 /\ hdisj h2 h4); [apply hdisj_split_L with h1; auto | intuition]
    | [H1 : hdisj ?h1 ?h2, H2 : hdisj ?h3 ?h4, 
      H3 : heap_eq (happend ?h3 ?h4) ?h1 |- hdisj ?h2 ?h4] =>
    assert (hdisj h2 h3 /\ hdisj h2 h4); [apply hdisj_split_L with h1; auto | intuition]
  end.

(* lll *)
Lemma cross_happend :
  forall (h : heap) (h1 h2 : heap) (h3 h4 : heap),
    hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
    hdisj h3 h4 -> heap_eq (happend h3 h4) h1 ->
    heap_eq (happend h3 (happend h2 h4)) h.
Proof.
intros.
assert (hdisj h2 h3 /\ hdisj h2 h4);
 [ apply disj_append_R; auto | inversion_clear H3 ].
 apply disj_heap_eq with h1;
  [ apply disj_sym; auto | apply heap_eq_sym; auto ].
 apply heap_eq_trans with (happend h3 (happend h4 h2)).
  apply heap_eq_trans with (happend (happend h4 h2) h3);
   [ auto | apply append_sym; auto ].
   apply heap_eq_trans with (happend (happend h2 h4) h3);
    [ apply append_sym; auto | auto ].
    apply disj_append_L; constructor; auto.
      apply disj_sym; auto.
    apply heap_eq_append; apply append_sym; auto.
   apply disj_sym; apply disj_append_L; constructor; auto.
     apply disj_sym; auto.
  apply heap_eq_trans with (happend (happend h3 h4) h2);
   [ apply append_assoc; auto | auto ].
   apply disj_sym; auto.
   apply disj_sym; auto.
   apply heap_eq_trans with (happend h1 h2); auto.
     apply heap_eq_append; auto.
Save.

(* lll *)
(*Axiom EndL_pni_emp :
  forall (fl tl : addr) (h : heap) (rf : rfile),  
    EndL fl pnil tl (h, rf) -> h = hempty.*)

Lemma EndL_pni_emp' :
  forall (fl tl : addr) (h : heap) (rf : rfile),  
    EndL fl pnil tl (h, rf) -> heap_eq h hempty.
Proof.
intros.
inversion_clear H.
clear H1.
assert (MBlkLst 0 fl tl (h, rf)); [ apply H0; auto | clear H0 ].
inversion_clear H.
unfold emp in H0; simpl in H0; unfold hread in H0.
constructor; unfold hempty in |- *; auto.
Save.

Ltac fromendl0 :=
  match goal with
    | [ H1 : EndL ?fl 0 ?tl (?h, ?rf) |- _ ] =>
      assert (heap_eq h hempty); [apply EndL_pni_emp' with fl tl rf; auto| idtac]
  end.

Ltac revheq' H := 
  match type of H with
    heap_eq ?h1 ?h2 => assert (heap_eq h2 h1); [apply heap_eq_sym; trivial | clear H]
  end.

Ltac heqtrans :=
  match goal with
    | [ H1 : heap_eq ?h1 ?h2, H2 : heap_eq ?h2 ?h3 |- heap_eq ?h1 ?h3] =>
      apply heap_eq_trans with h2; auto
  end.


(** assertions on equal heaps **)

Lemma h_subst_star : forall (h1 h2 : heap) (rf : rfile) (a1 a2 : assert),
  star a1 a2 (h1, rf) -> heap_eq h1 h2 -> star a1 a2 (h2, rf).
Proof.
intros.
splitstar.
getstar x x0;getland.
apply heap_eq_trans with h1; auto.
Save.

Ltac aeqh :=
  match goal with
    | [ H1 : heap_eq ?h1 ?h2, H2 : star ?a1 ?a2 (?h1, ?rf) 
      |- star ?a1 ?a2 (?h2, ?rf)] => apply h_subst_star with h1; auto
  end.


Lemma MblkPtrNotin : forall (h : heap) (rf :rfile) (hptr lnk po : addr) (siz :nat),
   MBlk (hptr + 2) lnk siz (h, rf) ->
   hnotindom po h -> po < hptr \/ po >= hptr + siz.
Proof.
intros.
assert (po >= hptr -> po < hptr + siz -> False).
 intros.
   assert (hindom po h);
    [ apply MblkPtrIn with rf hptr lnk siz; auto | idtac ].
   inversion_clear H0; inversion_clear H3; rewrite H4 in H0; discriminate.
 omega.
Save.


Lemma MBlkLoNotOverlap : forall (h1 h2 : heap) (rf : rfile) 
  (lptr tp bptr lnk : addr) (siz len : nat),
  hdisj h1 h2 -> bptr < lptr -> 
  MBlk (bptr + 2) lnk siz (h1, rf) ->
  MBlkLst (S len) lptr tp (h2, rf) ->
  bptr + siz <= lptr.
Proof.
intros.
assert (lptr < bptr \/ lptr >= bptr + siz);
 [ apply MblkPtrNotin with h1 rf lnk; auto | omega ].
apply disj_inr_notinl with h2; [ exact H | idtac ].
inversion_clear H2.
apply h_subst_indom with (happend h0 h3); auto.
apply append_preserve_indom_L.
apply MblkPtrIn with rf lptr p' n; auto.
inversion_clear H6; inversion_clear H2; omega.
Save.

      
Ltac mblklonotoverlap := 
  match goal with
    | [ H1 : MBlk (?bptr + 2) ?lnk ?siz (?h1, ?rf),
      H2 : MBlkLst (S ?len) ?lptr ?tp (?h2, ?rf) |- _] =>
    apply MBlkLoNotOverlap with h1 h2 rf tp lnk len; auto
  end.

Lemma MBlk_Write_Lnk : forall (h : heap) (rf : rfile) (siz : nat) 
  (lnk lnk' : nat) (hptr : addr),
  MBlk (hptr + 2) lnk siz (h, rf) ->
  MBlk (hptr + 2) lnk' siz (hwrite h hptr lnk', rf).
Proof.
intros.
replace (hptr + 2) with (2 + hptr); [ idtac | omega ].
apply Write_MBlk_lnk with lnk.
replace (hptr + 2) with (2 + hptr) in H; [ exact H | omega ].
Save.



Ltac mblklnk :=
  match goal with
    | [H1 : MBlk (?hptr + 2) ?lnk ?siz (?h, ?rf) |-
      MBlk (?hptr + 2) ?lnk' ?siz (hwrite ?h ?hptr ?lnk', ?rf)] =>
    apply MBlk_Write_Lnk with lnk; auto
  end.

Lemma MBlkLst_MBlk_MBlkLst :
  forall (len : nat) (lptr : addr) (tptr : addr) (h1 h2 h : heap) (rf : rfile)
    (lnk : addr) (siz : nat),
    hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
    (tptr < lnk \/ lnk = pnil) ->
    MBlkLst len lptr tptr (h1, rf) ->
    MBlk (tptr + 2) lnk siz (h2, rf) ->
    MBlkLst (S len) lptr lnk (h, rf).
Proof.
intros.
apply Compose_MBlkLst_MBlk_to_MBlkLst with tptr siz;
 [ getstar h1 h2; getland | exact H1 ].
replace (tptr + 2) with (2 + tptr) in H3; [ exact H3 | omega ].
Save.

Lemma MBlkLst3' :
  forall (len2 : nat), 
    forall (h1 h2 h : heap) (len1 : nat) (lptr1 lptr2 : addr) (rf : rfile),
      hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
      MBlkLst len1 lptr1 lptr2 (h1, rf) ->
      MBlkLst len2 lptr2 pnil (h2, rf) ->
      MBlkLst (len1 + len2) lptr1 pnil (h, rf).
Proof.
induction len2.
 intros.
   inversion H2.
   unfold emp in H3; simpl in H3; unfold hread in H3.
   apply h_subst_MBlkLst with (happend h1 h2); auto.
   rewrite H6 in H1; replace (len1 + 0) with len1; [ idtac | auto ].
   apply h_subst_MBlkLst with h1; auto.
   apply heap_eq_trans with (happend h2 h1);
    [ unfold happend in |- *; constructor; intros
    | apply append_sym; apply disj_sym; auto ].
   assert (h2 i = None); [ apply H3 | rewrite H4; auto ].
 intros.
   inversion_clear H2.
   assert (hdisj h1 h3 /\ hdisj h1 h4);
    [ apply disj_append_R; auto | inversion_clear H2 ].
  apply disj_heap_eq with h2; [ auto | apply heap_eq_sym; auto ].
  replace (len1 + S len2) with (S len1 + len2);
   [ apply IHlen2 with (happend h1 h3) h4 p'; auto | omega ].
   apply disj_sym; apply disj_append_L; constructor; apply disj_sym; auto.
   apply heap_eq_trans with (happend h1 h2); auto.
     apply heap_eq_trans with (happend h1 (happend h3 h4)).
    apply heap_eq_sym; apply append_assoc; auto.
    apply heap_eq_trans with (happend (happend h3 h4) h1);
     [ apply append_sym; auto | auto ].
     apply disj_heap_eq with h2; [ auto | apply heap_eq_sym; auto ].
     apply heap_eq_trans with (happend h2 h1);
      [ auto | apply append_sym; apply disj_sym; auto ].
       apply heap_eq_append; auto.
   apply Compose_MBlkLst_MBlk_to_MBlkLst with lptr2 n; auto.
     getstar h1 h3; getland.
    apply heap_eq_refl.
    replace (2 + lptr2) with (lptr2 + 2); [ exact H6 | omega ].
Save.

Lemma MBlkLst3 :
  forall (h1 h2 h : heap) (len1 len2 : nat) (lptr1 lptr2 : addr) (rf : rfile),
    hdisj h1 h2 -> heap_eq (happend h1 h2) h ->
    MBlkLst len1 lptr1 lptr2 (h1, rf) ->
    MBlkLst len2 lptr2 pnil (h2, rf) ->
    MBlkLst (len1 + len2) lptr1 pnil (h, rf).
Proof.
intros.
apply MBlkLst3' with h1 h2 lptr2;auto.
Save.


Ltac mblk3 :=
  match goal with 
    | [H1 : MBlk (?pr1 + 2) ?lnk1 ?siz1 (?h1, ?rf), H2 : MBlk _ ?lnk2 ?siz2 (?h2, ?rf)
      |- MBlk (?pr1 + 2) ?lnk1 _ (_, ?rf)] =>
    apply compose_MBlk_MBlk_MBlk with siz1 siz2 lnk2; auto
  end.

Lemma EndL_Po_gt_Pr : forall (h : heap) (rf : rfile) (fv pr po : addr),
  EndL fv pr po (h, rf) -> pr > 0 -> po > 0 -> po > pr.
Proof.
intros.
inversion_clear H; clear H2.
elim H3; [ intros; clear H3 | inversion H0; auto ].
inversion_clear H; auto.
rewrite H3 in H1; inversion H1.
Save.


Lemma MBLkLst_EndL_Good :
  forall (h1 h2 h : heap) (rf : rfile) (fv lp pr po : addr) (len : nat),
    hdisj h1 h2 ->
    heap_eq (happend h1 h2) h ->
    MBlkLst len lp pnil (h1, rf) ->
    EndL fv pr po (h2, rf) ->
    lp > pr \/ lp = pnil ->
    pr > pnil ->
    Good fv (hwrite h pr lp, rf).
Proof.
intros.
apply MidL_Good with pr lp.
unfold MidL in |- *.
assert (hdisj (hwrite h2 pr lp) h1).
 apply in_write_preserve_disj_L;
  [ apply disj_sym; auto | apply EndLIndom with rf fv po; auto ].
   inversion H4; auto.
 exists len; getstar (hwrite h2 pr lp) h1; getland.
  apply heap_eq_trans with (happend h1 (hwrite h2 pr lp));
   [ apply append_sym; auto | apply heap_eq_sym ].
    apply append_disj_write; auto.
    apply EndLIndom with rf fv po; auto.
    inversion H4; auto.
  inversion_clear H2.
    elim H7; [ intros; clear H7 | inversion H4; auto ].
    unfold EndL in |- *; getland; intros.
   rewrite H7 in H4; inversion H4.
   getland.
     inversion_clear H8; inversion_clear H9.
     exists x; exists x0.
     splitstar.
     assert (hindom (pr + 0) x2 /\ hindom (pr + 1) x2).
    apply MBlkInDom with po x rf; auto.
    inversion_clear H11.
      replace (pr + 0) with pr in H13; [ idtac | auto with arith ].
      getstar x1 (hwrite x2 pr lp); getland.
     apply disj_sym; apply in_write_preserve_disj_L;
      [ apply disj_sym; auto | auto ].
     apply heap_eq_sym; apply append_disj_write; auto.
     apply MBlk_Write_Lnk with po; auto.
Save.

Lemma EndL_MBlkLst_Good :
  forall (h1 h2 h : heap) (rf : rfile) (fv pr po : addr) (len : nat),
    hdisj h1 h2 ->
    heap_eq (happend h1 h2) h ->
    EndL fv pr po (h1, rf) ->
    MBlkLst len po pnil (h2, rf) ->
    Good fv (h, rf).
Proof.
intros.
apply MidL_Good with pr po.
unfold MidL in |- *.
exists len.
getstar h1 h2; auto.
Save.

Ltac elg :=
  match goal with
    | [ H1 : EndL ?fv ?pr ?po (?h1, ?rf), H2 : MBlkLst ?len ?po pnil (?h2, ?rf)
      |- _] =>
    apply EndL_MBlkLst_Good with h1 h2 pr po len; auto
  end.

Lemma h_subst_ptoany : forall (h h' :heap) (rf :rfile) (hptr :addr),
  heap_eq h h' -> ptoany hptr (h, rf) -> ptoany hptr (h', rf).
Proof.
intros.
inversion_clear H0.
unfold ptoany in |- *; exists x.
inversion_clear H1.
simpl in H0.
constructor; intros.
 simpl in |- *; rewrite <- H0; apply h_subst_read; apply heap_eq_sym; exact H.
 elim H2 with j; auto.
   simpl in |- *; apply h_subst_read; apply heap_eq_sym; exact H.
Save.

Lemma pnal_sep_tail : forall (siz : nat),
  forall  (h : heap) (rf : rfile) (lnk : addr) (hptr : addr),
  ptoanynlist hptr (S siz) (h,rf) -> 
  star (ptoanynlist hptr siz) 
  (ptoany (hptr + siz)) (h, rf).
Proof.
induction siz; intros.
 getstar hempty h; getland.
  apply disj_emp.
  apply append_emp.
  unfold ptoanynlist in |- *; unfold emp in |- *; intros; simpl in |- *;
   unfold hread in |- *; unfold hempty in |- *; tauto.
  inversion_clear H; inversion_clear H0; intuition.
    replace (hptr + 0) with hptr; [ idtac | omega ].
    apply h_subst_ptoany with x; auto.
    unfold emp in H3; simpl in H3; unfold hread in H3.
    apply heap_eq_trans with (happend x0 x).
   apply heap_eq_sym; apply append_nullheap.
     constructor; unfold hempty in |- *; symmetry  in |- *; apply H3.
   apply heap_eq_trans with (happend x x0);
    [ apply append_sym; apply disj_sym; exact H0 | exact H ].
 unfold ptoanynlist in |- *; fold ptoanynlist in |- *.
   apply StarAssocR.
   unfold ptoanynlist in H; fold ptoanynlist in H.
   splitstar.
   getstar x x0; getland.
   replace (hptr + S siz) with (S hptr + siz); [ apply IHsiz; auto | omega ].
Save.

Lemma Sep_pnal : forall (h : heap) (rf : rfile) (siz siz0 siz1 : nat)
  (lnk : addr) (hptr : addr),
  ptoanynlist hptr siz (h,rf) -> 
  siz = siz0 + siz1 ->
  star (ptoanynlist hptr siz1) 
  (ptoanynlist (hptr + siz1) siz0) (h, rf).
Proof.
induction siz0; intros.
 simpl in H0; subst siz1.
   apply StarSym; getstar hempty h; getland.
  apply disj_emp.
  apply append_emp.
  unfold ptoanynlist in |- *; unfold emp in |- *; intros; simpl in |- *;
   unfold hread in |- *; unfold hempty in |- *; tauto.
 unfold ptoanynlist in |- *; fold ptoanynlist in |- *.
   apply StarAssocL.
   replace (S (hptr + siz1)) with (hptr + S siz1); [ idtac | omega ].
   assert
    (star (ptoanynlist hptr (S siz1)) (ptoanynlist (hptr + S siz1) siz0)
       (h, rf)); [ apply IHsiz0; auto | clear IHsiz0 ].
  omega.
  splitstar.
    getstar x x0; getland.
    apply pnal_sep_tail; auto.
Save.

Lemma notin_part_L :
  forall (h1 h2 h : heap) (p :addr),
    hnotindom p h -> heap_eq (happend h1 h2) h -> hnotindom p h1.
Proof.
intros.
assert (hindom p h1 \/ hnotindom p h1);[apply indom_or_notin|idtac].
intuition.
inversion_clear H2;inversion_clear H;inversion_clear H0.
assert (happend h1 h2 p = h p);[apply H|clear H].
unfold happend in H0;rewrite H1 in H0;rewrite H2 in H0;discriminate.
Save.


Axiom h_subst_assert : forall (a : assert) (h h' : heap) (rf :rfile),
  a (h, rf) -> heap_eq h h' -> a (h', rf).


Lemma star1_star :forall (a1 a2:assert) (p : addr) (n : nat) (h : heap) (rf :rfile),
  star1 p n (star a1 a2) (h, rf) -> star (star1 p n a1) a2 (h, rf).
Proof.
intros.
splitstar1.
unfold removeS in H1.
splitstar.
getstar (happend (hsingleton p n) x) x0; getland.
 apply disj_sym; apply disj_append_L.
   constructor.
  apply disj_notboth; intros.
    simpl in H0; unfold hread in H0.
    compare p a; [ intros; left | intros; right ].
   apply notin_part_L with x (hfree h p).
    constructor; unfold hfree in |- *.
      subst a; simpleq; apply EqNat.beq_nat_refl.
    apply heap_eq_trans with (happend x x0); [ apply append_sym | auto ].
      apply disj_sym; auto.
   constructor; unfold hsingleton in |- *.
     simplneq; apply beq_nat_neq; auto.
  apply disj_sym; auto.
 apply heap_eq_trans with (happend (hsingleton p n) (happend x x0)).
  assert (hdisj (hsingleton p n) x /\ hdisj (hsingleton p n) x0).
   apply disj_append_R; apply disj_heap_eq with (hfree h p);
    [ idtac | apply heap_eq_sym; auto ].
     apply disj_notboth; intros.
     simpl in H0; unfold hread in H0.
     compare p a; [ intros; right | intros; left ].
    constructor; unfold hfree in |- *.
      subst a; simpleq; apply EqNat.beq_nat_refl.
    constructor; unfold hsingleton in |- *.
      simplneq; apply beq_nat_neq; auto.
   inversion_clear H3.
     apply heap_eq_sym; apply append_assoc; auto.
  apply heap_eq_trans with (happend (hfree h p) (hsingleton p n)).
   apply heap_eq_trans with (happend (happend x x0) (hsingleton p n)).
    apply append_sym.
      apply disj_heap_eq with (hfree h p).
     apply disj_notboth; intros.
       simpl in H0; unfold hread in H0.
       compare p a; [ intros; right | intros; left ].
      constructor; unfold hfree in |- *.
        subst a; simpleq; apply EqNat.beq_nat_refl.
      constructor; unfold hsingleton in |- *.
        simplneq; apply beq_nat_neq; auto.
     apply heap_eq_sym; auto.
    apply heap_eq_append; auto.
   constructor; intros.
     unfold happend in |- *; unfold hsingleton in |- *; unfold hfree in |- *.
     compare p i; intros.
    subst i; simpleq; apply EqNat.beq_nat_refl.
    simplneq; [ induction (h i); auto | apply beq_nat_neq; auto ].
 getstar1.
  simpl in |- *; unfold hread in |- *; unfold happend in |- *;
   unfold hsingleton in |- *.
    simpleq; apply EqNat.beq_nat_refl.
  unfold removeS in |- *.
    apply h_subst_assert with x; auto.
    constructor; intros.
    simpl in H0; unfold hread in H0.
    unfold happend in |- *; unfold hsingleton in |- *; unfold hfree in |- *.
    compare p i; intros.
   subst i; simpleq.
    assert (hdisj (hsingleton p n) x /\ hdisj (hsingleton p n) x0).
     apply disj_append_R; apply disj_heap_eq with (hfree h p);
      [ idtac | apply heap_eq_sym; auto ].
       apply disj_notboth; intros.
       simpl in H0; unfold hread in H0.
       compare p a; [ intros; right | intros; left ].
      constructor; unfold hfree in |- *.
        subst a; simpleq; apply EqNat.beq_nat_refl.
      constructor; unfold hsingleton in |- *.
        simplneq; apply beq_nat_neq; auto.
     inversion_clear H3.
       assert (hnotindom p x);
        [ apply disj_inl_notinr with (hsingleton p n); auto
        | inversion H3; auto ].
       unfold hsingleton in |- *.
       apply indom_some with n.
       simpleq; apply EqNat.beq_nat_refl.
    apply EqNat.beq_nat_refl.
   simplneq; apply beq_nat_neq; auto.
Save.
Lemma star1_strengthen : forall (a1 a2:assert) (p : addr) (n : nat) (h : heap) (rf :rfile),
  star1 p n a1 (h, rf) -> (forall (h :heap) (rf :rfile), a1 (h, rf) -> a2 (h, rf)) -> star1 p n a2 (h, rf).
Proof.
intros.
splitstar1.
unfold removeS in H2.
assert (a2 (hfree h p, rf)).
 apply H0;auto.
 getstar1.
Save.

Lemma star1_star1_star :forall (a1 a2:assert) (p1 p2: addr) (n1 n2: nat) (h : heap) (rf :rfile),
  star1 p1 n1 (star1 p2 n2 (star a1 a2)) (h, rf) -> 
  star (star1 p1 n1 (star1 p2 n2 a1)) a2 (h, rf).
Proof.
intros.
assert (star1 p1 n1 (star (star1 p2 n2 a1) a2) (h, rf)).
apply star1_strengthen with (star1 p2 n2 (star a1 a2));[auto|apply star1_star].
apply star1_star;auto.
Save.



Lemma Sep_MBlk : forall (h : heap) (rf : rfile) (siz siz0 siz1 : nat)
  (lnk : addr) (hptr : addr),
  MBlk (hptr + 2) lnk siz (h, rf) -> 
  siz = siz0 + siz1 ->
  siz1 > 2 ->
  star (MBlk (hptr + 2) lnk siz1) 
  (ptoanynlist (hptr + siz1) siz0) (hwrite h (hptr + 1) siz1, rf).
Proof.
intros.
inversion_clear H; inversion_clear H3; splitstar1.
cut (hptr + 2 - 2 = hptr);
 [ intro e; rewrite e in H4; rewrite e in H6; rewrite e in H3; clear e
 | omega ].
cut (hptr + 2 - 1 = hptr + 1);
 [ intro e; rewrite e in H4; rewrite e in H6; clear e | omega ].
unfold removeS in H4; unfold removeS in H6.
assert
 (star (ptoanynlist (hptr + 2) (siz1 - 2))
    (ptoanynlist (hptr + 2 + (siz1 - 2)) siz0)
    (hfree (hfree h hptr) (hptr + 1), rf));
 [ apply Sep_pnal with (siz - 2); auto; try omega | idtac ].
assert
 (star1 hptr lnk
    (star1 (hptr + 1) siz
       (star (ptoanynlist (hptr + 2) (siz1 - 2))
          (ptoanynlist (hptr + 2 + (siz1 - 2)) siz0))) 
    (h, rf)).
 getstar1.
 assert
  (star1 hptr lnk
     (star1 (hptr + 1) siz1
        (star (ptoanynlist (hptr + 2) (siz1 - 2))
           (ptoanynlist (hptr + 2 + (siz1 - 2)) siz0)))
     (hwrite h (hptr + 1) siz1, rf)).
  simpl in H3; simpl in H4; unfold hread in H3; unfold hread in H4;
   unfold hfree in H4.
    cut (false = EqNat.beq_nat hptr (hptr + 1));
     [ intro e; rewrite <- e in H4; clear e | apply beq_nat_neq; omega ].
    getstar1.
   simpl in |- *; unfold hread in |- *; unfold hwrite in |- *.
     simplneq; apply beq_nat_neq; omega.
   simpl in |- *; unfold hread in |- *; unfold hfree in |- *.
     simplneq; [ unfold hwrite in |- * | apply beq_nat_neq; omega ].
     simpleq; apply EqNat.beq_nat_refl.
   unfold removeS in |- *.
     apply h_subst_star with (hfree (hfree h hptr) (hptr + 1)); auto.
     constructor; intros.
     unfold hfree in |- *; unfold hwrite in |- *.
     compare (hptr + 1) i; intros.
    subst i; simpleq; apply EqNat.beq_nat_refl.
    simplneq; apply beq_nat_neq; auto.
  clear H7; clear H5; clear H3; clear H4; clear H6.
    assert
     (star
        (star1 hptr lnk
           (star1 (hptr + 1) siz1 (ptoanynlist (hptr + 2) (siz1 - 2))))
        (ptoanynlist (hptr + 2 + (siz1 - 2)) siz0)
        (hwrite h (hptr + 1) siz1, rf)).
   apply star1_star1_star; auto.
   clear H8.
     splitstar.
     getstar x x0; getland.
    unfold MBlk in |- *; getland.
      replace (hptr + 2 - 2) with hptr; [ idtac | omega ].
      replace (hptr + 2 - 1) with (hptr + 1); [ exact H5 | omega ].
    replace (hptr + 2 + (siz1 - 2)) with (hptr + siz1) in H7;
     [ exact H7 | omega ].
Save.

Ltac pnalindom :=
  match goal with
    | [ H : ptoanynlist ?m ?len (?h, ?rf) |- _] =>
      apply ListInDom with len m rf; auto
  end.

Lemma Pnal_MBlk : forall (h : heap) (rf : rfile) (len : nat) (hptr : addr),
  ptoanynlist hptr len (h, rf) ->
  hptr > pnil ->
  len > 2 ->
  MBlk (hptr + 2) (hread' h hptr) len (hwrite h (hptr + 1) len, rf).
Proof.
intros.
cut (len = S (S (len - 2))); [ intro e; rewrite e in H; clear e | omega ].
unfold ptoanynlist in H; fold ptoanynlist in H.
replace (S (S hptr)) with (hptr + 2) in H; [ splitstar | omega ].
splitstar.
constructor.
 omega.
 constructor; auto.
   replace (hptr + 2 - 2) with hptr; [ idtac | omega ].
   replace (hptr + 2 - 1) with (hptr + 1); [ idtac | omega ].
   getstar1.
  inversion_clear H3; inversion_clear H7.
    simpl in H3.
    assert (hread h hptr = Some x3).
   rewrite <- H3.
     assert (hread (happend x x0) hptr = hread h hptr).
    apply h_subst_read; auto.
    rewrite <- H7.
      apply append_preserve_read_L.
      apply disj_inl_notinr with x; auto.
      apply indom_some with x3; unfold hread in H3; auto.
   unfold hread' in |- *; rewrite H7.
     simpl in |- *; unfold hread in |- *; unfold hwrite in |- *.
     simplneq; apply beq_nat_neq; omega.
  unfold removeS in |- *; simpl in |- *; unfold hread in |- *;
   unfold hfree in |- *; unfold hwrite in |- *.
    simplneq; [ idtac | apply beq_nat_neq; omega ].
    simpleq; apply EqNat.beq_nat_refl.
  unfold removeS in |- *.
    apply h_subst_ptoanynlist with x2; auto.
    constructor; intros.
    unfold hfree in |- *; unfold hwrite in |- *.
    compare (hptr + 1) i; intros.
   subst i; simpleq.
    assert (hnotindom (hptr + 1) x2).
     apply disj_inl_notinr with x1; auto.
       inversion_clear H6; inversion_clear H7.
       simpl in H6; unfold hread in H6.
       replace (hptr + 1) with (S hptr);
        [ apply indom_some with x3; auto | omega ].
     inversion_clear H7; auto.
    apply EqNat.beq_nat_refl.
   simplneq; [ idtac | apply beq_nat_neq; auto ].
     compare hptr i; intros.
    subst i; simpleq.
     assert (hnotindom hptr x2).
      assert (hnotindom hptr x0).
       apply disj_inl_notinr with x; auto.
         inversion_clear H3; inversion_clear H7.
         simpl in H3; unfold hread in H3.
         apply indom_some with x3; auto.
       assert (heap_eq (happend x2 x1) x0).
        apply heap_eq_trans with (happend x1 x2); [ apply append_sym | auto ].
          apply disj_sym; auto.
        apply notin_part_L with x1 x0; auto.
      inversion_clear H7; auto.
     apply EqNat.beq_nat_refl.
    simplneq.
     inversion_clear H5; inversion_clear H.
       assert (happend x1 x2 i = x0 i); [ apply H7 | clear H7 ].
       assert (happend x x0 i = h i); [ apply H5 | clear H5 ].
       inversion_clear H3; inversion_clear H6.
       inversion_clear H5; inversion_clear H3.
       simpl in H9; simpl in H10; unfold hread in H9; unfold hread in H10.
       assert (x i = None); [ apply H9; auto | clear H9 ].
       assert (x1 i = None); [ apply H10; omega | clear H10 ].
       unfold happend in H; unfold happend in H7.
       rewrite H9 in H.
       rewrite H3 in H7.
       rewrite H; rewrite <- H7; tauto.
     apply beq_nat_neq; auto.
Save.

