Require Import natutil.
Require Import heap2.
Require Import scap.
Require Import scap_rules.
Require Import Arith.

(** Program Lemma *)
(** if Psi |- C : Psi ) and l in dom(Psi) then l in dom(C) **)

Lemma exists_iseq_0 : 
  forall (ch ch0 : codeheap) (ct0 ct : codeheapty)
    (l : lab) (theta : codespec),
    WF_CH_0 ct0 ct ch -> lookupCT ct l theta -> exists is : iseq, lookupCH ch l is.
Proof.
induction ch.
intros.
inversion H.
rewrite <- H2 in H0.
inversion H0.
intros.
assert (n = l \/ n <> l).
apply eqorneq_nat.
inversion_clear H1.
rewrite H2.
exists i.
constructor.
inversion H.
assert (exists is : iseq, lookupCH ch l is).
apply IHch with ct0 ct2 theta.
auto.
auto.
rewrite H1 in H5.
rewrite <- H5 in H0.
inversion H0.
contradiction.
auto.
inversion_clear H9.
exists x.
constructor; auto.
Save.

Lemma exists_iseq :
  forall (ch : codeheap) (ct : codeheapty) (l : lab) (theta : codespec),
    WF_CH ct ch -> lookupCT ct l theta -> exists is : iseq, lookupCH ch l is.
Proof.
intros.
inversion_clear H.
apply exists_iseq_0 with ct ct theta.
auto.
auto.
auto.
Save.


(** Progress Lemma *)
Lemma progress :
  forall (s : state) (ch : codeheap) (is : iseq)
    (ct : codeheapty) (a : interp),
    WF_WLD ct a (s, (ch, is)) ->
    exists w : world, step (s, (ch, is)) w.
Proof.
intros s ch is ct a wfw.
inversion_clear wfw.
inversion_clear H0.

(* j f *)
assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct s).
apply H2; auto.
inversion_clear H0.
inversion_clear H3.
assert (exists is : iseq, lookupCH ch f is).
apply exists_iseq with ct x; auto.
inversion_clear H3.
exists (s, (ch, x0)).
induction s.
constructor; auto.

(* jal f, fret *)
induction s as [h rf].
assert (exists theta : codespec,
         lookupCT ct f theta /\ itpx1 theta ct (h, updateR rf ra fret)).
apply H2; auto.
inversion_clear H0.
inversion_clear H3.
assert (exists is : iseq, lookupCH ch f is).
apply exists_iseq with ct x; auto.
inversion_clear H3.
exists ((h, updateR rf ra fret), (ch, x0)).
constructor; auto.

(* jalr rd rs fret *)
induction s as [h rf].
assert (exists theta : codespec,
         lookupCT ct (lookupR rf rs) theta /\
         itpx1 theta ct (h, updateR rf rd fret)).
apply H2; auto.
inversion_clear H0.
inversion_clear H3.
assert (exists is : iseq, lookupCH ch (lookupR rf rs) is).
apply exists_iseq with ct x; auto.
inversion_clear H3.
exists ((h, updateR rf rd fret), (ch, x0)).
constructor; auto.

(* jr rs *)
induction s as [h rf].
assert (exists theta : codespec,
  lookupCT ct (lookupR rf rs) theta /\
  itpx1 theta ct (h, rf)).
apply H2; auto.
inversion_clear H0.
inversion_clear H3.
assert (exists is : iseq, lookupCH ch (lookupR rf rs) is).
apply exists_iseq with ct x; auto.
inversion_clear H3.
exists ((h, rf), (ch, x0)).
constructor; auto.

(* beq rs rt f *)
induction s as [h rf].
assert ((lookupR rf rs <> lookupR rf rt -> a' ct (h, rf)) /\
       (lookupR rf rs = lookupR rf rt ->
        exists theta : codespec,
          lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H2; auto.
assert (lookupR rf rs = lookupR rf rt \/ lookupR rf rs <> lookupR rf rt).
apply eqorneq_nat.
inversion_clear H0.
inversion_clear H4.
  (*case rs = rt *)
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H6; auto.
  inversion_clear H4.
  inversion_clear H7.
  assert (exists is : iseq, lookupCH ch f is).
  apply exists_iseq with ct x; auto.
  inversion_clear H7.
  exists ((h, rf), (ch, x0)).
  constructor; assumption.
  (*case rs <> rt *)
  exists ((h, rf), (ch, is0)).
  constructor; assumption.

(* bne rs rt f *)
induction s as [h rf].
assert ((lookupR rf rs = lookupR rf rt -> a' ct (h, rf)) /\
       (lookupR rf rs <> lookupR rf rt ->
        exists theta : codespec,
          lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H2; auto.
assert (lookupR rf rs = lookupR rf rt \/ lookupR rf rs <> lookupR rf rt).
apply eqorneq_nat.
inversion_clear H0.
inversion_clear H4.
  (*case rs <> rt *)
  exists ((h, rf), (ch, is0)).
  constructor; assumption.
  (*case rs = rt *)
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H6; auto.
  inversion_clear H4.
  inversion_clear H7.
  assert (exists is : iseq, lookupCH ch f is).
  apply exists_iseq with ct x; auto.
  inversion_clear H7.
  exists ((h, rf), (ch, x0)).
  constructor; assumption.

(* bgez rs f *)
induction s as [h rf].
assert ((lookupR rf rs < 0 -> a' ct (h, rf)) /\
       (lookupR rf rs >= 0 ->
        exists theta : codespec,
          lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H2; auto.
assert (0 > lookupR rf rs \/ 0 <= lookupR rf rs).
apply gt_or_le.
inversion_clear H0.
inversion_clear H4.
  (*case rs < 0 *)
  exists ((h, rf), (ch, is0)).
  constructor; assumption.
  (*case rs >= 0 *)
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H6; auto.
  inversion_clear H4.
  inversion_clear H7.
  assert (exists is : iseq, lookupCH ch f is).
  apply exists_iseq with ct x; auto.
  inversion_clear H7.
  exists ((h, rf), (ch, x0)).
  constructor; auto.

(* bgtz rs f *)
induction s as [h rf].
assert ((lookupR rf rs <= 0 -> a' ct (h, rf)) /\
       (lookupR rf rs > 0 ->
        exists theta : codespec,
          lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H2; auto.
assert (lookupR rf rs <= 0 \/ lookupR rf rs > 0).
apply le_or_gt.
inversion_clear H0.
inversion_clear H4.
  (*case rs <= 0 *)
  exists ((h, rf), (ch, is0)).
  constructor; assumption.
  (*case rs > 0 *)
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H6; auto.
  inversion_clear H4.
  inversion_clear H7.
  assert (exists is : iseq, lookupCH ch f is).
  apply exists_iseq with ct x; auto.
  inversion_clear H7.
  exists ((h, rf), (ch, x0)).
  constructor; auto.

(* blez rs f *)
induction s as [h rf].
assert ((lookupR rf rs > 0 -> a' ct (h, rf)) /\
       (lookupR rf rs <= 0 ->
        exists theta : codespec,
          lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H2; auto.
assert (lookupR rf rs > 0  \/ lookupR rf rs <= 0).
apply gt_or_le.
inversion_clear H0.
inversion_clear H4.
  (*case rs > 0 *)
  exists ((h, rf), (ch, is0)).
  constructor; assumption.
  (*case rs <= rt *)
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H6; auto.
  inversion_clear H4.
  inversion_clear H7.
  assert (exists is : iseq, lookupCH ch f is).
  apply exists_iseq with ct x; auto.
  inversion_clear H7.
  exists ((h, rf), (ch, x0)).
  constructor; auto.

(*bltz rs f *)
induction s as [h rf].
assert ((lookupR rf rs >= 0 -> a' ct (h, rf)) /\
       (lookupR rf rs < 0 ->
        exists theta : codespec,
          lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H2; auto.
assert (0 <= lookupR rf rs \/ 0 > lookupR rf rs).
apply le_or_gt.
inversion_clear H0.
inversion_clear H4.
  (*case rs >= 0 *)
  exists ((h, rf), (ch, is0)).
  constructor; assumption.
  (*case rs < 0  *)
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H6; auto.
  inversion_clear H4.
  inversion_clear H7.
  assert (exists is : iseq, lookupCH ch f is).
  apply exists_iseq with ct x; auto.
  inversion_clear H7.
  exists ((h, rf), (ch, x0)).
  constructor; auto.

(* c in addu ...; *)
assert (exists s' : state, next c s s' /\ a' ct s').
apply H2; auto.
inversion_clear H0.
inversion_clear H4.
exists (x, (ch, is0)).
constructor; auto.
Save.


(** *** Preservation lemma **)

(* well formed codeheap implies all iseq in the codeheap are well formed *)
Lemma wfch0_wfiseq :
  forall (ct0 ct : codeheapty) (ch : codeheap) (l : lab) 
    (theta : codespec) (is : iseq),
    WF_CH_0 ct0 ct ch -> lookupCT ct l theta -> lookupCH ch l is ->
    Infer (itpx2 (itpx1 theta) ct0) is.
Proof.
intros.
induction H as [emp | nonemp].
inversion_clear H0.
assert (f = l \/ f <> l).
apply eqorneq_nat.
inversion_clear H3.
  (* case f =l *)
  rewrite H4 in H0.
  rewrite H4 in H1.
  inversion H0.
  inversion H1.
  rewrite H13 in H.
  rewrite H8 in H; assumption.
  assert (l = l); auto.
  contradiction.
  assert (l = l); auto.
  contradiction.
  (* case f <> l *)
  inversion H0.
  contradiction.
  inversion H1.
  contradiction.
  apply IHWF_CH_0.
  assumption.
  assumption.
Save.

Lemma wfch_wfiseq :
  forall (ct : codeheapty) (ch : codeheap) (l : lab) (theta : codespec) (is : iseq),
    WF_CH ct ch -> lookupCT ct l theta -> lookupCH ch l is -> 
    Infer (itpx2 (itpx1 theta) ct) is.
Proof.
intros.
inversion_clear H.
apply wfch0_wfiseq with ct ch l.
assumption.
assumption.
assumption.
Save.

(** preservation lemma *)
Lemma preservation :
  forall (ch : codeheap) (ct : codeheapty) (a : interp)
    (h h' : heap) (rf rf' : rfile) (is is' : iseq),
    WF_WLD ct a ((h, rf), (ch, is)) ->
    step ((h, rf), (ch, is)) ((h', rf'), (ch, is')) ->
    exists a' : interp, WF_WLD ct a' ((h', rf'), (ch, is')).
Proof.
intros.
inversion_clear H.
inversion H2.
(* j f *)
rewrite <- H5 in H0.
assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
apply H; auto.
inversion_clear H6.
inversion_clear H7.
inversion H0.
rewrite <- H13; rewrite <- H14.
exists (itpx2 (itpx1 x) ct).
constructor.
unfold itpx2.
split.
apply psi_eq_subseteq.
assumption.
apply wfch_wfiseq with ch f; assumption.
assumption.

(* jal f fret *)
rewrite <- H5 in H0.
assert (exists theta : codespec,
        lookupCT ct f theta /\ itpx1 theta ct (h, updateR rf ra fret)).
apply H; assumption.
inversion_clear H6.
inversion_clear H7.
inversion H0.
rewrite <- H14.
exists (itpx2 (itpx1 x) ct).
constructor.
unfold itpx2.
split.
apply psi_eq_subseteq.
assumption.
apply wfch_wfiseq with ch f; assumption.
assumption.

(* jalr rd rs fret *)
rewrite <- H5 in H0.
assert (exists theta : codespec,
        lookupCT ct (lookupR rf rs) theta /\
        itpx1 theta ct (h, updateR rf rd fret)).
apply H; assumption.
inversion_clear H6.
inversion_clear H7.
inversion H0.
rewrite <- H15.
exists (itpx2 (itpx1 x) ct).
constructor.
unfold itpx2.
split.
apply psi_eq_subseteq.
assumption.
apply wfch_wfiseq with ch (lookupR rf rs); auto.
assumption.

(* jr rs *)
rewrite <- H5 in H0.
assert (exists theta : codespec,
        lookupCT ct (lookupR rf rs) theta /\
        itpx1 theta ct (h, rf)).
apply H; assumption.
inversion_clear H6.
inversion_clear H7.
inversion H0.
rewrite <- H15.
exists (itpx2 (itpx1 x) ct).
constructor.
unfold itpx2.
split.
apply psi_eq_subseteq.
rewrite <- H13; rewrite <- H14; auto.
apply wfch_wfiseq with ch (lookupR rf rs); auto.
rewrite H15; rewrite H14; auto.
assumption.
(* beq rs rt f *)
rewrite <- H6 in H0.
assert ((lookupR rf rs <> lookupR rf rt -> a' ct (h, rf)) /\
      (lookupR rf rs = lookupR rf rt ->
       exists theta : codespec,
         lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H.
clear H; assumption.
inversion_clear H7.
inversion H0.
  (*case rs <> rt *)
  rewrite <- H18; rewrite <- H18 in H10; rewrite H19 in H4.
  rewrite <- H17.
  clear H; clear H9; clear H0; clear H5; clear H6.
  clear H7; clear H11; clear H12; clear H13; clear H14.
  clear H15; clear H16; clear H17; clear H18; clear H19.
  exists a'.
  constructor.
  apply H8; auto.
  assumption.
  assumption.
  (* case rs = rt *)
  rewrite <- H17; rewrite <- H18; rewrite <- H18 in H11.
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H9; assumption.
  inversion_clear H21.
  inversion_clear H22.
  exists (itpx2 (itpx1 x) ct).
  constructor.
  unfold itpx2.
  split.
  apply psi_eq_subseteq.
  assumption.
  apply wfch_wfiseq with ch f; auto.
  assumption.
  inversion_clear H10.
  
(* bne rs rt f *)
rewrite <- H6 in H0.
assert ((lookupR rf rs = lookupR rf rt -> a' ct (h, rf)) /\
      (lookupR rf rs <> lookupR rf rt ->
       exists theta : codespec,
         lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H.
clear H; assumption.
inversion_clear H7.
inversion H0.
  (*case rs = rt *)
  rewrite <- H18; rewrite <- H18 in H10; rewrite H19 in H4.
  rewrite <- H17.
  exists a'.
  constructor.
  apply H8; auto.
  assumption.
  assumption.
  (* case rs <> rt *)
  rewrite <- H17; rewrite <- H18; rewrite <- H18 in H11.
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H9; assumption.
  inversion_clear H21.
  inversion_clear H22.
  exists (itpx2 (itpx1 x) ct).
  constructor.
  unfold itpx2.
  split.
  apply psi_eq_subseteq.
  assumption.
  apply wfch_wfiseq with ch f; auto.
  assumption.
  inversion_clear H10.

(* bgez rs f *)
rewrite <- H6 in H0.
assert ((lookupR rf rs < 0 -> a' ct (h, rf)) /\
      (lookupR rf rs >= 0 ->
       exists theta : codespec,
         lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H.
clear H; assumption.
inversion_clear H7.
inversion H0.
  (*case rs < 0 *)
  rewrite <- H17; rewrite <- H17 in H10; rewrite H18 in H4.
  rewrite <- H16.
  exists a'.
  constructor.
  apply H8; auto.
  assumption.
  assumption.
  (* case rs >= 0  *)
  rewrite <- H16; rewrite <- H17; rewrite <- H17 in H11.
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H9; assumption.
  inversion_clear H20.
  inversion_clear H21.
  exists (itpx2 (itpx1 x) ct).
  constructor.
  unfold itpx2.
  split.
  apply psi_eq_subseteq.
  assumption.
  apply wfch_wfiseq with ch f; auto.
  assumption.
  inversion_clear H10.

(* bgtz rs f *)
rewrite <- H6 in H0.
assert ((lookupR rf rs <= 0 -> a' ct (h, rf)) /\
      (lookupR rf rs > 0 ->
       exists theta : codespec,
         lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H.
clear H; assumption.
inversion_clear H7.
inversion H0.
  (*case rs <= 0 *)
  rewrite <- H17; rewrite <- H17 in H10; rewrite H18 in H4.
  rewrite <- H16.
  exists a'.
  constructor.
  apply H8; auto.
  assumption.
  assumption.
  (* case rs > 0  *)
  rewrite <- H16; rewrite <- H17; rewrite <- H17 in H11.
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H9; assumption.
  inversion_clear H20.
  inversion_clear H21.
  exists (itpx2 (itpx1 x) ct).
  constructor.
  unfold itpx2.
  split.
  apply psi_eq_subseteq.
  assumption.
  apply wfch_wfiseq with ch f; auto.
  assumption.
  inversion_clear H10.

(* blez rs f *)
rewrite <- H6 in H0.
assert ((lookupR rf rs > 0 -> a' ct (h, rf)) /\
      (lookupR rf rs <= 0 ->
       exists theta : codespec,
         lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H.
clear H; assumption.
inversion_clear H7.
inversion H0.
  (*case rs > 0 *)
  rewrite <- H17; rewrite <- H17 in H10; rewrite H18 in H4.
  rewrite <- H16.
  exists a'.
  constructor.
  apply H8; auto.
  assumption.
  assumption.
  (* case rs <= 0  *)
  rewrite <- H16; rewrite <- H17; rewrite <- H17 in H11.
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H9; assumption.
  inversion_clear H20.
  inversion_clear H21.
  exists (itpx2 (itpx1 x) ct).
  constructor.
  unfold itpx2.
  split.
  apply psi_eq_subseteq.
  assumption.
  apply wfch_wfiseq with ch f; auto.
  assumption.
  inversion_clear H10.

(* bltz rs f *)
rewrite <- H6 in H0.
assert ((lookupR rf rs >=  0 -> a' ct (h, rf)) /\
      (lookupR rf rs < 0 ->
       exists theta : codespec,
         lookupCT ct f theta /\ itpx1 theta ct (h, rf))).
apply H.
clear H; assumption.
inversion_clear H7.
inversion H0.
  (*case rs >= 0 *)
  rewrite <- H17; rewrite <- H17 in H10; rewrite H18 in H4.
  rewrite <- H16.
  exists a'.
  constructor.
  apply H8; auto.
  assumption.
  assumption.
  (* case rs < 0  *)
  rewrite <- H16; rewrite <- H17; rewrite <- H17 in H11.
  assert (exists theta : codespec, lookupCT ct f theta /\ itpx1 theta ct (h, rf)).
  apply H9; assumption.
  inversion_clear H20.
  inversion_clear H21.
  exists (itpx2 (itpx1 x) ct).
  constructor.
  unfold itpx2.
  split.
  apply psi_eq_subseteq.
  assumption.
  apply wfch_wfiseq with ch f; auto.
  assumption.
  inversion_clear H10.

(* c in addu subu ... *)
rewrite <- H6 in H0.
assert (exists s' : state, next c (h, rf) s' /\ a' ct s').
apply H; assumption.
inversion_clear H7.
inversion_clear H8.
inversion H0.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite <- H13 in H7.
inversion_clear H7.
rewrite H15 in H4.
exists a'.
constructor.
assert ((h', rf') = x).
apply next_state_eq with c (h, rf); auto.
rewrite H16; auto.
auto.
auto.
Save.






















